
#include "Manifolds/CFixedRank2Factors/CFixedRank2Factors.h"

/*Define the namespace*/
namespace ROPTLIB{

	CFixedRank2Factors::CFixedRank2Factors(integer inm, integer inn, integer inr) : ProductManifold(2,
		new Euclidean(2 * inm, inr), static_cast<integer> (1), new Euclidean(2 * inn, inr), static_cast<integer> (1))
	{
		m = inm;
		n = inn;
		r = inr;
		name.assign("Complex fixed-rank manifold by 2-factor representation");
		IsIntrApproach = true;
		delete EMPTYEXTR;
		delete EMPTYINTR;
		EMPTYEXTR = new CFR2Vector(m, n, r);
		EMPTYINTR = new CFR2Vector(m, n - r, r);
	};

	CFixedRank2Factors::~CFixedRank2Factors()
	{
		for (integer i = 0; i < numofmani; i++)
		{
			delete manifolds[i];
		}
	};

	void CFixedRank2Factors::ObtainIntr(Variable *x, Vector *etax, Vector *result) const
	{
		CFR2Variable *CFR2x = dynamic_cast<CFR2Variable *> (x);
		Variable* G = CFR2x->GetElement(0);
		Variable* H = CFR2x->GetElement(1);
		if (!G->TempDataExist("HHR"))
		{
			CpxNStQOrth::ComputeHHR(G);
		}
		if (!H->TempDataExist("HHR"))
		{
			CpxNStQOrth::ComputeHHR(H);
		}

		const realdp *etaxGTV = etax->ObtainReadData();
		const realdp *etaxHTV = etaxGTV + 2 * m * r;
		const SharedSpace *GHHR = G->ObtainReadTempData("HHR");
		const SharedSpace *GHHRTau = G->ObtainReadTempData("HHRTau");
		const SharedSpace *HHHR = H->ObtainReadTempData("HHR");
		const SharedSpace *HHHRTau = H->ObtainReadTempData("HHRTau");
		realdp *resultTV = result->ObtainWriteEntireData();
		const realdp *ptrGHHR = GHHR->ObtainReadData();
		const realdp *ptrGHHRTau = GHHRTau->ObtainReadData();
		const realdp *ptrHHHR = HHHR->ObtainReadData();
		const realdp *ptrHHHRTau = HHHRTau->ObtainReadData();

		/*for G component*/
		integer info;
		integer lwork = -1, Length = 2 * m * r;
		realdpcomplex lworkopt;
		realdpcomplex *Gtempspace = new realdpcomplex[2 * m * r];
		// compute the size of space required in the ormqr
		unmqr_(GLOBAL::L, GLOBAL::C, &m, &r, &r, (realdpcomplex *) (const_cast<realdp *> (ptrGHHR)), &m,
			(realdpcomplex *)(const_cast<realdp *> (ptrGHHRTau)), Gtempspace, &m, &lworkopt, &lwork, &info);

		lwork = static_cast<integer> (lworkopt.r);
		realdpcomplex *work = new realdpcomplex[lwork];
		// tempspace <- etaxTV, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		copy_(&Length, const_cast<realdp *> (etaxGTV), &GLOBAL::IONE, (realdp *)Gtempspace, &GLOBAL::IONE);
		// tempspace <- Q^T * tempspace, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html

		unmqr_(GLOBAL::L, GLOBAL::C, &m, &r, &r, (realdpcomplex *)(const_cast<realdp *> (ptrGHHR)), &m,
			(realdpcomplex *)(const_cast<realdp *> (ptrGHHRTau)), Gtempspace, &m, work, &lwork, &info);

		delete[] work;
		realdpcomplex sign = { 0, 0 };
		for (integer i = 0; i < r; i++)
		{
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrGHHR)[i + m * i].r >= 0) ? 1 : -1);
			scal_(&r, &sign, Gtempspace + i, &m);
		}

		realdpcomplex *LG = new realdpcomplex[r * r];
		for (integer i = 0; i < r; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				LG[j + i * r].r = 0;
				LG[j + i * r].i = 0;
			}
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrGHHR)[i + m * i].r >= 0) ? 1 : -1);
			for (integer j = i; j < r; j++)
			{
				LG[j + i * r].r = ((realdpcomplex *)ptrGHHR)[i + m * j].r * sign.r;
				LG[j + i * r].i = ((realdpcomplex *)ptrGHHR)[i + m * j].i * (-sign.r);
			}
		}

		/*for H component*/
		realdpcomplex *Htempspace = new realdpcomplex[2 * n * r];
		lwork = -1;
		unmqr_(GLOBAL::L, GLOBAL::C, &n, &r, &r, (realdpcomplex *)(const_cast<realdp *> (ptrHHHR)), &n,
			(realdpcomplex *)(const_cast<realdp *> (ptrHHHRTau)), Htempspace, &n, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt.r);
		work = new realdpcomplex[lwork];
		Length = 2 * n * r;
		// tempspace <- etaxTV, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		copy_(&Length, const_cast<realdp *> (etaxHTV), &GLOBAL::IONE, (realdp *)Htempspace, &GLOBAL::IONE);
		// tempspace <- Q^T * tempspace, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		unmqr_(GLOBAL::L, GLOBAL::C, &n, &r, &r, (realdpcomplex *)(const_cast<realdp *> (ptrHHHR)), &n,
			(realdpcomplex *)(const_cast<realdp *> (ptrHHHRTau)), Htempspace, &n, work, &lwork, &info);
		delete[] work;
		sign.r = 0;
		sign.i = 0;
		for (integer i = 0; i < r; i++)
		{
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrHHHR)[i + n * i].r >= 0) ? 1 : -1);
			scal_(&r, &sign, Htempspace + i, &n);
		}

		realdpcomplex *LH = new realdpcomplex[r * r];
		for (integer i = 0; i < r; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				LH[j + i * r].r = 0;
				LH[j + i * r].i = 0;
			}
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrHHHR)[i + n * i].r >= 0) ? 1.0 : -1.0);
			for (integer j = i; j < r; j++)
			{
				LH[j + i * r].r = ((realdpcomplex *)ptrHHHR)[i + n * j].r * sign.r;
				LH[j + i * r].i = ((realdpcomplex *)ptrHHHR)[i + n * j].i * (-sign.r);
			}
		}

		/*Compute representations*/

		realdpcomplex *GtempspaceL = new realdpcomplex[m * r];

		Matrix GMtL((realdp *)GtempspaceL, m, r), MLH((realdp *)LH, r, r), MGtempspace((realdp *)Gtempspace, m, r);
		Matrix::CGEMM(GLOBAL::ZONE, MGtempspace, false, MLH, false, GLOBAL::ZZERO, GMtL);

		realdpcomplex *HtempspaceL = new realdpcomplex[n * r];

		Matrix HMtL((realdp *)HtempspaceL, n, r), MLG((realdp *)LG, r, r), MHtempspace((realdp *)Htempspace, n, r);
		Matrix::CGEMM(GLOBAL::ZONE, MHtempspace, false, MLG, false, GLOBAL::ZZERO, HMtL);
		delete[] LG;
		delete[] LH;
		delete[] Gtempspace;
		delete[] Htempspace;

		//ForDebug::Print("GtempspaceL", (realdp *)GtempspaceL, 2 * m, r);//---
		//ForDebug::Print("HtempspaceL", (realdp *)HtempspaceL, 2 * n, r);//---

		realdp r2 = static_cast<realdp> (sqrt(2.0));
		integer idx = 0;
		for (integer i = 0; i < r; i++)
		{
			for (integer j = 0; j < r; j++)
			{
				resultTV[idx] = (GtempspaceL[j + i * m].r + HtempspaceL[i + j * n].r) / r2;
				idx++;
				resultTV[idx] = (GtempspaceL[j + i * m].i - HtempspaceL[i + j * n].i) / r2;
				idx++;
			}
		}

		for (integer i = 0; i < r; i++)
		{
			for (integer j = r; j < m; j++)
			{
				resultTV[idx] = GtempspaceL[j + i * m].r;
				idx++;
				resultTV[idx] = GtempspaceL[j + i * m].i;
				idx++;
			}
		}
		for (integer i = 0; i < r; i++)
		{
			for (integer j = r; j < n; j++)
			{
				resultTV[idx] = HtempspaceL[j + i * n].r;
				idx++;
				resultTV[idx] = HtempspaceL[j + i * n].i;
				idx++;
			}
		}
		delete[] GtempspaceL;
		delete[] HtempspaceL;
	};

	void CFixedRank2Factors::ObtainExtr(Variable *x, Vector *intretax, Vector *result) const
	{
		CFR2Variable *CFR2x = dynamic_cast<CFR2Variable *> (x);
		Variable* G = CFR2x->GetElement(0);
		Variable* H = CFR2x->GetElement(1);
		if (!G->TempDataExist("HHR"))
		{
			CpxNStQOrth::ComputeHHR(G);
		}
		if (!H->TempDataExist("HHR"))
		{
			CpxNStQOrth::ComputeHHR(H);
		}

		const realdp *ptrintretax = intretax->ObtainReadData();
		const SharedSpace *GHHR = G->ObtainReadTempData("HHR");
		const SharedSpace *GHHRTau = G->ObtainReadTempData("HHRTau");
		const SharedSpace *HHHR = H->ObtainReadTempData("HHR");
		const SharedSpace *HHHRTau = H->ObtainReadTempData("HHRTau");
		realdpcomplex *Gresult = (realdpcomplex *) (result->ObtainWriteEntireData());
		realdpcomplex *Hresult = Gresult + m * r;
		const realdp *ptrGHHR = GHHR->ObtainReadData();
		const realdp *ptrGHHRTau = GHHRTau->ObtainReadData();
		const realdp *ptrHHHR = HHHR->ObtainReadData();
		const realdp *ptrHHHRTau = HHHRTau->ObtainReadData();



		integer info;
		integer idx = 0;
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		for (integer i = 0; i < r; i++)
		{
			for (integer j = 0; j < r; j++)
			{
				Gresult[j + i * m].r = ptrintretax[idx] / r2;
				Hresult[i + j * n].r = Gresult[j + i * m].r;
				idx++;
				Gresult[j + i * m].i = ptrintretax[idx] / r2;
				Hresult[i + j * n].i = -Gresult[j + i * m].i;
				idx++;
			}
		}

		for (integer i = 0; i < r; i++)
		{
			for (integer j = r; j < m; j++)
			{
				Gresult[j + i * m].r = ptrintretax[idx];
				idx++;
				Gresult[j + i * m].i = ptrintretax[idx];
				idx++;
			}
		}
		for (integer i = 0; i < r; i++)
		{
			for (integer j = r; j < n; j++)
			{
				Hresult[j + i * n].r = ptrintretax[idx];
				idx++;
				Hresult[j + i * n].i = ptrintretax[idx];
				idx++;
			}
		}

		//ForDebug::Print("Gresult", (realdp *)Gresult, 2 * m, r);//---
		//ForDebug::Print("Hresult", (realdp *)Hresult, 2 * n, r);//---

		/*for eta_G*/

		realdpcomplex sign = { 0, 0 };
		for (integer i = 0; i < r; i++)
		{
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrGHHR)[i + m * i].r >= 0) ? 1 : -1);
			// result(i, :) <- sign * result(i, :), details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
			scal_(&r, &sign, Gresult + i, &m);
		}
		integer lwork = -1;
		realdpcomplex lworkopt;
		// compute the size of space required in the ormqr
		unmqr_(GLOBAL::L, GLOBAL::N, &m, &r, &r, (realdpcomplex *)(const_cast<realdp *> (ptrGHHR)), &m,
			(realdpcomplex *)(const_cast<realdp *> (ptrGHHRTau)), Gresult, &m, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt.r);

		realdpcomplex *work = new realdpcomplex[lwork];
		// resultTV <- Q * resultTV, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		unmqr_(GLOBAL::L, GLOBAL::N, &m, &r, &r, (realdpcomplex *)(const_cast<realdp *> (ptrGHHR)), &m,
			(realdpcomplex *)(const_cast<realdp *> (ptrGHHRTau)), Gresult, &m, work, &lwork, &info);
		delete[] work;

		/*for eta_H*/
		sign.r = 0;
		sign.i = 0;
		for (integer i = 0; i < r; i++)
		{
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrHHHR)[i + n * i].r >= 0) ? 1 : -1);
			// result(i, :) <- sign * result(i, :), details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
			scal_(&r, &sign, Hresult + i, &n);
		}
		lwork = -1;
		// compute the size of space required in the ormqr
		unmqr_(GLOBAL::L, GLOBAL::N, &n, &r, &r, (realdpcomplex *)(const_cast<realdp *> (ptrHHHR)), &n,
			(realdpcomplex *)(const_cast<realdp *> (ptrHHHRTau)), Hresult, &n, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt.r);

		work = new realdpcomplex[lwork];
		// resultTV <- Q * resultTV, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		unmqr_(GLOBAL::L, GLOBAL::N, &n, &r, &r, (realdpcomplex *)(const_cast<realdp *> (ptrHHHR)), &n,
			(realdpcomplex *)(const_cast<realdp *> (ptrHHHRTau)), Hresult, &n, work, &lwork, &info);
		delete[] work;

		realdpcomplex *LG = new realdpcomplex[2 * r * r + n * r + m * r];
		realdpcomplex *LH = LG + r * r;
		realdpcomplex *Gr_T = LH + r * r;
		realdpcomplex *Hr_T = Gr_T + m * r;
		for (integer i = 0; i < r; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				LG[j + i * r].r = 0;
				LG[j + i * r].i = 0;
			}
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrGHHR)[i + m * i].r >= 0) ? 1 : -1);
			for (integer j = i; j < r; j++)
			{
				LG[j + i * r].r = ((realdpcomplex *)ptrGHHR)[i + m * j].r * sign.r;
				LG[j + i * r].i = ((realdpcomplex *)ptrGHHR)[i + m * j].i * (-sign.r);
			}
		}

		/*Gr_T <-  Gresult transpose conjugate*/
		for (integer i = 0; i < m; i++)
		{
			for (integer j = 0; j < r; j++)
			{
				Gr_T[j + i * r].r = ((realdpcomplex *)Gresult)[i + j * m].r;
				Gr_T[j + i * r].i = -((realdpcomplex *)Gresult)[i + j * m].i;
			}
		}

		for (integer i = 0; i < r; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				LH[j + i * r].r = 0;
				LH[j + i * r].i = 0;
			}
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrHHHR)[i + n * i].r >= 0) ? 1 : -1);
			for (integer j = i; j < r; j++)
			{
				LH[j + i * r].r = ((realdpcomplex *)ptrHHHR)[i + n * j].r * sign.r;
				LH[j + i * r].i = ((realdpcomplex *)ptrHHHR)[i + n * j].i * (-sign.r);
			}
		}

		/*Hr_T <-  Gresult transpose conjugate*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < r; j++)
			{
				Hr_T[j + i * r].r = ((realdpcomplex *)Hresult)[i + j * n].r;
				Hr_T[j + i * r].i = -((realdpcomplex *)Hresult)[i + j * n].i;
			}
		}

		/*solve linear system LH^H M = Gr_T, the solution M is stored in Gr_T*/
		trtrs_(GLOBAL::L, GLOBAL::C, GLOBAL::N, &r, &m, LH, &r, Gr_T, &r, &info);
		/*resultTV <-  r_T transpose conjugate*/
		for (integer i = 0; i < m; i++)
		{
			for (integer j = 0; j < r; j++)
			{
				((realdpcomplex *)Gresult)[i + j * m].r = Gr_T[j + i * r].r;
				((realdpcomplex *)Gresult)[i + j * m].i = -Gr_T[j + i * r].i;
			}
		}

		/*solve linear system LH^H M = Gr_T, the solution M is stored in Gr_T*/
		trtrs_(GLOBAL::L, GLOBAL::C, GLOBAL::N, &r, &n, LG, &r, Hr_T, &r, &info);
		/*resultTV <-  r_T transpose conjugate*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < r; j++)
			{
				((realdpcomplex *)Hresult)[i + j * n].r = Hr_T[j + i * r].r;
				((realdpcomplex *)Hresult)[i + j * n].i = -Hr_T[j + i * r].i;
			}
		}
		delete[] LG;

	};

	void CFixedRank2Factors::Retraction(Variable *x, Vector *etax, Variable *result, realdp stepsize) const
	{
		if (IsIntrApproach)
		{
			Vector *exetax = EMPTYEXTR->ConstructEmpty();
			ObtainExtr(x, etax, exetax);
			SetIsIntrApproach(false);
			ProductManifold::Retraction(x, exetax, result, stepsize);
			SetIsIntrApproach(true);
			delete exetax;
		}
		else
		{
			ProductManifold::Retraction(x, etax, result, stepsize);
		}
	};

	void CFixedRank2Factors::coTangentVector(Variable *x, Vector *etax, Variable *y, Vector *xiy, Vector *result) const
	{
		xiy->CopyTo(result);
		printf("Warning: CFixedRank2Factors::coTangentVector has not been done!\n");
	};

	void CFixedRank2Factors::DiffRetraction(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result, bool IsEtaXiSameDir) const
	{
		xix->CopyTo(result);
	};

	void CFixedRank2Factors::ExtrProjection(Variable *x, Vector *etax, Vector *result) const
	{
		CFR2Variable *CFR2x = dynamic_cast<CFR2Variable *> (x);
		Variable* G = CFR2x->GetElement(0);
		Variable* H = CFR2x->GetElement(1);
		const realdp *Gptr = G->ObtainReadData();
		const realdp *Hptr = H->ObtainReadData();
		const realdp *etaxGTV = etax->ObtainReadData();
		const realdp *etaxHTV = etaxGTV + 2 * m * r;
		

		/*Compute (H^*H)^{-1} H^* eta_H */
		realdpcomplex *HH = new realdpcomplex[4 * r * r];
		realdpcomplex *HV = HH + r * r;
		realdpcomplex *GG = HV + r * r;
		realdpcomplex *GV = GG + r * r;
		Matrix MH(Hptr, n, r), MHH((realdp*)HH, r, r), MV(etaxHTV, n, r), MHV((realdp*)HV, r, r);
		// MHH <- MH^* MH
		Matrix::CGEMM(GLOBAL::ZONE, MH, true, MH, false, GLOBAL::ZZERO, MHH);
		// MHV <- MH^* MV
		Matrix::CGEMM(GLOBAL::ZONE, MH, true, MV, false, GLOBAL::ZZERO, MHV);

		integer info;
		// solve for (MHH)^{-1} MHV
		potrf_(GLOBAL::L, &r, HH, &r, &info);
		potrs_(GLOBAL::L, &r, &r, HH, &r, HV, &r, &info);

		if (info != 0)
		{
			printf("warning: potrs failed in CFixedRank2Factors::ExtrProjection with info:%d!\n", info);
		}

		/*Compute (G^*G)^{-1} G^* eta_G */
		Matrix MG(Gptr, m, r), MGG((realdp*)GG, r, r), MV2(etaxGTV, m, r), MGV((realdp*)GV, r, r);
		// MGG <- MG^* MG
		Matrix::CGEMM(GLOBAL::ZONE, MG, true, MG, false, GLOBAL::ZZERO, MGG);
		// MGV <- MG^* MV2
		Matrix::CGEMM(GLOBAL::ZONE, MG, true, MV2, false, GLOBAL::ZZERO, MGV);

		// solve for (MGG)^{-1} MGV
		potrf_(GLOBAL::L, &r, GG, &r, &info);
		potrs_(GLOBAL::L, &r, &r, GG, &r, GV, &r, &info);

		if (info != 0)
		{
			printf("warning: potrs failed in CFixedRank2Factors::ExtrProjection with info:%d!\n", info);
		}

		/*GV = (- HV^* + GV)/2 */
		for (integer i = 0; i < r; i++)
		{
			for (integer j = 0; j < r; j++)
			{
				GV[i + j * r].r -= HV[j + i * r].r;
				GV[i + j * r].r /= 2.0;
				GV[i + j * r].i += HV[j + i * r].i;
				GV[i + j * r].i /= 2.0;
			}
		}

		etax->CopyTo(result);
		realdp *resultTV = result->ObtainWritePartialData();
		Matrix MGresultTV(resultTV, m, r), MHresultTV(resultTV + 2 * m * r, n, r);
		Matrix::CGEMM(GLOBAL::ZNONE, MG, false, MGV, false, GLOBAL::ZONE, MGresultTV);
		Matrix::CGEMM(GLOBAL::ZONE, MH, false, MGV, true, GLOBAL::ZONE, MHresultTV);

		delete[] HH;
	};

	void CFixedRank2Factors::EucGradToGrad(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		if (prob->GetUseHess())
		{
			Vector *segf = egf->ConstructEmpty();
			segf->NewMemoryOnWrite(); // I don't remember the reason. It seems to be required.
			egf->CopyTo(segf);
			SharedSpace *Sharedegf = new SharedSpace(segf);
			x->AddToTempData("EGrad", Sharedegf);
		}
		ExtrProjection(x, egf, gf);
	};

	void CFixedRank2Factors::EucHvToHv(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{
		ExtrProjection(x, exix, xix);
	};
}; /*end of ROPTLIB namespace*/
