
#include "Manifolds/CpxNStQOrth/CpxNStQOrth.h"

/*Define the namespace*/
namespace ROPTLIB{

	CpxNStQOrth::CpxNStQOrth(integer r, integer c)
	{
		n = r;
		p = c;
		IsIntrApproach = true;
		HasHHR = false;
        HasLockCon = false;
		UpdBetaAlone = false;
		name.assign("Complex Noncompact Stiefel manifold quotient unitary group");
		IntrinsicDim = 2 * r * c - c * c;
		ExtrinsicDim = 2 * r * c;
		EMPTYEXTR = new CSOVector(2 * r, c);
		EMPTYINTR = new CSOVector(IntrinsicDim);
	};

	CpxNStQOrth::~CpxNStQOrth(void)
	{
		delete EMPTYEXTR;
		delete EMPTYINTR;
	};

	realdp CpxNStQOrth::Metric(Variable *x, Vector *etax, Vector *xix) const
	{
		//Vector *exetax = EMPTYEXTR->ConstructEmpty();
		//Vector *exxix = EMPTYEXTR->ConstructEmpty();
		//ObtainExtr(x, etax, exetax);
		//ObtainExtr(x, xix, exxix);
		//realdp result = Manifold::Metric(x, exetax, exxix);
		//delete exetax;
		//delete exxix;
		if (IsIntrApproach)
			return Manifold::Metric(x, etax, xix);
		printf("Warning: Metric for extrinsic representation has not been done!\n");
		return 0;
	};

	void CpxNStQOrth::CheckParams(void) const
	{
		Manifold::CheckParams();
		printf("%s PARAMETERS:\n", name.c_str());
		if (p == 1)
			printf("n           :%15d\n", n);
		else
		{
			printf("n           :%15d,\t", n);
			printf("p           :%15d\n", p);
		}
	};

	void CpxNStQOrth::EucGradToGrad(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		if (prob->GetUseHess())
		{
			Vector *segf = egf->ConstructEmpty();
			segf->NewMemoryOnWrite(); // I don't remember the reason. It seems to be required.
			egf->CopyTo(segf);
			SharedSpace *Sharedegf = new SharedSpace(segf);
			x->AddToTempData("EGrad", Sharedegf);
		}
		ExtrProjection(x, egf, gf);
	};

	void CpxNStQOrth::EucHvToHv(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{/*The following is the EucHvToHv for the metric tr(eta^H xi), not the used metric. It needs be modified.*/
		const realdp *xM = x->ObtainReadData();
		const SharedSpace *Sharedegf = x->ObtainReadTempData("EGrad");
		Vector *segf = Sharedegf->GetSharedElement();
		const realdp *egf = segf->ObtainReadData();

		realdpcomplex *XHX = new realdpcomplex[2 * p * p];
		realdpcomplex *XHM = XHX + p * p;

		// XHX <- XM^H XM
		Matrix MxM(xM, n, p), MXHX((realdp*)XHX, p, p), Megf(egf, n, p), MXHM((realdp*)XHM, p, p);
		Matrix::CGEMM(GLOBAL::ZONE, MxM, true, MxM, false, GLOBAL::ZZERO, MXHX);
		// XHM <- XM^H egf
		Matrix::CGEMM(GLOBAL::ZONE, MxM, true, Megf, false, GLOBAL::ZZERO, MXHM);

		// XHM <- XHM - XHM^H
		for (integer i = 0; i < p; i++)
		{
			for (integer j = i; j < p; j++)
			{
				XHM[i + j * p].r -= XHM[j + i * p].r;
				XHM[i + j * p].i += XHM[j + i * p].i;
				XHM[j + i * p].r = -XHM[i + j * p].r;
				XHM[j + i * p].i = XHM[i + j * p].i;
			}
		}
		Matrix::CSYL(MXHX, MXHX, MXHM);

		exix->CopyTo(xix);
		realdp *xixTV = xix->ObtainWritePartialData();
		const realdp *etaxTV = etax->ObtainReadData();
		Matrix MetaxTV(etaxTV, n, p), MxixTV(xixTV, n, p);
		Matrix::CGEMM(GLOBAL::ZNONE, MetaxTV, false, MXHM, false,
			GLOBAL::ZONE, MxixTV);

		delete[] XHX;

		ExtrProjection(x, xix, xix);
	};

	void CpxNStQOrth::Projection(Variable *x, Vector *v, Vector *result) const
	{
		if (IsIntrApproach)
			IntrProjection(x, v, result);
		else
			ExtrProjection(x, v, result);
	};

	void CpxNStQOrth::Retraction(Variable *x, Vector *etax, Variable *result, realdp stepsize) const
	{
		if (IsIntrApproach)
		{
			Vector *exetax = EMPTYEXTR->ConstructEmpty();
			ObtainExtr(x, etax, exetax);
			Manifold::Retraction(x, exetax, result, 1);
			delete exetax;
		}
		else
		{
			Manifold::Retraction(x, etax, result, 1);
		}
	};

	void CpxNStQOrth::coTangentVector(Variable *x, Vector *etax, Variable *y, Vector *xiy, Vector *result) const
	{
		printf("TODO\n");//----
		xiy->CopyTo(result);
	};

	void CpxNStQOrth::DiffRetraction(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result, bool IsEtaXiSameDir) const
	{
		Vector *exxix = EMPTYEXTR->ConstructEmpty();
		Vector *exresult = EMPTYEXTR->ConstructEmpty();
		ObtainExtr(x, xix, exxix);
		ExtrProjection(y, exxix, exresult);
		ObtainIntr(y, exresult, result);
		delete exxix;
		delete exresult;
	};

	realdp CpxNStQOrth::Beta(Variable *x, Vector *etax) const
	{
		return 1;
	};


	void CpxNStQOrth::ComputeHHR(Variable *x)
	{
		const realdp *xM = x->ObtainReadData();
		SharedSpace *HouseHolderResult = new SharedSpace(2, x->Getsize()[0], x->Getsize()[1]);
		realdp *ptrHHR = HouseHolderResult->ObtainWriteEntireData();
		SharedSpace *HHRTau = new SharedSpace(1, x->Getsize()[1] * 2);
		realdp *tau = HHRTau->ObtainWriteEntireData();

		// for complex, N is the size divided by 2
		integer N = x->Getsize()[0] / 2, P = x->Getsize()[1], inc = 1;
		// ptrHHR <- xM, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		integer Length = 2 * N * P;
		copy_(&Length, const_cast<realdp *> (xM), &inc, ptrHHR, &inc);
		integer *jpvt = new integer[P];
		integer info;
		integer lwork = -1;
		realdpcomplex lworkopt;
		realdp *rwork = new realdp[P];
		// compute the size of space required in the geqp3
		geqp3_(&N, &P, (realdpcomplex *)ptrHHR, &N, jpvt, (realdpcomplex *)tau, &lworkopt, &lwork, rwork, &info);
		lwork = static_cast<integer> (lworkopt.r);
		realdpcomplex *work = new realdpcomplex[lwork];
		for (integer i = 0; i < P; i++)
			jpvt[i] = i + 1;
		// QR decomposition for ptrHHR using Householder reflections. Householder reflectors and R are stored in ptrHHR.
		// details: http://www.netlib.org/lapack/explore-html/db/de5/geqp3_8f.html
		geqp3_(&N, &P, (realdpcomplex *)ptrHHR, &N, jpvt, (realdpcomplex *)tau, work, &lwork, rwork, &info);

		x->AddToTempData("HHR", HouseHolderResult);
		x->AddToTempData("HHRTau", HHRTau);
		if (info < 0)
			printf("Error in qr decomposition!\n");
		for (integer i = 0; i < P; i++)
		{
			if (jpvt[i] != (i + 1))
				printf("Error in qf retraction!\n");
		}
		delete[] jpvt;
		delete[] work;
		delete[] rwork;
	};

	void CpxNStQOrth::ObtainIntr(Variable *x, Vector *etax, Vector *result) const
	{
		if (!x->TempDataExist("HHR"))
		{
			ComputeHHR(x);
		}

		const realdp *xM = x->ObtainReadData();
		const realdp *etaxTV = etax->ObtainReadData();
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
		realdp *resultTV = result->ObtainWriteEntireData();
		const realdp *ptrHHR = HHR->ObtainReadData();
		const realdp *ptrHHRTau = HHRTau->ObtainReadData();

		// for complex, N is the size divided by 2
		integer N = x->Getsize()[0] / 2, P = x->Getsize()[1], inc = 1, Length = 2 * N * P;
		integer info;
		integer lwork = -1;
		realdpcomplex lworkopt;
		realdpcomplex *tempspace = new realdpcomplex[n * p];
		// compute the size of space required in the ormqr
		unmqr_(GLOBAL::L, GLOBAL::C, &N, &P, &P, (realdpcomplex *)(const_cast<realdp *> (ptrHHR)), &N,
			(realdpcomplex *)(const_cast<realdp *> (ptrHHRTau)), tempspace, &N, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt.r);
		realdpcomplex *work = new realdpcomplex[lwork];
		// tempspace <- etaxTV, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		copy_(&Length, const_cast<realdp *> (etaxTV), &inc, (realdp *)tempspace, &inc);
		// tempspace <- Q^T * tempspace, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		unmqr_(GLOBAL::L, GLOBAL::C, &N, &P, &P, (realdpcomplex *)(const_cast<realdp *> (ptrHHR)), &N,
			(realdpcomplex *)(const_cast<realdp *> (ptrHHRTau)), tempspace, &N, work, &lwork, &info);
		realdpcomplex sign = { 0, 0 };
		for (integer i = 0; i < p; i++)
		{
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrHHR)[i + n * i].r >= 0) ? 1 : -1);
			scal_(&P, &sign, tempspace + i, &N);
		}

		realdpcomplex *L = new realdpcomplex[p * p];
		for (integer i = 0; i < p; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				L[j + i * p].r = 0;
				L[j + i * p].i = 0;
			}
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrHHR)[i + n * i].r >= 0) ? 1 : -1);
			for (integer j = i; j < p; j++)
			{
				L[j + i * p].r = ((realdpcomplex *)ptrHHR)[i + n * j].r * sign.r;
				L[j + i * p].i = ((realdpcomplex *)ptrHHR)[i + n * j].i * (-sign.r);
			}
		}

		realdpcomplex *tempspaceL = new realdpcomplex[n * p];

		Matrix MtL((realdp *)tempspaceL, n, p), ML((realdp *)L, p, p), Mtempspace((realdp *)tempspace, n, p);
		Matrix::CGEMM(GLOBAL::ZONE, Mtempspace, false, ML, false, GLOBAL::ZZERO, MtL);
		
		delete[] L;
		delete[] tempspace;

		/*Matrix MxM(xM, n, p), MetaxTV(etaxTV, n, p), Mtempspace((realdp *)tempspace, p, p, n);
		Matrix::CGEMM(GLOBAL::ZONE, MxM, true, MetaxTV, false, GLOBAL::ZZERO, Mtempspace);*/
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		realdp factor = 1;//-- sqrt(Manifold::Metric(x, x, x));
		integer idx = 0;
		for (integer i = 0; i < p; i++)
		{
			resultTV[idx] = tempspaceL[i + i * n].r / factor;
			idx++;
		}
		for (integer i = 0; i < p; i++)
		{
			for (integer j = i + 1; j < p; j++)
			{
				resultTV[idx] = r2 * tempspaceL[j + i * n].r / factor;
				idx++;
				resultTV[idx] = r2 * tempspaceL[j + i * n].i / factor;
				idx++;
			}
		}

		for (integer i = 0; i < p; i++)
		{
			for (integer j = p; j < n; j++)
			{
				resultTV[idx] = tempspaceL[j + i * n].r;
				idx++;
				resultTV[idx] = tempspaceL[j + i * n].i;
				idx++;
			}
		}
		delete[] work;
		delete[] tempspaceL;
	};

	void CpxNStQOrth::ObtainExtr(Variable *x, Vector *intretax, Vector *result) const
	{
		if (!x->TempDataExist("HHR"))
		{
			ComputeHHR(x);
		}

		const realdp *xM = x->ObtainReadData();
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
		const realdp *ptrHHR = HHR->ObtainReadData();
		const realdp *ptrHHRTau = HHRTau->ObtainReadData();
		const realdp *intretaxTV = intretax->ObtainReadData();
		realdp *resultTV = result->ObtainWriteEntireData();

		integer N = x->Getsize()[0] / 2, P = x->Getsize()[1], inc = 1, Length = 2 * N * P;
		integer info;
		integer idx = 0;
//		realdpcomplex *S = new realdpcomplex[p * p];
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		realdp factor = 1;//-- sqrt(Manifold::Metric(x, x, x));
		for (integer i = 0; i < p; i++)
		{
			((realdpcomplex *)resultTV)[i + i * n].r = intretaxTV[idx] * factor;
			((realdpcomplex *)resultTV)[i + i * n].i = 0;
			idx++;
		}

		for (integer i = 0; i < p; i++)
		{
			for (integer j = i + 1; j < p; j++)
			{
				((realdpcomplex *)resultTV)[j + i * n].r = intretaxTV[idx] / r2 * factor;
				((realdpcomplex *)resultTV)[i + j * n].r = intretaxTV[idx] / r2 * factor;
				idx++;
				((realdpcomplex *)resultTV)[j + i * n].i = intretaxTV[idx] / r2 * factor;
				((realdpcomplex *)resultTV)[i + j * n].i = -intretaxTV[idx] / r2 * factor;
				idx++;
			}
		}
		for (integer i = 0; i < p; i++)
		{
			for (integer j = p; j < n; j++)
			{
				((realdpcomplex *)resultTV)[j + i * n].r = intretaxTV[idx];
				idx++;
				((realdpcomplex *)resultTV)[j + i * n].i = intretaxTV[idx];
				idx++;
			}
		}

		realdpcomplex sign = { 0, 0 };
		for (integer i = 0; i < p; i++)
		{
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrHHR)[i + n * i].r >= 0) ? 1 : -1);
			// result(i, :) <- sign * result(i, :), details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
			scal_(&P, &sign, ((realdpcomplex *)resultTV) + i, &N);
		}
		integer lwork = -1;
		realdpcomplex lworkopt;
		// compute the size of space required in the ormqr
		unmqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (realdpcomplex *)(const_cast<realdp *> (ptrHHR)), &N,
			(realdpcomplex *)(const_cast<realdp *> (ptrHHRTau)), (realdpcomplex *)resultTV, &N, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt.r);

		realdpcomplex *work = new realdpcomplex[lwork];
		// resultTV <- Q * resultTV, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		unmqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (realdpcomplex *)(const_cast<realdp *> (ptrHHR)), &N,
			(realdpcomplex *)(const_cast<realdp *> (ptrHHRTau)), (realdpcomplex *)resultTV, &N, work, &lwork, &info);
		delete[] work;

		realdpcomplex *L = new realdpcomplex[p * p + n * p];
		realdpcomplex *r_T = L + p * p;
		for (integer i = 0; i < p; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				L[j + i * p].r = 0;
				L[j + i * p].i = 0;
			}
			sign.r = static_cast<realdp> ((((realdpcomplex *)ptrHHR)[i + n * i].r >= 0) ? 1 : -1);
			for (integer j = i; j < p; j++)
			{
				L[j + i * p].r = ((realdpcomplex *)ptrHHR)[i + n * j].r * sign.r;
				L[j + i * p].i = ((realdpcomplex *)ptrHHR)[i + n * j].i * (-sign.r);
			}
		}

		/*r_T <-  resultTV transpose conjugate*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				r_T[j + i * p].r = ((realdpcomplex *)resultTV)[i + j * n].r;
				r_T[j + i * p].i = -((realdpcomplex *)resultTV)[i + j * n].i;
			}
		}
		/*solve linear system L^H M = r_T, the solution M is stored in r_T*/
		trtrs_(GLOBAL::L, GLOBAL::C, GLOBAL::N, &P, &N, L, &P, r_T, &P, &info);
		/*resultTV <-  r_T transpose conjugate*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				((realdpcomplex *)resultTV)[i + j * n].r = r_T[j + i * p].r;
				((realdpcomplex *)resultTV)[i + j * n].i = -r_T[j + i * p].i;
			}
		}
		delete[] L;
	};

	void CpxNStQOrth::IntrProjection(Variable *x, Vector *v, Vector *result) const
	{
		v->CopyTo(result);
	};

	void CpxNStQOrth::ExtrProjection(Variable *x, Vector *v, Vector *result) const
	{
		const realdp *xM = x->ObtainReadData();
		const realdp *V = v->ObtainReadData();

		realdpcomplex *XHX = new realdpcomplex[2 * p * p];
		realdpcomplex *XHV = XHX + p * p;
		Matrix MxM(xM, n, p), MXHX((realdp*)XHX, p, p), MV(V, n, p), MXHV((realdp*)XHV, p, p);
		// XHX <- XM^H XM
		Matrix::CGEMM(GLOBAL::ZONE, MxM, true, MxM, false, GLOBAL::ZZERO, MXHX);
		// XHV <- XM^H V
		Matrix::CGEMM(GLOBAL::ZONE, MxM, true, MV, false, GLOBAL::ZZERO, MXHV);

		integer N = n, P = p, info;
		// solve for (XHX)^{-1} XHV
		potrf_(GLOBAL::L, &P, XHX, &P, &info);
		potrs_(GLOBAL::L, &P, &P, XHX, &P, XHV, &P, &info);

		if(info != 0)
		{
			printf("warning: potrs failed in CpxNStQOrth::ExtrProjection with info:%d!\n", info);
		}

		// XHV <- (XHV - XHV^H)/2
		for (integer i = 0; i < p; i++)
		{
			for (integer j = i; j < p; j++)
			{
				XHV[i + j * p].r -= XHV[j + i * p].r;
				XHV[i + j * p].i += XHV[j + i * p].i;
				XHV[j + i * p].r = -XHV[i + j * p].r;
				XHV[j + i * p].i = XHV[i + j * p].i;
			}
		}
		for (integer i = 0; i < p * p; i++)
		{
			XHV[i].r /= 2;
			XHV[i].i /= 2;
		}
		v->CopyTo(result);
		realdp *resultTV = result->ObtainWritePartialData();
		Matrix MresultTV(resultTV, n, p);
		Matrix::CGEMM(GLOBAL::ZNONE, MxM, false, MXHV, false, GLOBAL::ZONE, MresultTV);

		delete[] XHX;
	};
}; /*end of ROPTLIB namespace*/
