/*
This file defines the class for the manifold of multinomial manifold.
See details in: 
	[SGHMY16] Y. Sun, J. Gao, X. Hong, B. Mishra, and B. Yin. "Heterogeneous Tensor Decomposition
	for Clustering via Manifold Optimization".

Manifold --> SPD

---- WH
*/
#ifndef MULTINOMIAL_H
#define MULTINOMIAL_H

#include "Manifolds/Multinomial/MNVariable.h"
#include "Manifolds/Multinomial/MNVector.h"
#include "Manifolds/Manifold.h"
#include "Others/MyMatrix.h"
#include "Others/def.h"

/*Define the namespace*/
namespace ROPTLIB{

	class Multinomial : public Manifold{
	public:
		/*Construct the Multinomial manifold*/
		Multinomial(integer inn, integer inm);

		/*Delete EMPTYINTR and EMPTYEXTR*/
		virtual ~Multinomial(void);

		/*Check whether all the parameters are legal or not.*/
		virtual void CheckParams(void) const;

		/*The second order approximation of the exponential mapping, i.e., result = x + etax + 0.5 etax x^{-1} etax */
		virtual void Retraction(Variable *x, Vector *etax, Variable *result, realdp stepsize) const;

		/*The fisher metric*/
		virtual realdp Metric(Variable *x, Vector *etax, Vector *xix) const;

		/*computes beta = \|etax\| / \|\mathcal{T}_{R_etax} etax\|
		Default: beta <-- 1*/
		virtual realdp Beta(Variable *x, Vector *etax) const;

		/*The vector transport by differentiated the retraction, which is R_x(etax) = x + etax + 0.5 etax x^{-1} etax.
		Therefore, it is xix + 0.5 xix x^{-1} etax + 0.5 etax x^{-1} xix */
		virtual void DiffRetraction(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result, bool IsEtaXiSameDir = false) const;

		/*TODO*/
		virtual void coTangentVector(Variable *x, Vector *etax, Variable *y, Vector *xiy, Vector *result) const;

		/*The orthogonal projection onto the tangent space using the affine invariance metric is: result = (etax + etax^T) / 2*/
		virtual void ExtrProjection(Variable *x, Vector *etax, Vector *result) const;

		/*gf = x * egf * x */
		virtual void EucGradToGrad(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const;

		/*Not done yet. Temporarily use: xix <-- exix*/
		virtual void EucHvToHv(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const;

	protected:
		mutable integer n; /*The size of the space, i.e., the number of row */
		mutable integer m; /*The size of the space, i.e., the number of column */
	};
}; /*end of ROPTLIB namespace*/
#endif // end of SPDMANIFOLD_H
