
#include "Manifolds/NStQOrth/NStQOrth.h"

/*Define the namespace*/
namespace ROPTLIB{

	NStQOrth::NStQOrth(integer r, integer c)
	{
		n = r;
		p = c;
		IsIntrApproach = true;
		HasHHR = false;
        HasLockCon = false;
		UpdBetaAlone = false;
		name.assign("Noncompact Stiefel manifold quotient orthogonal group");
		IntrinsicDim = r * c - c * (c - 1) / 2;
		ExtrinsicDim = r * c;
		metric = NSOHGZ;
		VecTran = NSOPARALLELIZATION;
		EMPTYEXTR = new NSOVector(r, c);
		EMPTYINTR = new NSOVector(IntrinsicDim);
	};

	void NStQOrth::SetParams(PARAMSMAP params)
	{
		Manifold::SetParams(params);
		PARAMSMAP::iterator iter;
		for (iter = params.begin(); iter != params.end(); iter++)
		{
			if (iter->first == static_cast<std::string> ("ParamSet"))
			{
				switch (static_cast<integer> (iter->second))
				{
				case 1:
					ChooseNSOParamsSet1();
					break;
				case 2:
					ChooseNSOParamsSet2();
					break;
				case 3:
					ChooseNSOParamsSet3();
					break;
				case 4:
					ChooseNSOParamsSet4();
					break;
				case 5:
					ChooseNSOParamsSet5();
					break;
				case 6:
					ChooseNSOParamsSet6();
					break;
				default:
					printf("warning: invalid index for parameter set! use the default one.");
					ChooseNSOParamsSet1();
					break;
				}
			}
		}
	};
	// Choose the default parameters
	void NStQOrth::ChooseNSOParamsSet1(void)
	{
		metric = NSOHGZ;
		VecTran = NSOPARALLELIZATION;
	};

	void NStQOrth::ChooseNSOParamsSet2(void)
	{
		metric = NSOEUC;
		VecTran = NSOPARALLELIZATION;
	};

	void NStQOrth::ChooseNSOParamsSet3(void)
	{
		metric = NSOQEUC;
		VecTran = NSOPARALLELIZATION;
	};

	void NStQOrth::ChooseNSOParamsSet4(void)
	{
		metric = NSOHGZ;
		VecTran = NSOPROJECTION;
	};

	void NStQOrth::ChooseNSOParamsSet5(void)
	{
		metric = NSOEUC;
		VecTran = NSOPROJECTION;
	};

	void NStQOrth::ChooseNSOParamsSet6(void)
	{
		metric = NSOQEUC;
		VecTran = NSOPROJECTION;
	};

	NStQOrth::~NStQOrth(void)
	{
		delete EMPTYEXTR;
		delete EMPTYINTR;
	};

	realdp NStQOrth::Metric(Variable *x, Vector *etax, Vector *xix) const
	{
		//Vector *exetax = EMPTYEXTR->ConstructEmpty();
		//Vector *exxix = EMPTYEXTR->ConstructEmpty();
		//ObtainExtr(x, etax, exetax);
		//ObtainExtr(x, xix, exxix);
		//realdp result = Manifold::Metric(x, exetax, exxix);
		//delete exetax;
		//delete exxix;
		if (IsIntrApproach)
			return Manifold::Metric(x, etax, xix);
		printf("Warning: Metric for extrinsic representation has not been done!\n");
		return 0;
	};

	void NStQOrth::CheckParams(void) const
	{
		std::string NSOMetricnames[NSOMETRICLENGTH] = { "NSOEUC", "NSOQEUC", "NSOHGZ" };
		std::string NSOVTnames[NSOVECTORTRANSPORTLENGTH] = { "NSOPARALLELIZATION", "NSOPROJECTION" };
		Manifold::CheckParams();
		printf("%s PARAMETERS:\n", name.c_str());
		if (p == 1)
		{
			printf("n           :%15d,\t", n);
			printf("metric      :%15s\n", NSOMetricnames[metric].c_str());
			printf("VecTran     :%15s\n", NSOVTnames[VecTran].c_str());
		}
		else
		{
			printf("n           :%15d,\t", n);
			printf("p           :%15d\n", p);
			printf("metric      :%15s\t", NSOMetricnames[metric].c_str());
			printf("VecTran     :%15s\n", NSOVTnames[VecTran].c_str());
		}
	};

	void NStQOrth::EucGradToGrad(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		if (metric == NSOEUC)
			EucGradToGradEUC(x, egf, gf, prob);
		else
			if (metric == NSOHGZ)
				EucGradToGradHGZ(x, egf, gf, prob);
			else
			if (metric == NSOQEUC)
				EucGradToGradQEUC(x, egf, gf, prob);
			else
				printf("warning: NStQOrth::EucGradToGrad has not been done for this metric!\n");
	};

	void NStQOrth::EucHvToHv(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{
		if (metric == NSOEUC)
			EucHvToHvEUC(x, etax, exix, xix, prob);
		else
			if (metric == NSOHGZ)
				EucHvToHvHGZ(x, etax, exix, xix, prob);
			else
				if (metric == NSOQEUC)
					EucHvToHvQEUC(x, etax, exix, xix, prob);
			else
				printf("warning: NStQOrth::EucHvToHv has not been done for this metric!\n");
	};

	void NStQOrth::Projection(Variable *x, Vector *v, Vector *result) const
	{
		if (IsIntrApproach)
			IntrProjection(x, v, result);
		else
			ExtrProjection(x, v, result);
	};

	void NStQOrth::Retraction(Variable *x, Vector *etax, Variable *result, realdp stepsize) const
	{
		if (IsIntrApproach)
		{
			//Vector *exetax = EMPTYEXTR->ConstructEmpty();
			//ObtainExtr(x, etax, exetax);
			//Manifold::Retraction(x, exetax, result, 1);
			//delete exetax;
			//return;

//			x->Print("x:");
//			exetax->Print("exetax:");//---

            // The following retraction is the same as the retraction by projection
            // when NSOEUC metric is used.
			// Compute Yp and S such that etax = x * S + Yp, where S is symmetric and x' * Yp = 0
			Vector *Yp = EMPTYEXTR->ConstructEmpty();
            ObtainExtr(x, etax, Yp);//----
            x->Print("x:");//---
            Yp->Print("exetax:");//----
            
			if (!x->TempDataExist("HHR"))
			{
				ComputeHHR(x);
			}
			const realdp *xM = x->ObtainReadData();
			const SharedSpace *HHR = x->ObtainReadTempData("HHR");
			const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
			const realdp *ptrHHR = HHR->ObtainReadData();
			const realdp *ptrHHRTau = HHRTau->ObtainReadData();
			const realdp *intretaxTV = etax->ObtainReadData();
			realdp *Ypptr = Yp->ObtainWriteEntireData();

			integer N = x->Getsize()[0], P = x->Getsize()[1], inc = 1, Length = N * P;
			integer info;
			integer idx = 0;
			//		realdpcomplex *S = new realdpcomplex[p * p];
			realdp r2 = static_cast<realdp> (sqrt(2.0));
			realdp factor = 1;//-- sqrt(Manifold::Metric(x, x, x));
			realdp *S = new realdp[p * p];

			for (integer i = 0; i < p; i++)
			{
				Ypptr[i + i * n] = intretaxTV[idx] * factor / 2;
				idx++;
			}

			for (integer i = 0; i < p; i++)
			{
				for (integer j = i + 1; j < p; j++)
				{
					Ypptr[j + i * n] = intretaxTV[idx] / r2 / 2 * factor;
					Ypptr[i + j * n] = intretaxTV[idx] / r2 / 2 * factor;
					idx++;
				}
			}
			for (integer i = 0; i < p; i++)
			{
				for (integer j = p; j < n; j++)
				{
					Ypptr[j + i * n] = intretaxTV[idx] / r2;
					idx++;
				}
			}

			realdp sign = 0;
			//for (integer i = 0; i < p; i++)
			//{
			//	sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			//	// result(i, :) <- sign * result(i, :), details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
			//	scal_(&P, &sign, Ypptr + i, &N);
			//}

			realdp *L = new realdp[p * p + n * p];
			realdp *r_T = L + p * p;
			for (integer i = 0; i < p; i++)
			{
				for (integer j = 0; j < i; j++)
				{
					L[j + i * p] = 0;
				}
				sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
				for (integer j = i; j < p; j++)
				{
					L[j + i * p] = ptrHHR[i + n * j] * sign;
				}
			}
			/*r_T <-  resultTV transpose*/
			for (integer i = 0; i < n; i++)
			{
				for (integer j = 0; j < p; j++)
				{
					r_T[j + i * p] = Ypptr[i + j * n];
				}
			}
			/*solve linear system L^H M = r_T, the solution M is stored in r_T*/
			trtrs_(GLOBAL::L, GLOBAL::C, GLOBAL::N, &P, &N, L, &P, r_T, &P, &info);
			/*resultTV <-  r_T transpose*/
			for (integer i = 0; i < n; i++)
			{
				for (integer j = 0; j < p; j++)
				{
					Ypptr[i + j * n] = r_T[j + i * p];
				}
			}
			for (integer i = 0; i < p; i++)
			{
				for (integer j = 0; j < p; j++)
				{
					S[i + j * p] = Ypptr[i + j * n];
					Ypptr[i + j * n] = 0;
				}
			}
			trtrs_(GLOBAL::L, GLOBAL::C, GLOBAL::N, &P, &P, L, &P, S, &P, &info);
			delete[] L;

			integer lwork = -1;
			realdp lworkopt;
			// compute the size of space required in the ormqr
			ormqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
				(const_cast<realdp *> (ptrHHRTau)), Ypptr, &N, &lworkopt, &lwork, &info);
			lwork = static_cast<integer> (lworkopt);

			realdp *work = new realdp[lwork];
			// resultTV <- Q * resultTV, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
			// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
			ormqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
				(const_cast<realdp *> (ptrHHRTau)), Ypptr, &N, work, &lwork, &info);
			delete[] work;
			ForDebug::Print("Yp:", Ypptr, n, p);
			ForDebug::Print("S:", S, p, p);
			// Obtain S and Yp

			// Start computing retraction
			realdp *YYp = new realdp[n * p * 4 + 16 * p * p];
			realdp *YYp_Q = YYp + n * p * 2;
			realdp *YYp_R = YYp_Q + n * p * 2;
			realdp *Z = YYp_R + 4 * p * p;
			realdp *tmp1 = Z + 4 * p * p;
			realdp *tmp2 = tmp1 + 4 * p * p;
			copy_(&Length, const_cast<realdp *> (xM), &GLOBAL::IONE, YYp, &GLOBAL::IONE);
			copy_(&Length, Ypptr, &GLOBAL::IONE, YYp + N * P, &GLOBAL::IONE);
			Matrix M_YYp(YYp, n, 2 * p), M_YYp_Q(YYp_Q, n, 2 * p), M_YYp_R(YYp_R, 2 * p, 2 * p);
			Matrix::QRDecom(M_YYp, M_YYp_Q, M_YYp_R);

//			std::cout << "YYp:" << M_YYp << std::endl;//--
//			std::cout << "YYp_Q:" << M_YYp_Q << std::endl;//--
//			std::cout << "YYp_R:" << M_YYp_R << std::endl;//--

			//Z = [I + 2 * S, I; I, 0];
			for (integer i = 0; i < 4 * p * p; i++)
				Z[i] = 0;
			for (integer i = 0; i < p; i++)
			{
				Z[i + i * 2 * p] = 1;
				Z[i + (i + p) * 2 * p] = 1;
				Z[i + p + i * 2 * p] = 1;
				for (integer j = 0; j < p; j++)
				{
					Z[i + j * 2 * p] += S[i + j * p] + S[j + i * p];
				}
			}
//			ForDebug::Print("Z:", Z, 2 * p, 2 * p);
			// Eigendecomposition for R Z R';
			integer twop = 2 * p;
			gemm_(GLOBAL::N, GLOBAL::N, &twop, &twop, &twop, &GLOBAL::DONE, YYp_R, &twop, Z, &twop, &GLOBAL::DZERO, tmp1, &twop);
//			ForDebug::Print("R Z:", tmp1, 2 * p, 2 * p);

			gemm_(GLOBAL::N, GLOBAL::T, &twop, &twop, &twop, &GLOBAL::DONE, tmp1, &twop, YYp_R, &twop, &GLOBAL::DZERO, Z, &twop);
//			ForDebug::Print("R Z R':", Z, 2 * p, 2 * p);
			Matrix M_Z(Z, 2 * p, 2 * p), M_ZU(tmp1, 2 * p, 2 * p), M_ZD(tmp2, 2 * p, 1);
			Matrix::EigenSymmetricM(GLOBAL::L, M_Z, M_ZD, M_ZU);
//			std::cout << "ZU:" << M_ZU << std::endl;//---
//			std::cout << "ZD:" << M_ZD << std::endl;//---
			for (integer i = 0; i < p; i++)
			{
				realdp sigma = sqrt(tmp2[2 * p - 1 - i]); // singular values are in ascending order, so pick the large ones
				scal_(&twop, &sigma, tmp1 + (2 * p - 1 - i) * 2 * p, &GLOBAL::IONE);
			}
//			std::cout << "ZUD:" << M_ZU << std::endl;//---
			double *resultTV = result->ObtainWriteEntireData();
			Matrix M_UDp(tmp1 + 2 * p * p, 2 * p, p), M_result(resultTV, n, p);
//			std::cout << "Q:" << M_YYp_Q << std::endl;//---
			Matrix::DGEMM(1, M_YYp_Q, false, M_UDp, false, 0, M_result);
//			std::cout << "result:" << M_result << std::endl;//---

			delete[] S;
			delete Yp;
			delete [] YYp;

			realdp *O = new realdp[p * p * 4 + p + n * p];
			realdp *U = O + p * p;
			realdp *Vt = U + p * p;
			realdp *Sigma = Vt + p * p;
			realdp *XYt = Sigma + p;
			realdp *ResultO = XYt + p * p;
			XYt[0] = 1;
			Matrix M_O(O, p, p), M_U(U, p, p), M_Vt(Vt, p, p), M_S(Sigma, p, 1), M_XYt(XYt, p, p), M_X(xM, n, p), M_RO(ResultO, n, p);
			Matrix::DGEMM(1, M_X, true, M_result, false, 0, M_XYt);
			Matrix::SVDDecom(M_XYt, M_U, M_S, M_Vt);
			Matrix::DGEMM(1, M_Vt, true, M_U, true, 0, M_O);
			Matrix::DGEMM(1, M_result, false, M_O, false, 0, M_RO);
			copy_(&Length, M_RO.matrix, &GLOBAL::IONE, M_result.matrix, &GLOBAL::IONE);
			delete[] O;
		}
		else
		{
			Manifold::Retraction(x, etax, result, 1);
		}
	};

	void NStQOrth::coTangentVector(Variable *x, Vector *etax, Variable *y, Vector *xiy, Vector *result) const
	{
		printf("warning:NStQOrth::coTangentVector has not been done!\n");
		xiy->CopyTo(result);
	};

	void NStQOrth::DiffRetraction(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result, bool IsEtaXiSameDir) const
	{
		if (IsIntrApproach)
		{
			Vector *exxix = EMPTYEXTR->ConstructEmpty();
			Vector *exresult = EMPTYEXTR->ConstructEmpty();
			ObtainExtr(x, xix, exxix);
			ExtrProjection(y, exxix, exresult);
			ObtainIntr(y, exresult, result);
			delete exxix;
			delete exresult;
		}
		else
		{
			ExtrProjection(y, xix, result);
		}

		if (IsEtaXiSameDir && (HasHHR || UpdBetaAlone))
		{
			const realdp *etaxTV = etax->ObtainReadData();
			const realdp *xixTV = xix->ObtainReadData();
			realdp EtatoXi = sqrt(Metric(x, etax, etax) / Metric(x, xix, xix));
			SharedSpace *beta = new SharedSpace(1, 3);
			realdp *betav = beta->ObtainWriteEntireData();
			betav[0] = sqrt(Metric(x, etax, etax) / Metric(x, result, result)) / EtatoXi;
			betav[1] = Metric(x, etax, etax);
			betav[2] = Metric(x, result, result) * EtatoXi * EtatoXi;
			etax->AddToTempData("beta", beta);

			if (HasHHR)
			{
				Vector *TReta = result->ConstructEmpty();
				result->CopyTo(TReta);
				ScaleTimesVector(x, betav[0] * EtatoXi, TReta, TReta);
				SharedSpace *SharedTReta = new SharedSpace(TReta);
				etax->AddToTempData("betaTReta", SharedTReta);
			}
		}
	};

	realdp NStQOrth::Beta(Variable *x, Vector *etax) const
	{
		return 1;
	};


	void NStQOrth::ComputeHHR(Variable *x)
	{
		const realdp *xM = x->ObtainReadData();
		SharedSpace *HouseHolderResult = new SharedSpace(2, x->Getsize()[0], x->Getsize()[1]);
		realdp *ptrHHR = HouseHolderResult->ObtainWriteEntireData();
		SharedSpace *HHRTau = new SharedSpace(1, x->Getsize()[1]);
		realdp *tau = HHRTau->ObtainWriteEntireData();

		integer N = x->Getsize()[0], P = x->Getsize()[1], inc = 1;
		// ptrHHR <- xM, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		integer Length = N * P;
		copy_(&Length, const_cast<realdp *> (xM), &inc, ptrHHR, &inc);
		integer *jpvt = new integer[P];
		integer info;
		integer lwork = -1;
		realdp lworkopt;
		// compute the size of space required in the geqp3
		geqp3_(&N, &P, ptrHHR, &N, jpvt, tau, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt);
		realdp *work = new realdp[lwork];
		for (integer i = 0; i < P; i++)
			jpvt[i] = i + 1;
		// QR decomposition for ptrHHR using Householder reflections. Householder reflectors and R are stored in ptrHHR.
		// details: http://www.netlib.org/lapack/explore-html/db/de5/geqp3_8f.html
		geqp3_(&N, &P, ptrHHR, &N, jpvt, tau, work, &lwork, &info);

		x->AddToTempData("HHR", HouseHolderResult);
		x->AddToTempData("HHRTau", HHRTau);
		if (info < 0)
			printf("Error in qr decomposition in NStQOrth::ComputeHHR!\n");
		for (integer i = 0; i < P; i++)
		{
			if (jpvt[i] != (i + 1))
				printf("Error in qf retraction in NStQOrth::ComputeHHR!\n");
		}
		delete[] jpvt;
		delete[] work;
	};

	void NStQOrth::ObtainIntr(Variable *x, Vector *etax, Vector *result) const
	{
		if (metric == NSOEUC)
			ObtainIntrEUC(x, etax, result);
		else
			if (metric == NSOHGZ)
				ObtainIntrHGZ(x, etax, result);
			else
				if (metric == NSOQEUC)
					ObtainIntrQEUC(x, etax, result);
			else
				printf("warning: NStQOrth::ObtainIntr has not been done for this metric!\n");
	};

	void NStQOrth::ObtainExtr(Variable *x, Vector *intretax, Vector *result) const
	{
		if (metric == NSOEUC)
			ObtainExtrEUC(x, intretax, result);
		else
			if (metric == NSOHGZ)
				ObtainExtrHGZ(x, intretax, result);
			else
				if (metric == NSOQEUC)
					ObtainExtrQEUC(x, intretax, result);
			else
				printf("warning: NStQOrth::ObtainExtr has not been done for this metric!\n");
	};

	void NStQOrth::IntrProjection(Variable *x, Vector *v, Vector *result) const
	{
		v->CopyTo(result);
	};

	void NStQOrth::ExtrProjection(Variable *x, Vector *v, Vector *result) const
	{
		if (metric == NSOEUC || metric == NSOHGZ)
			ExtrProjectionEUCorHGZ(x, v, result);
		else
			if (metric == NSOQEUC)
				ExtrProjectionQEUC(x, v, result);
			else
				printf("Warning: NStQOrth::ExtrProjection has not been done for this Riemannian metric!\n");
	};

	void NStQOrth::ExtrProjectionQEUC(Variable *x, Vector *v, Vector *result) const
	{
		const realdp *xM = x->ObtainReadData();
		const realdp *vptr = v->ObtainReadData();

		realdp *XTX = new realdp[p * p * 2];
		realdp *XTV = XTX + p * p;

		// XTX <- XM^T XM
		Matrix MxM(xM, n, p), MXTX(XTX, p, p), MV(vptr, n, p), MXTV(XTV, p, p);
		Matrix::DGEMM(GLOBAL::DONE, MxM, true, MxM, false, GLOBAL::DZERO, MXTX);
		// XTV <- XM^T V
		Matrix::DGEMM(GLOBAL::DONE, MxM, true, MV, false, GLOBAL::DZERO, MXTV);

		// XTV <- XTV - XTV^T
		for (integer i = 0; i < p; i++)
		{
			for (integer j = i; j < p; j++)
			{
				XTV[i + j * p] -= XTV[j + i * p];
				XTV[j + i * p] = -XTV[i + j * p];
			}
		}
		v->CopyTo(result);
		realdp *resultTV = result->ObtainWritePartialData();
		Matrix Mresult(resultTV, n, p);
		Matrix::DSYL(MXTX, MXTX, MXTV);
		Matrix::DGEMM(GLOBAL::DNONE, MxM, false, MXTV, false,
			GLOBAL::DONE, Mresult);

		delete[] XTX;
	};

	void NStQOrth::ExtrProjectionEUCorHGZ(Variable *x, Vector *v, Vector *result) const
	{
		const realdp *xM = x->ObtainReadData();
		const realdp *V = v->ObtainReadData();

		realdp *XTX = new realdp[2 * p * p];
		realdp *XTV = XTX + p * p;
		Matrix MxM(xM, n, p), MXTX(XTX, p, p), MV(V, n, p), MXTV(XTV, p, p);
		// XHX <- XM^T XM
		Matrix::DGEMM(GLOBAL::DONE, MxM, true, MxM, false, GLOBAL::DZERO, MXTX);
		// XHV <- XM^H V
		Matrix::DGEMM(GLOBAL::DONE, MxM, true, MV, false, GLOBAL::DZERO, MXTV);

		integer N = n, P = p, info;
		// solve for (XTX)^{-1} XTV
		potrf_(GLOBAL::L, &P, XTX, &P, &info);
		potrs_(GLOBAL::L, &P, &P, XTX, &P, XTV, &P, &info);

		if (info != 0)
		{
			printf("warning: potrs failed in NStQOrth::ExtrProjection with info:%d!\n", info);
	}

		// XTV <- (XTV - XTV^T)/2
		for (integer i = 0; i < p; i++)
		{
			for (integer j = i; j < p; j++)
			{
				XTV[i + j * p] -= XTV[j + i * p];
				XTV[j + i * p] = -XTV[i + j * p];
			}
		}
		for (integer i = 0; i < p * p; i++)
		{
			XTV[i] /= 2;
		}
		v->CopyTo(result);
		realdp *resultTV = result->ObtainWritePartialData();
		Matrix MresultTV(resultTV, n, p);
		Matrix::DGEMM(GLOBAL::DNONE, MxM, false, MXTV, false, GLOBAL::DONE, MresultTV);

		delete[] XTX;
	};

	void NStQOrth::EucGradToGradEUC(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		if (prob->GetUseHess())
		{
			Vector *segf = egf->ConstructEmpty();
			segf->NewMemoryOnWrite(); // I don't remember the reason. It seems to be required.
			egf->CopyTo(segf);
			SharedSpace *Sharedegf = new SharedSpace(segf);
			x->AddToTempData("EGrad", Sharedegf);
		}

		egf->CopyTo(gf);
		realdp *gfv = gf->ObtainWriteEntireData();

		//Y / (x'*x) use compute HHR in NstQorth
		if (!x->TempDataExist("HHR"))
			ComputeHHR(x);
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const realdp *ptrHHR = HHR->ObtainReadData();

		realdp *YT = new realdp[n * p];

		/*YT = Y^T*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				YT[j + i * p] = gfv[i + j * n];
			}
		}
		integer info, N = n, P = p;
		trtrs_(GLOBAL::U, GLOBAL::T, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, YT, &P, &info);
		trtrs_(GLOBAL::U, GLOBAL::N, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, YT, &P, &info);

		/*YT = Y^T*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				gfv[i + j * n] = YT[j + i * p];
			}
		}
		delete[] YT;
		realdp half = 0.5;
		integer length = n * p;
		scal_(&length, &half, gfv, &GLOBAL::IONE);

		realdp *tempspace = new realdp[p * p];

		const realdp *Y = x->ObtainReadData();

		/*tempspace2 = Y^T * gfv */
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (Y), &N, gfv, &N, &GLOBAL::DZERO, tempspace, &P);

		/*compute (Y^T Y)^{-1} tempspace2. It is also R^{-1} R^{-T} tempspace2 */
		trtrs_(GLOBAL::U, GLOBAL::T, GLOBAL::N, &P, &P, const_cast<realdp *> (ptrHHR), &N, tempspace, &P, &info);
		trtrs_(GLOBAL::U, GLOBAL::N, GLOBAL::N, &P, &P, const_cast<realdp *> (ptrHHR), &N, tempspace, &P, &info);

		realdp nhalf = -0.5;
		/*final result: gfv = (I - 1/2 * Y * (Y^T Y)^{-1}Y^T) gfv */
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &nhalf, const_cast<realdp *> (Y), &N, tempspace, &P, &GLOBAL::DONE, gfv, &N);

		delete[] tempspace;
		//ExtrProjection(x, gf, gf); Not necessary since gf is already in the horizontal space.
	};

	void NStQOrth::EucHvToHvEUC(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{
		const realdp *Y = x->ObtainReadData();
		const realdp *etaxTV = etax->ObtainReadData();

		const SharedSpace *Sharedegf = nullptr;
		Vector *egfx = nullptr;
		realdp *egfTV = nullptr;
		if (x->TempDataExist("EGrad"))
		{
			Sharedegf = x->ObtainReadTempData("EGrad");
			egfx = Sharedegf->GetSharedElement();
			egfTV = const_cast<realdp *> (egfx->ObtainReadData());
		}
		else
		{
			egfTV = new realdp[x->Getlength()];
			for (integer i = 0; i < x->Getlength(); i++)
				egfTV[i] = 0;
		}
		//realdp *egfTV = new realdp[n * p];//----
		//for (integer i = 0; i < n * p; i++)
		//	egfTV[i] = 0;//----

		if (!x->TempDataExist("HHR"))
			ComputeHHR(x);
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const realdp *ptrHHR = HHR->ObtainReadData();

		realdp *tmp = new realdp[n * p * 2 + p * p * 4];
		realdp *YYinvYTEH = tmp + n * p;
		realdp *EGYYinv = YYinvYTEH + p * p;
		realdp *YTeta = EGYYinv + n * p;
		realdp *EtaTEGYYinv = YTeta + p * p;
		realdp *YYinvYTEGYYinv = EtaTEGYYinv + p * p;
		/*tmp = Egrad^T*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				tmp[j + i * p] = egfTV[i + j * n];
			}
		}
		integer info, N = n, P = p;
		trtrs_(GLOBAL::U, GLOBAL::T, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, tmp, &P, &info);
		trtrs_(GLOBAL::U, GLOBAL::N, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, tmp, &P, &info);
		/*EGYYinv = Egrad * (Y^T * Y)^{-1}*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				EGYYinv[i + j * n] = tmp[j + i * p];
			}
		}

		/*YYinvYTEH = (Y^T * Y)^{-1} * Y^T * EHess */
		const realdp *exixTV = exix->ObtainReadData();
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (Y), &N, const_cast<realdp *> (exixTV), &N, &GLOBAL::DZERO, YYinvYTEH, &P);
		trtrs_(GLOBAL::U, GLOBAL::T, GLOBAL::N, &P, &P, const_cast<realdp *> (ptrHHR), &N, YYinvYTEH, &P, &info);
		trtrs_(GLOBAL::U, GLOBAL::N, GLOBAL::N, &P, &P, const_cast<realdp *> (ptrHHR), &N, YYinvYTEH, &P, &info);

		/*YYinvYTEGYYinv = (Y^T * Y)^{-1} * Y^T * Egrad * (Y^T * Y)^{-1}*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (Y), &N, EGYYinv, &N, &GLOBAL::DZERO, YYinvYTEGYYinv, &P);
		trtrs_(GLOBAL::U, GLOBAL::T, GLOBAL::N, &P, &P, const_cast<realdp *> (ptrHHR), &N, YYinvYTEGYYinv, &P, &info);
		trtrs_(GLOBAL::U, GLOBAL::N, GLOBAL::N, &P, &P, const_cast<realdp *> (ptrHHR), &N, YYinvYTEGYYinv, &P, &info);

		/*EtaTEGYYinv = eta^T * Egrad * (Y^T * Y)^{-1}*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (etaxTV), &N, EGYYinv, &N, &GLOBAL::DZERO, EtaTEGYYinv, &P);

		/*YTeta = Y^T * eta */
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (Y), &N, const_cast<realdp *> (etaxTV), &N, &GLOBAL::DZERO, YTeta, &P);

		exix->CopyTo(xix);
		realdp *xixTV = xix->ObtainWritePartialData();
		realdp half = 0.5, nhalf = -0.5;

		/*xix = EucHess - 0.5 * Y * (Y^T * Y)^{-1} * Y^T * EucHess */
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &nhalf, const_cast<realdp *> (Y), &N, YYinvYTEH, &P, &GLOBAL::DONE, xixTV, &N);

		/*xix =  EucHess - 0.5 * Y * (Y^T * Y)^{-1} * Y^T * EucHess - 0.5 * Y * (Y^T * Y)^{-1} * Egrad^T * eta */
		gemm_(GLOBAL::N, GLOBAL::T, &N, &P, &P, &nhalf, const_cast<realdp *>(Y), &N, EtaTEGYYinv, &P, &GLOBAL::DONE, xixTV, &N);

		/*xix =  EucHess - 0.5 * Y * (Y^T * Y)^{-1} * Y^T * EucHess - 0.5 * Y * (Y^T * Y)^{-1} * Egrad^T * eta - Egrad * (Y^T * Y)^{-1} * Y * eta */
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DNONE, const_cast<realdp *>(EGYYinv), &N, YTeta, &P, &GLOBAL::DONE, xixTV, &N);

		/*xix =  EucHess - 0.5 * Y * (Y^T * Y)^{-1} * Y^T * EucHess - 0.5 * Y * (Y^T * Y)^{-1} * Egrad^T * eta - Egrad * (Y^T * Y)^{-1} * Y * eta 
		         + Y * (Y^T * Y)^{-1} * Y^T * Egrad * (Y^T * Y)^{-1} * Y^T * eta */
		gemm_(GLOBAL::N, GLOBAL::N, &P, &P, &P, &GLOBAL::DONE, YYinvYTEGYYinv, &P, YTeta, &P, &GLOBAL::DZERO, tmp, &P);
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DONE, const_cast<realdp *>(Y), &N, tmp, &P, &GLOBAL::DONE, xixTV, &N);

		integer length = n * p;
		scal_(&length, &half, xixTV, &GLOBAL::IONE);

		/*tmp = xix^T*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				tmp[j + i * p] = xixTV[i + j * n];
			}
		}
		trtrs_(GLOBAL::U, GLOBAL::T, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, tmp, &P, &info);
		trtrs_(GLOBAL::U, GLOBAL::N, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, tmp, &P, &info);

		/*xix = xix * (Y^T * Y)^{-1}*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				xixTV[i + j * n] = tmp[j + i * p];
			}
		}

		delete[] tmp;

		ExtrProjection(x, xix, xix);
	};

	void NStQOrth::ObtainIntrEUC(Variable *x, Vector *etax, Vector *result) const
	{
		if (!x->TempDataExist("HHR"))
		{
			ComputeHHR(x);
		}

		const realdp *xM = x->ObtainReadData();
		const realdp *etaxTV = etax->ObtainReadData();
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
		realdp *resultTV = result->ObtainWriteEntireData();
		const realdp *ptrHHR = HHR->ObtainReadData();
		const realdp *ptrHHRTau = HHRTau->ObtainReadData();

		integer N = x->Getsize()[0], P = x->Getsize()[1], inc = 1, Length = N * P;
		integer info;
		integer lwork = -1;
		realdp lworkopt;
		realdp *tempspace = new realdp[n * p];
		// compute the size of space required in the ormqr
		ormqr_(GLOBAL::L, GLOBAL::T, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), tempspace, &N, &lworkopt, &lwork, &info);

		lwork = static_cast<integer> (lworkopt);
		realdp *work = new realdp[lwork];
		// tempspace <- etaxTV, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		copy_(&Length, const_cast<realdp *> (etaxTV), &inc, tempspace, &inc);
		// tempspace <- Q^T * tempspace, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		ormqr_(GLOBAL::L, GLOBAL::T, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), tempspace, &N, work, &lwork, &info);
		realdp sign = 0;
		for (integer i = 0; i < p; i++)
		{
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			scal_(&P, &sign, tempspace + i, &N);
		}

		realdp *L = new realdp[p * p];
		for (integer i = 0; i < p; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				L[j + i * p] = 0;
			}
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			for (integer j = i; j < p; j++)
			{
				L[j + i * p] = ptrHHR[i + n * j] * sign;
			}
		}

		realdp *tempspaceL = new realdp[n * p];

		Matrix MtL(tempspaceL, n, p), ML(L, p, p), Mtempspace(tempspace, n, p);
		Matrix::DGEMM(GLOBAL::DONE, Mtempspace, false, ML, false, GLOBAL::DZERO, MtL);

		delete[] L;
		delete[] tempspace;

		/*Matrix MxM(xM, n, p), MetaxTV(etaxTV, n, p), Mtempspace((realdp *)tempspace, p, p, n);
		Matrix::CGEMM(GLOBAL::ZONE, MxM, true, MetaxTV, false, GLOBAL::ZZERO, Mtempspace);*/
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		realdp factor = 1;//-- sqrt(Manifold::Metric(x, x, x));
		integer idx = 0;
		for (integer i = 0; i < p; i++)
		{
			resultTV[idx] = 2 * tempspaceL[i + i * n] / factor;
			idx++;
		}
		for (integer i = 0; i < p; i++)
		{
			for (integer j = i + 1; j < p; j++)
			{
				resultTV[idx] = 2 * r2 * tempspaceL[j + i * n] / factor;
				idx++;
			}
		}

		for (integer i = 0; i < p; i++)
		{
			for (integer j = p; j < n; j++)
			{
				resultTV[idx] = r2 * tempspaceL[j + i * n];
				idx++;
			}
		}
		delete[] work;
		delete[] tempspaceL;
	};

	void NStQOrth::ObtainExtrEUC(Variable *x, Vector *intretax, Vector *result) const
	{
		if (!x->TempDataExist("HHR"))
		{
			ComputeHHR(x);
		}

		const realdp *xM = x->ObtainReadData();
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
		const realdp *ptrHHR = HHR->ObtainReadData();
		const realdp *ptrHHRTau = HHRTau->ObtainReadData();
		const realdp *intretaxTV = intretax->ObtainReadData();
		realdp *resultTV = result->ObtainWriteEntireData();

		integer N = x->Getsize()[0], P = x->Getsize()[1], inc = 1, Length = N * P;
		integer info;
		integer idx = 0;
		//		realdpcomplex *S = new realdpcomplex[p * p];
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		realdp factor = 1;//-- sqrt(Manifold::Metric(x, x, x));
		for (integer i = 0; i < p; i++)
		{
			resultTV[i + i * n] = intretaxTV[idx] * factor / 2;
			idx++;
		}

		for (integer i = 0; i < p; i++)
		{
			for (integer j = i + 1; j < p; j++)
			{
				resultTV[j + i * n] = intretaxTV[idx] / r2 / 2 * factor;
				resultTV[i + j * n] = intretaxTV[idx] / r2 / 2 * factor;
				idx++;
			}
		}
		for (integer i = 0; i < p; i++)
		{
			for (integer j = p; j < n; j++)
			{
				resultTV[j + i * n] = intretaxTV[idx] / r2;
				idx++;
			}
		}

		realdp sign = 0;
		for (integer i = 0; i < p; i++)
		{
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			// result(i, :) <- sign * result(i, :), details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
			scal_(&P, &sign, resultTV + i, &N);
		}

		realdp *L = new realdp[p * p + n * p];
		realdp *r_T = L + p * p;
		for (integer i = 0; i < p; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				L[j + i * p] = 0;
			}
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			for (integer j = i; j < p; j++)
			{
				L[j + i * p] = ptrHHR[i + n * j] * sign;
			}
		}

		/*r_T <-  resultTV transpose*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				r_T[j + i * p] = resultTV[i + j * n];
			}
		}
		/*solve linear system L^H M = r_T, the solution M is stored in r_T*/
		trtrs_(GLOBAL::L, GLOBAL::C, GLOBAL::N, &P, &N, L, &P, r_T, &P, &info);
		/*resultTV <-  r_T transpose*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				resultTV[i + j * n] = r_T[j + i * p];
			}
		}
		delete[] L;

		integer lwork = -1;
		realdp lworkopt;
		// compute the size of space required in the ormqr
		ormqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), resultTV, &N, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt);

		realdp *work = new realdp[lwork];
		// resultTV <- Q * resultTV, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		ormqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), resultTV, &N, work, &lwork, &info);
		delete[] work;
	};

	void NStQOrth::EucGradToGradQEUC(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		if (prob->GetUseHess())
		{
			Vector *segf = egf->ConstructEmpty();
			segf->NewMemoryOnWrite(); // I don't remember the reason. It seems to be required.
			egf->CopyTo(segf);
			SharedSpace *Sharedegf = new SharedSpace(segf);
			x->AddToTempData("EGrad", Sharedegf);
		}
		ExtrProjection(x, egf, gf);
	};

	void NStQOrth::EucHvToHvQEUC(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{
		ExtrProjection(x, exix, xix);
	};

	void NStQOrth::ObtainIntrQEUC(Variable *x, Vector *etax, Vector *result) const
	{
		if (!x->TempDataExist("HHR"))
		{
			ComputeHHR(x);
		}

		const realdp *xM = x->ObtainReadData();
		const realdp *etaxTV = etax->ObtainReadData();
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
		realdp *resultTV = result->ObtainWriteEntireData();
		const realdp *ptrHHR = HHR->ObtainReadData();
		const realdp *ptrHHRTau = HHRTau->ObtainReadData();

		integer N = x->Getsize()[0], P = x->Getsize()[1], inc = 1, Length = N * P;
		integer info;
		integer lwork = -1;
		realdp lworkopt;
		realdp *tempspace = new realdp[n * p];
		// compute the size of space required in the ormqr
		ormqr_(GLOBAL::L, GLOBAL::T, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), tempspace, &N, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt);
		realdp *work = new realdp[lwork];
		// tempspace <- etaxTV, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		copy_(&Length, const_cast<realdp *> (etaxTV), &inc, tempspace, &inc);
		// tempspace <- Q^T * tempspace, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		ormqr_(GLOBAL::L, GLOBAL::T, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), tempspace, &N, work, &lwork, &info);
		delete[] work;

		realdp sign = 0;
		for (integer i = 0; i < p; i++)
		{
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			scal_(&P, &sign, tempspace + i, &N);
		}

		realdp *L = new realdp[p * p];
		for (integer i = 0; i < p; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				L[j + i * p] = 0;
			}
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			for (integer j = i; j < p; j++)
			{
				L[j + i * p] = ptrHHR[i + n * j] * sign;
			}
		}

		/* Ltempspace = L* tempspace(1:p, 1:p) */

		realdp *Ltempspace = new realdp[p * p];
		gemm_(GLOBAL::N, GLOBAL::N, &P, &P, &P, &GLOBAL::DONE, L, &P, tempspace, &N, &GLOBAL::DZERO, Ltempspace, &P);
		for (integer i = 0; i < P; i++)
		{
			for (integer j = 0; j < P; j++)
				tempspace[j + i * n] = Ltempspace[j + i * p];
		}
		delete[] Ltempspace;

		realdp *resultptr = result->ObtainWriteEntireData();
		integer idx = 0;
		for (integer i = 0; i < p; i++)
		{
			for (integer j = i; j < p; j++)
			{
				resultptr[idx] = tempspace[j + i * n];
				idx++;
			}
		}
		for (integer i = 0; i < p; i++)
		{
			for (integer j = p; j < n; j++)
			{
				resultptr[idx] = tempspace[j + i * n];
				idx++;
			}
		}
		delete[] tempspace;
		/* Above: result is then R^{-T} S + K. Next, we first find the coefficients of the etax under a non-orthonormal basis.
		Then orthonormalize the basis and update the coefficient. To the end, we find the non-orthogonal basis and store it in B.
		Then use qr decomposition to find the matrix R. Then use the matrix R to find the coefficient vector under the orthonormalized
		basis. total complexity is O(p^6) + O(n p)*/


		/* cholesky decomposition */
		realdp *B = new realdp[p * p * p * (p + 1) / 2];
		idx = 0;
		for (integer i = 0; i < p * p * p * (p + 1) / 2; i++)
			B[i] = 0;

		for (integer i = 0; i < p; i++)
		{
			for (integer j = i; j < p; j++)
			{
				B[j + i * p + idx * p * p] = 1;
				B[i + j * p + idx * p * p] = 1;
				trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &P, &P, L, &P, B + idx * p * p, &P, &info);
				idx++;
			}
		}
		delete[] L;

		integer *jpvt = new integer[p * (p + 1) / 2];
		realdp *tau = new realdp[p * (p + 1) / 2];
		lwork = -1;
		integer dim = p * (p + 1) / 2;
		integer Psquare = p * p;
		// compute the size of space required in the geqp3
		geqp3_(&Psquare, &dim, B, &Psquare, jpvt, tau, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt);
		work = new realdp[lwork];
		for (integer i = 0; i < p * (p + 1) / 2; i++)
			jpvt[i] = i + 1;
		// QR decomposition for ptrHHR using Householder reflections. Householder reflectors and R are stored in ptrHHR.
		// details: http://www.netlib.org/lapack/explore-html/db/de5/geqp3_8f.html
		geqp3_(&Psquare, &dim, B, &Psquare, jpvt, tau, work, &lwork, &info);
		delete[] tau;
		delete[] work;
		delete[]jpvt;


		SharedSpace *SharedBL = new SharedSpace(2, dim, dim);
		realdp *BL = SharedBL->ObtainWriteEntireData();
		for (integer i = 0; i < dim; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				BL[j + i * dim] = 0;
			}
			sign = static_cast<realdp> ((B[i + p * p * i] >= 0) ? 1 : -1);
			for (integer j = i; j < dim; j++)
			{
				BL[j + i * dim] = B[i + p * p * j] * sign;
			}
		}

		delete[] B;

		tempspace = new realdp[dim];
		gemm_(GLOBAL::T, GLOBAL::N, &dim, &GLOBAL::IONE, &dim, &GLOBAL::DONE, BL, &dim, resultptr, &dim, &GLOBAL::DZERO, tempspace, &dim);
		copy_(&dim, tempspace, &GLOBAL::IONE, resultptr, &GLOBAL::IONE);
		delete[] tempspace;
		x->AddToTempData("BL", SharedBL);
	};

	void NStQOrth::ObtainExtrQEUC(Variable *x, Vector *intretax, Vector *result) const
	{
		if (!x->TempDataExist("HHR"))
		{
			ComputeHHR(x);
		}

		const realdp *xM = x->ObtainReadData();
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
		const SharedSpace *BL = x->ObtainReadTempData("BL");
		const realdp *ptrHHR = HHR->ObtainReadData();
		const realdp *ptrHHRTau = HHRTau->ObtainReadData();
		const realdp *ptrBL = BL->ObtainReadData();
		const realdp *intretaxTV = intretax->ObtainReadData();
		realdp *resultTV = result->ObtainWriteEntireData();

		integer dim = p * (p + 1) / 2, info;
		realdp *tempspace = new realdp[dim];
		copy_(&dim, const_cast<realdp *> (intretaxTV), &GLOBAL::IONE, tempspace, &GLOBAL::IONE);
		trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &dim, &GLOBAL::IONE, const_cast<realdp *>(ptrBL), &dim, tempspace, &dim, &info);

		integer N = n, P = p, Length = N * P;
		integer idx = 0;

		for (integer i = 0; i < p; i++)
		{
			for (integer j = i; j < p; j++)
			{
				resultTV[j + i * n] = tempspace[idx];
				resultTV[i + j * n] = resultTV[j + i * n];
				idx++;
			}
		}
		for (integer i = 0; i < p; i++)
		{
			for (integer j = p; j < n; j++)
			{
				resultTV[j + i * n] = intretaxTV[idx];
				idx++;
			}
		}
		delete[] tempspace;
		realdp *L = new realdp[p * p];
		realdp sign = 0;
		for (integer i = 0; i < p; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				L[j + i * p] = 0;
			}
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			for (integer j = i; j < p; j++)
			{
				L[j + i * p] = ptrHHR[i + n * j] * sign;
			}
		}

		trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &P, &P, L, &P, resultTV, &N, &info);

		delete[] L;

		/*=========*/
		sign = 0;
		for (integer i = 0; i < p; i++)
		{
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			// result(i, :) <- sign * result(i, :), details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
			scal_(&P, &sign, resultTV + i, &N);
		}
		integer lwork = -1;
		realdp lworkopt;
		// compute the size of space required in the ormqr
		ormqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), resultTV, &N, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt);

		realdp *work = new realdp[lwork];
		// resultTV <- Q * resultTV, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		ormqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), resultTV, &N, work, &lwork, &info);
		delete[] work;
	};

	void NStQOrth::EucGradToGradHGZ(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		if (prob->GetUseHess())
		{
			Vector *segf = egf->ConstructEmpty();
			segf->NewMemoryOnWrite(); // I don't remember the reason. It seems to be required.
			egf->CopyTo(segf);
			SharedSpace *Sharedegf = new SharedSpace(segf);
			x->AddToTempData("EGrad", Sharedegf);
		}
		egf->CopyTo(gf);
		realdp *gfv = gf->ObtainWriteEntireData();

		//Y / (x'*x) use compute HHR in NstQorth
		if (!x->TempDataExist("HHR"))
			ComputeHHR(x);
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const realdp *ptrHHR = HHR->ObtainReadData();

		realdp *YT = new realdp[n * p];

		/*YT = Y^T*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				YT[j + i * p] = gfv[i + j * n];
			}
		}
		integer info, N = n, P = p;
		trtrs_(GLOBAL::U, GLOBAL::T, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, YT, &P, &info);
		trtrs_(GLOBAL::U, GLOBAL::N, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, YT, &P, &info);
		/*YT = Y^T*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				gfv[i + j * n] = YT[j + i * p];
			}
		}
		delete[] YT;

		//ExtrProjection(x, gf, gf); Not necessary since gf is already in the horizontal space.
	};

	void NStQOrth::EucHvToHvHGZ(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{
		const realdp *Y = x->ObtainReadData();
		const realdp *etaxTV = etax->ObtainReadData();

		const SharedSpace *Sharedegf = x->ObtainReadTempData("EGrad");
		Vector *egfx = Sharedegf->GetSharedElement();
		const realdp *egfTV = egfx->ObtainReadData();

		if (!x->TempDataExist("HHR"))
			ComputeHHR(x);
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const realdp *ptrHHR = HHR->ObtainReadData();

		realdp *tmp = new realdp[n * p * 2 + p * p * 3];
		realdp *EGYYinv = tmp + n * p;
		realdp *YTEGYYinv = EGYYinv + n * p;
		realdp *EtaTEGYYinv = YTEGYYinv + p * p;
		realdp *EtaTY = EtaTEGYYinv + p * p;
		/*tmp = Egrad^T*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				tmp[j + i * p] = egfTV[i + j * n];
			}
		}
		integer info, N = n, P = p;
		trtrs_(GLOBAL::U, GLOBAL::T, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, tmp, &P, &info);
		trtrs_(GLOBAL::U, GLOBAL::N, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, tmp, &P, &info);

		/*EGYYinv = Egrad * (Y^T * Y)^{-1}*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				EGYYinv[i + j * n] = tmp[j + i * p];
			}
		}

		/*YTEGYYinv = Y^T * Egrad * (Y^T * Y)^{-1}*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (Y), &N, EGYYinv, &N, &GLOBAL::DZERO, YTEGYYinv, &P);

		/*EtaTEGYYinv = eta^T * Egrad * (Y^T * Y)^{-1}*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (etaxTV), &N, EGYYinv, &N, &GLOBAL::DZERO, EtaTEGYYinv, &P);

		/*EtaTY = eta^T * Y */
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (etaxTV), &N, const_cast<realdp *> (Y), &N, &GLOBAL::DZERO, EtaTY, &P);

		exix->CopyTo(xix);
		realdp *xixTV = xix->ObtainWritePartialData();
		realdp half = 0.5, nhalf = -0.5;

		/*xix = EucHess - 0.5 * Egrad * (Y^T * Y)^{-1} * Y^T * eta */
		gemm_(GLOBAL::N, GLOBAL::T, &N, &P, &P, &nhalf, EGYYinv, &N, EtaTY, &P, &GLOBAL::DONE, xixTV, &N);

		/*xix =  EucHess - 0.5 * Egrad * (Y^T * Y)^{-1} * Y^T * eta - 0.5 * Y * (Y^T * Y)^{-1} Egrad^T * eta */
		gemm_(GLOBAL::N, GLOBAL::T, &N, &P, &P, &nhalf, const_cast<realdp *>(Y), &N, EtaTEGYYinv, &P, &GLOBAL::DONE, xixTV, &N);

		/*xix =  EucHess - 0.5 * Egrad * (Y^T * Y)^{-1} * Y^T * eta - 0.5 * Y * (Y^T * Y)^{-1} Egrad^T * eta + 0.5 * eta * Y^T * Egrad * (Y^T * Y)^{-1} */
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &half, const_cast<realdp *>(etaxTV), &N, YTEGYYinv, &P, &GLOBAL::DONE, xixTV, &N);

		/*xix =  EucHess - 0.5 * Egrad * (Y^T * Y)^{-1} * Y^T * eta - 0.5 * Y * (Y^T * Y)^{-1} Egrad^T * eta + 0.5 * eta * Y^T * Egrad * (Y^T * Y)^{-1} 
		         - 0.5 * Y * eta^T * Egrad * (Y^T * Y)^{-1} */
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &nhalf, const_cast<realdp *>(Y), &N, EtaTEGYYinv, &P, &GLOBAL::DONE, xixTV, &N);

		/*xix =  EucHess - 0.5 * Egrad * (Y^T * Y)^{-1} * Y^T * eta - 0.5 * Y * (Y^T * Y)^{-1} Egrad^T * eta + 0.5 * eta * Y^T * Egrad * (Y^T * Y)^{-1}
				 - 0.5 * Y * eta^T * Egrad * (Y^T * Y)^{-1} + 0.5 * eta *(Y^T * Y)^{-1} Egrad^T Y */
		gemm_(GLOBAL::N, GLOBAL::T, &N, &P, &P, &half, const_cast<realdp *>(etaxTV), &N, YTEGYYinv, &P, &GLOBAL::DONE, xixTV, &N);

		/*xix =  EucHess - 0.5 * Egrad * (Y^T * Y)^{-1} * Y^T * eta - 0.5 * Y * (Y^T * Y)^{-1} Egrad^T * eta + 0.5 * eta * Y^T * Egrad * (Y^T * Y)^{-1}
				 - 0.5 * Y * eta^T * Egrad * (Y^T * Y)^{-1} + 0.5 * eta *(Y^T * Y)^{-1} Egrad^T Y - 0.5 * Egrad * (Y^T * Y)^{-1} * eta^T * Y */
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &nhalf, EGYYinv, &N, EtaTY, &P, &GLOBAL::DONE, xixTV, &N);
		
		/*tmp = xix^T*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				tmp[j + i * p] = xixTV[i + j * n];
			}
		}
		trtrs_(GLOBAL::U, GLOBAL::T, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, tmp, &P, &info);
		trtrs_(GLOBAL::U, GLOBAL::N, GLOBAL::N, &P, &N, const_cast<realdp *> (ptrHHR), &N, tmp, &P, &info);

		/*xix = xix * (Y^T * Y)^{-1}*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				xixTV[i + j * n] = tmp[j + i * p];
			}
		}

		delete[] tmp;
		
		ExtrProjection(x, xix, xix);
	};

	void NStQOrth::ObtainIntrHGZ(Variable *x, Vector *etax, Vector *result) const
	{
		if (!x->TempDataExist("HHR"))
		{
			ComputeHHR(x);
		}

		const realdp *xM = x->ObtainReadData();
		const realdp *etaxTV = etax->ObtainReadData();
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
		realdp *resultTV = result->ObtainWriteEntireData();
		const realdp *ptrHHR = HHR->ObtainReadData();
		const realdp *ptrHHRTau = HHRTau->ObtainReadData();

		integer N = x->Getsize()[0], P = x->Getsize()[1], inc = 1, Length = N * P;
		integer info;
		integer lwork = -1;
		realdp lworkopt;
		realdp *tempspace = new realdp[n * p];
		// compute the size of space required in the ormqr
		ormqr_(GLOBAL::L, GLOBAL::T, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), tempspace, &N, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt);
		realdp *work = new realdp[lwork];
		// tempspace <- etaxTV, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		copy_(&Length, const_cast<realdp *> (etaxTV), &inc, tempspace, &inc);
		// tempspace <- Q^T * tempspace, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		ormqr_(GLOBAL::L, GLOBAL::T, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), tempspace, &N, work, &lwork, &info);
		realdp sign = 0;
		for (integer i = 0; i < p; i++)
		{
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			scal_(&P, &sign, tempspace + i, &N);
		}

		realdp *L = new realdp[p * p];
		for (integer i = 0; i < p; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				L[j + i * p] = 0;
			}
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			for (integer j = i; j < p; j++)
			{
				L[j + i * p] = ptrHHR[i + n * j] * sign;
			}
		}

		realdp *tempspaceL = new realdp[n * p];

		Matrix MtL(tempspaceL, n, p), ML(L, p, p), Mtempspace(tempspace, n, p);
		Matrix::DGEMM(GLOBAL::DONE, Mtempspace, false, ML, false, GLOBAL::DZERO, MtL);

		delete[] L;
		delete[] tempspace;

		/*Matrix MxM(xM, n, p), MetaxTV(etaxTV, n, p), Mtempspace((realdp *)tempspace, p, p, n);
		Matrix::CGEMM(GLOBAL::ZONE, MxM, true, MetaxTV, false, GLOBAL::ZZERO, Mtempspace);*/
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		realdp factor = 1;//-- sqrt(Manifold::Metric(x, x, x));
		integer idx = 0;
		for (integer i = 0; i < p; i++)
		{
			resultTV[idx] = tempspaceL[i + i * n] / factor;
			idx++;
		}
		for (integer i = 0; i < p; i++)
		{
			for (integer j = i + 1; j < p; j++)
			{
				resultTV[idx] = r2 * tempspaceL[j + i * n] / factor;
				idx++;
			}
		}

		for (integer i = 0; i < p; i++)
		{
			for (integer j = p; j < n; j++)
			{
				resultTV[idx] = tempspaceL[j + i * n];
				idx++;
			}
		}
		delete[] work;
		delete[] tempspaceL;
	};

	void NStQOrth::ObtainExtrHGZ(Variable *x, Vector *intretax, Vector *result) const
	{
		if (!x->TempDataExist("HHR"))
		{
			ComputeHHR(x);
		}

		const realdp *xM = x->ObtainReadData();
		const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
		const realdp *ptrHHR = HHR->ObtainReadData();
		const realdp *ptrHHRTau = HHRTau->ObtainReadData();
		const realdp *intretaxTV = intretax->ObtainReadData();
		realdp *resultTV = result->ObtainWriteEntireData();

		integer N = x->Getsize()[0], P = x->Getsize()[1], inc = 1, Length = N * P;
		integer info;
		integer idx = 0;
		//		realdpcomplex *S = new realdpcomplex[p * p];
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		realdp factor = 1;//-- sqrt(Manifold::Metric(x, x, x));
		for (integer i = 0; i < p; i++)
		{
			resultTV[i + i * n] = intretaxTV[idx] * factor;
			idx++;
		}

		for (integer i = 0; i < p; i++)
		{
			for (integer j = i + 1; j < p; j++)
			{
				resultTV[j + i * n] = intretaxTV[idx] / r2 * factor;
				resultTV[i + j * n] = intretaxTV[idx] / r2 * factor;
				idx++;
			}
		}
		for (integer i = 0; i < p; i++)
		{
			for (integer j = p; j < n; j++)
			{
				resultTV[j + i * n] = intretaxTV[idx];
				idx++;
			}
		}

		realdp sign = 0;
		for (integer i = 0; i < p; i++)
		{
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			// result(i, :) <- sign * result(i, :), details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
			scal_(&P, &sign, resultTV + i, &N);
		}
		integer lwork = -1;
		realdp lworkopt;
		// compute the size of space required in the ormqr
		ormqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), resultTV, &N, &lworkopt, &lwork, &info);
		lwork = static_cast<integer> (lworkopt);

		realdp *work = new realdp[lwork];
		// resultTV <- Q * resultTV, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		ormqr_(GLOBAL::L, GLOBAL::N, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
			(const_cast<realdp *> (ptrHHRTau)), resultTV, &N, work, &lwork, &info);
		delete[] work;

		realdp *L = new realdp[p * p + n * p];
		realdp *r_T = L + p * p;
		for (integer i = 0; i < p; i++)
		{
			for (integer j = 0; j < i; j++)
			{
				L[j + i * p] = 0;
			}
			sign = static_cast<realdp> ((ptrHHR[i + n * i] >= 0) ? 1 : -1);
			for (integer j = i; j < p; j++)
			{
				L[j + i * p] = ptrHHR[i + n * j] * sign;
			}
		}

		/*r_T <-  resultTV transpose*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				r_T[j + i * p] = resultTV[i + j * n];
			}
		}
		/*solve linear system L^H M = r_T, the solution M is stored in r_T*/
		trtrs_(GLOBAL::L, GLOBAL::C, GLOBAL::N, &P, &N, L, &P, r_T, &P, &info);
		/*resultTV <-  r_T transpose*/
		for (integer i = 0; i < n; i++)
		{
			for (integer j = 0; j < p; j++)
			{
				resultTV[i + j * n] = r_T[j + i * p];
			}
		}
		delete[] L;
	};

	void NStQOrth::VectorTransport(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result) const
	{
		if (VecTran == NSOPARALLELIZATION && !HasHHR)
		{
			return Manifold::VectorTransport(x, etax, y, xix, result);
		}

		if (VecTran == NSOPROJECTION && !HasHHR)
		{
			if (IsIntrApproach)
			{
				VectorTransportProj(x, etax, y, xix, result);
			}
			else
				printf("Warning: Vector transport by projection has not been done for extrinsic approach!\n");
			return;
		}

		if (HasHHR)
			return LCVectorTransport(x, etax, y, xix, result);

		printf("Error: VectorTransport has not been done!\n");
	};

	void NStQOrth::VectorTransportProj(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result) const
	{
		Vector *exxix = EMPTYEXTR->ConstructEmpty();
		ObtainExtr(x, xix, exxix);
		ObtainIntr(y, exxix, result);
		delete exxix;
	};

}; /*end of ROPTLIB namespace*/
