
#include "Manifolds/ObliqueQ/ObliqueQ.h"

/*Define the namespace*/
namespace ROPTLIB{

	ObliqueQ::ObliqueQ(integer inn, integer innum) : ProductManifold(1, new Sphere(inn), innum)
	{
		name.assign("ObliqueQ");
		Sphere *S = dynamic_cast<Sphere *> (manifolds[0]);
		S->ChooseSphereParamsSet3();
		IsIntrApproach = false;
		metric = OQEUCLIDEAN;
		delete EMPTYEXTR;
		delete EMPTYINTR;
		n = inn;
		num = innum;
		EMPTYEXTR = new ObliqueQVector(n, num);
		EMPTYINTR = new ObliqueQVector(n - 1, num);
	};

	ObliqueQ::~ObliqueQ(void)
	{
		for (integer i = 0; i < numofmani; i++)
		{
			delete manifolds[i];
		}
	};

	void ObliqueQ::ChooseObliqueQParamsSet1(void)
	{
		Sphere *S = dynamic_cast<Sphere *> (manifolds[0]);
		S->ChooseSphereParamsSet3();
		metric = OQEUCLIDEAN;
		IsIntrApproach = false;
		delete EMPTYINTR;
		EMPTYINTR = new ObliqueQVector(n, num);
		//SetEMPTYINTR();
	};

	void ObliqueQ::ChooseObliqueQParamsSet2(void)
	{
		Sphere *S = dynamic_cast<Sphere *> (manifolds[0]);
		S->ChooseSphereParamsSet3();
		metric = OQSGP;
		IsIntrApproach = false;
		delete EMPTYINTR;
		EMPTYINTR = new ObliqueQVector(n, num);
		//SetEMPTYINTR();
	};

	void ObliqueQ::Projection(Variable *x, Vector *v, Vector *result) const
	{
		ProductManifold::Projection(x, v, result);
		const realdp *xM = x->ObtainReadData();
		const realdp *vptr = result->ObtainReadData();

		realdp *XXT = new realdp[n * n * 2];
		realdp *VXT = XXT + n * n;

		// XXT <- XM XM^T
		Matrix MxM(xM, n, num), MXXT(XXT, n, n), MV(vptr, n, num), MVXT(VXT, n, n);
		Matrix::DGEMM(GLOBAL::DONE, MxM, false, MxM, true, GLOBAL::DZERO, MXXT);
		// XTV <- V XM^T
		Matrix::DGEMM(GLOBAL::DONE, MV, false, MxM, true, GLOBAL::DZERO, MVXT);

		// VXT <= VXT - VXT^T
		for (integer i = 0; i < n; i++)
		{
			for (integer j = i; j < n; j++)
			{
				VXT[i + j * n] -= VXT[j + i * n];
				VXT[j + i * n] = -VXT[i + j * n];
			}
		}
		realdp *resultTV = result->ObtainWritePartialData();
		Matrix Mresult(resultTV, n, num);

		Matrix MXXT2(XXT, n - 1, n - 1, n), MVXT2(VXT, n - 1, n - 1, n);
		Matrix::DSYL(MXXT2, MXXT2, MVXT2);
		for (integer i = 0; i < n; i++)
		{
			VXT[n - 1 + i * n] = 0;
			VXT[i + (n - 1) * n] = 0;
		}
		Matrix::DGEMM(GLOBAL::DNONE, MVXT, false, MxM, false,
			GLOBAL::DONE, Mresult);

		delete[] XXT;

		//ProductManifold::Projection(x, v, result);
		//result->Print("in Z:");//---
		//double *Z = result->ObtainWritePartialData();
		//const double *xptr = x->ObtainReadData();
		//double *tmp = new double[n * n * 2 + n * num];
		//double *tmp2 = tmp + n * n;
		//double *tmp3 = tmp2 + n * n;

		///*tmp = x * x^T*/
		//gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &num, &GLOBAL::DONE, const_cast<double *> (xptr), &n, const_cast<double *> (xptr), &n, &GLOBAL::DZERO, tmp, &n);

		///*tmp2 = x * z^T*/
		//gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &num, &GLOBAL::DONE, const_cast<double *> (xptr), &n, Z, &n, &GLOBAL::DZERO, tmp2, &n);

		//integer info;
		//potrf_(GLOBAL::L, &n, tmp, &n, &info);/*Cholesky decomposition*/
		//potrs_(GLOBAL::L, &n, &n, tmp, &n, tmp2, &n, &info); /*tmp2 = (x * x^T)^{-1} (x * z^T) */
		//ForDebug::Print("tmp2:", tmp2, n, n);//----
		//for (integer i = 0; i < n; i++)
		//{
		//	tmp2[i + i * n] = 0;
		//	for (integer j = 0; j < i; j++)
		//	{
		//		tmp2[j + i * n] = (tmp2[i + j * n] - tmp2[j + i * n]) / 2;
		//		tmp2[i + j * n] = - tmp2[j + i * n];
		//	}
		//} /*tmp2 =  [(z * x^T) (x * x^T)^{-1} - (x * x^T)^{-1} (x * z^T)] / 2 */
		//for (integer i = 0; i < n; i++)
		//{
		//	tmp2[n - 1 + i * n] = 0;
		//	tmp2[i + (n - 1) * n] = 0;
		//}

		///*tmp3 = [Omega, 0; 0, 0] * X; tmp3 is in the vertical space*/
		//gemm_(GLOBAL::N, GLOBAL::N, &n, &num, &n, &GLOBAL::DONE, tmp2, &n, const_cast<double *> (xptr), &n, &GLOBAL::DZERO, tmp3, &n);

		///*result = Z - [Omega_Z, 0; 0, 0] * X*/
		//integer length = n * num;
		//axpy_(&length, &GLOBAL::DNONE, tmp3, &GLOBAL::IONE, Z, &GLOBAL::IONE);

		//delete[] tmp;
	};

	void ObliqueQ::EucGradToGrad(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		if (metric == OQSGP)
		{
			egf->CopyTo(gf);
		}
		else
		{
			ProductManifold::EucGradToGrad(x, egf, gf, prob);
			Projection(x, gf, gf);
			if(metric != OQEUCLIDEAN)
				printf("warning! metric is not specified! Use the Euclidean metric by default.");
		}
	};

	void ObliqueQ::Retraction(Variable *x, Vector *etax, Variable *result, realdp instepsize) const
	{
		if (metric == OQSGP)
		{
			Manifold::Retraction(x, etax, result, 1);
			realdp *resultptr = result->ObtainWritePartialData();
			integer length = n - 1;
			for (integer i = 0; i < num; i++)
			{
				realdp norm2 = dot_(&length, resultptr + i * n, &GLOBAL::IONE, resultptr + i * n, &GLOBAL::IONE);
				if (norm2 >= 1)
				{
					realdp coef = 1.0 / std::sqrt(norm2);
					dscal_(&length, &coef, resultptr + i * n, &GLOBAL::IONE);
					resultptr[i * n + n - 1] = 0;
				}
				else
				{
					resultptr[i * n + n - 1] = std::sqrt(1.0 - norm2);
				}
			}
		}
		else
		{
			ProductManifold::Retraction(x, etax, result, instepsize);
			if (metric != OQEUCLIDEAN)
				printf("warning! metric is not specified! Use the Euclidean metric by default.");
		}

	};

	void ObliqueQ::VectorTransport(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result) const
	{
		if (metric == OQSGP)
		{
			Manifold::VectorTransport(x, etax, y, xix, result);
		}
		else
		{
			ProductManifold::VectorTransport(x, etax, y, xix, result);
			if (metric != OQEUCLIDEAN)
				printf("warning! metric is not specified! Use the Euclidean metric by default.");
		}
	};

	void ObliqueQ::CheckParams(void) const
	{
		ProductManifold::CheckParams();

		std::string OQMetricnames[OQMETRICLENGTH] = { "EUCLIDEAN", "SGP" };
		printf("%s PARAMETERS:\n", name.c_str());
		printf("metric        :%15s,\n", OQMetricnames[metric].c_str());
	};
	void ObliqueQ::SetParams(PARAMSMAP params)
	{
		Manifold::SetParams(params);
		PARAMSMAP::iterator iter;
		for (iter = params.begin(); iter != params.end(); iter++)
		{
			if (iter->first == static_cast<std::string> ("ParamSet"))
			{
				switch (static_cast<integer> (iter->second))
				{
				case 1:
					ChooseObliqueQParamsSet1();
					break;
				case 2:
					ChooseObliqueQParamsSet2();
					break;
				default:
					break;
				}
			}
		}
	};
}; /*end of ROPTLIB namespace*/
