/*
This file defines the class for the manifold 
Ob(n, num) / O_{n-1} = \{[X] | X \in Ob(n, num) \} where
[X] = \{X [Q, 0; 0, 1] | Q^T Q = I \}.
num must be larger than n. We assume that X is always full row rank.

Manifold --> ProductManifold --> ObliqueQ

---- WH
*/

#ifndef OBLIQUEQ_H
#define OBLIQUEQ_H

#include "Manifolds/ProductManifold.h"
#include "Manifolds/Sphere/Sphere.h"
#include "Manifolds/ObliqueQ/ObliqueQVariable.h"
#include "Manifolds/ObliqueQ/ObliqueQVector.h"

/*Define the namespace*/
namespace ROPTLIB{

	enum ObliqueQMetric { OQEUCLIDEAN, OQSGP, OQMETRICLENGTH };

	class ObliqueQ : public ProductManifold{
	public:
		/*Construct an Oblique manifold, which is a product of spheres.
		the number of sphere is num and the unit sphere is in R^n.*/
		ObliqueQ(integer n, integer num);

		/*Delete each component manifold*/
		~ObliqueQ();

		/* Default one: choose qf, parallelization and intrinsic approach and use householder reflections*/
		virtual void ChooseObliqueQParamsSet1();

		/* SGP: spectral gradient descent. */
		virtual void ChooseObliqueQParamsSet2();

		virtual void Projection(Variable *x, Vector *v, Vector *result) const;

		virtual void EucGradToGrad(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const;

		virtual void Retraction(Variable *x, Vector *etax, Variable *result, realdp instepsize) const;

		virtual void VectorTransport(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result) const;

		/*Check whether all the parameters are legal or not.*/
		virtual void CheckParams(void) const;

		/*PARAMSMAP is defined in "def.h" and it is a map from string to realdp, i.e., std::map<std::string, realdp> .
		This function is used to set the parameters by the mapping*/
		virtual void SetParams(PARAMSMAP params);

		mutable integer n;
		mutable integer num;
		ObliqueQMetric metric;
	};
}; /*end of ROPTLIB namespace*/
#endif // end of OBLIQUE_H
