
% ARPGvsGPower

clear;
% % for producing Figure 2
dataset = 1;
rs = [4];
% % for producing  Figure 3
% dataset = 2;
% rs = [4];
% for producing  Figure 4
% dataset = 3;
% rs = [5];
%------------------

rng('shuffle');
seed = floor(rand() * 100000000);
seed = 10;
rng(seed);

figure(1);clf
set(0,'defaultaxesfontsize',10, ...
   'defaultaxeslinewidth',0.7, ...
   'defaultlinelinewidth',.8,'defaultpatchlinewidth',0.8);
set(0,'defaultlinemarkersize',10)
for i = 1 : length(rs)
    if(dataset == 1)

        n = 3000;
        m = 40;

        A = randn(m, n);
        A = A - repmat(mean(A, 1), m, 1);
        A = A ./ repmat(sqrt(sum(A .* A)), m, 1);
    elseif(dataset == 2)
        dat = importdata('dataBCT.txt');
        D = (dat.data)';
        [m, n] = size(D);
        A = D - ones(size(D,1),1)*mean(D, 1);
        tmp = sqrt(sum(A .* A));
        A = A ./ repmat(tmp, m, 1);
    elseif(dataset == 3)
        m = 400; n = 4000;
        lambda = 1.5;
        noiselevel = 0.5;
        t = linspace(0, 1, n);

        pc1 = max(0, ((t - 0.6)> 0) & (0.8 - t > 0))*0.5;
        pc2 = max(0, ((t - 0.2)> 0) & (0.4 - t > 0))*0.5;
        pc3 = 0.8*exp(-(t - 0.5).^2/5e-3);
        pc4 = 0.4*exp(-(t - 0.15).^2/1e-3) + 0.4*exp(-(t - 0.85).^2/1e-3);
        pc5 = 0.4*exp(-(t - 0.05).^2/1e-3) - 0.4*exp(-(t - 0.95).^2/1e-3);
        PC = [pc1; pc2; pc3; pc4; pc5];
        truesparsity = sum(sum(PC~=0)) / length(PC(:))
        A = [ones(m/5,1)*pc1 + randn(m/5,n) * noiselevel; ...
             ones(m/5,1)*pc2 + randn(m/5,n) * noiselevel; ...
             ones(m/5,1)*pc3 + randn(m/5,n) * noiselevel; ...
             ones(m/5,1)*pc4 + randn(m/5,n) * noiselevel; ...
             ones(m/5,1)*pc5 + randn(m/5,n) * noiselevel];
        tmp = sqrt(sum(A .* A));
        A = A ./ repmat(tmp, m, 1);
    else
        fprintf('no such data\n');
        return;
    end
    r = rs(i);
    
    [U, S, V] = svd(A, 'econ');
    PCAV = V(:, 1:r);
    tmp = A * PCAV; maxvar = sum(tmp(:) .* tmp(:));

    fprintf('n:%d, r:%d, m:%d, seed:%d\n', n, r, m, seed);

    N=[1:r].^(-1);
    spa = [0, 1];
    mus = [0, 1];
    tol = 0.05;

    Gpowerl1 = @(mu)GPower(A, mu * ones(1, r), r, 'l1', 1, N);
    [spal1, musl1, Zsl1] = SparsitiesForMus(Gpowerl1, spa, mus, tol, n, r);
    [navarl1, avarl1] = variance(Zsl1, A);
    [corvl1] = correlation(Zsl1, A);
    orthdiffl1 = orthogonality(Zsl1);

    Gpowerl0 = @(mu)GPower(A, mu * ones(1, r), r, 'l0', 1, N);
    [spal0, musl0, Zsl0] = SparsitiesForMus(Gpowerl0, spa, mus, tol, n, r);
    [navarl0, avarl0] = variance(Zsl0, A);
    [corvl0] = correlation(Zsl0, A);
    orthdiffl0 = orthogonality(Zsl0);


    norm_a_max = max(sum(A .* A).^0.5);
    phi_init = PCAV;
    mu_max = 5 * norm_a_max;
    
    ARPG = @(mu)driver_ManPG_invariants(5, A, phi_init, n, r, m, 1e-8*n*r, 3000, mu * mu_max, -inf);
    %driver_AManPG(A, phi_init, n, r, m, 1e-8*n*r, 3000, mu * mu_max, -inf);

    [~,~,~, spmax, ~] = ARPG(1); 
    while(spmax < 0.95)
        mu_max = mu_max * 2;
        ARPG = @(mu)driver_ManPG_invariants(5, A, phi_init, n, r, m, 1e-8*n*r, 3000, mu * mu_max, -inf);
        %@(mu)driver_AManPG(A, phi_init, n, r, m, 1e-8*n*r, 3000, mu * mu_max, -inf);
        [~,~,~, spmax, ~] = ARPG(1); 
        [mu_max, spmax]%%---
        fprintf('sparsity:%f\n', spmax);
    end
    fprintf('max sparsity:%f\n', spmax);
    spa(2) = spmax;
    [spaMF, musMF, ZsMF] = SparsitiesForMus(ARPG, spa, mus, tol, n, r);
    [navarMF, avarMF] = variance(ZsMF, A);
    [corvMF] = correlation(ZsMF, A);
    orthdiffMF = orthogonality(ZsMF);

    set(0,'defaultaxesfontsize',15, ...
       'defaultaxeslinewidth',0.7, ...
       'defaultlinelinewidth',.8,'defaultpatchlinewidth',0.8);
    set(0,'defaultlinemarkersize',10)
    subplot(length(rs), 3, (i-1) * 3 + 1); hold on;
    plot(spal1, navarl1 / maxvar, '*r');
    plot(spal0, navarl0 / maxvar, 'vb');
    plot(spaMF, navarMF / maxvar, 'xm');
    legend('GPower l1', 'GPower l0', 'AManPG-D')
    xlabel('Sparsity');
    ylabel('Non-adjusted variance');
    axis([0, 1.1, 0, 1.1]);

    subplot(length(rs), 3, (i-1) * 3 + 2); hold on;
    plot(spal1, avarl1 / maxvar, '*r');
    plot(spal0, avarl0 / maxvar, 'vb');
    plot(spaMF, avarMF / maxvar, 'xm');
    legend('GPower l1', 'GPower l0', 'AManPG-D')
    xlabel('Sparsity')
    ylabel('Adjusted variance');
    axis([0, 1.1, 0, 1.1]);

    subplot(length(rs), 3, (i-1) * 3 + 3); hold on;
    idx = find(orthdiffl1 < 0.5); plot(spal1(idx), orthdiffl1(idx), '*r');
    idx = find(orthdiffl0 < 0.5); plot(spal0(idx), orthdiffl0(idx), 'vb');
    plot(spaMF, orthdiffMF, 'xm');
    legend('GPower l1', 'GPower l0', 'AManPG-D')
    xlabel('Sparsity');
    ylabel('||X''X-I||');
    title('Non-orthogonality of the loadings');
end
return;

function orthdiff = orthogonality(Zs)
    n = length(Zs);
    r = size(Zs{1}, 2);
    orthdiff = zeros(n, 1);
    for i = 1 : n
        tmp = Zs{i}' * Zs{i};
        orthdiff(i) = norm(tmp - eye(size(tmp)));
    end
end

function [corv] = correlation(Zs, A)
    n = length(Zs);
    navar = zeros(n, 1);
    avar = zeros(n, 1);
    for i = 1 : n
        tmp = A * Zs{i}; tmp = tmp' * tmp;
        corv(i) = norm(tmp - diag(tmp), 'fro') / norm(diag(tmp));
    end
end

function [navar, avar] = variance(Zs, A)
    n = length(Zs);
    navar = zeros(n, 1);
    avar = zeros(n, 1);
    for i = 1 : n
        tmp = A * Zs{i};
        navar(i) = sum(tmp(:) .* tmp(:));
        [Q, R] = qr(tmp, 0);
        avar(i) = trace(R * R);
    end
end

function [spa, mus, Zs] = SparsitiesForMus(funhandle, spa, mus, tol, n, r)
    [v, idx] = max(spa(2:end) - spa(1:end-1));
    Zs = cell(1, 2);
    iter = 0;
    
    while(v > tol && min(abs(mus(2:end) - mus(1:end-1))) > 1e-4)
        mu = (mus(idx) + mus(idx + 1)) / 2;
        Z = funhandle(mu);
        sp = sum(abs(Z(:)) < 1e-7) / n / r;
        mus = [mus(1:idx), mu, mus(idx+1:end)];
        spa = [spa(1:idx), sp, spa(idx+1:end)];
        Zs = {Zs{1:idx}, Z, Zs{idx + 1:end}};
        [v, idx]= max(spa(2:end) - spa(1:end-1));
        fprintf('iter:%d, new mu:%e, new sparsity:%e\n', iter, mu, sp);
        iter = iter + 1;
    end
    Zs = Zs(2:end-1);
    spa = spa(2:end-1);
    mus = mus(2:end-1);
end


function [xopt, iter, Ds, sparsity, comtime, avar] = driver_ManPG_invariants(method, A, Xinitial, n, p, m, tolsq, maxiter, lambda, ftol)
    tol = sqrt(tolsq);
    if(norm(sum(A .* A) - ones(1, n)) < 1e-5) % if the columns of A are normalized
        LengthW = p;
    else
        LengthW = n * p;
    end
    %LengthW  %1     : use Lipschitz constant,        
              %n * r : Diagonal matrix inspired from Riemannian Hessian
              %r     : If matrix A is normalized, then the diagonal matrix can be simplified
    
    switch method
        case 1
            SolverParams.method = 'RPG'; % RPG D
            SolverParams.RPGVariant = 0;
            SolverParams.LengthW = LengthW;
        case 2
            SolverParams.method = 'RPG'; % RPG
            SolverParams.RPGVariant = 0;
            SolverParams.LengthW = 1;
        case 3
            SolverParams.method = 'RPG'; % RPG D ada
            SolverParams.RPGVariant = 1;
            SolverParams.LengthW = LengthW;
        case 4
            SolverParams.method = 'RPG'; % RPG ada
            SolverParams.RPGVariant = 1;
            SolverParams.LengthW = 1;
        case 5
            SolverParams.method = 'ARPG'; % ARPG D
            SolverParams.RPGVariant = 0;
            SolverParams.LengthW = LengthW;
%             SolverParams.SGIterGap = inf;
        case 6
            SolverParams.method = 'ARPG'; % ARPG
            SolverParams.RPGVariant = 0;
            SolverParams.LengthW = 1;
%             SolverParams.SGIterGap = 100000;
    end

    SolverParams.IsCheckParams = 1;
%     SolverParams.RPGVariant = 0; %0: RPG without adaptive stepsize, 1: RPG with adaptive stepsize
    SolverParams.OutputGap = 1;
    SolverParams.Max_Iteration = maxiter;
    SolverParams.DEBUG = 1;
    SolverParams.Tolerance = tol;
    [xopt, fs, Ds, gfgf0, iter, nf, ng, nR, nV, nVp, nH, comtime, fsall, Dsall, times] = ...
        TestStieSPCA(A, lambda, Xinitial, SolverParams);
    
    xopt.main(abs(xopt.main) < 1e-5) = 0;
    sparsity = sum(sum(abs(xopt.main) < 1e-5)) / (n * p);
    xopt = reshape(xopt.main, n, p);
    % adjusted variance
    [Q, R] = qr(A * xopt, 0);
    avar = trace(R * R);
end

