
#include "Problems/CFR2BlindDeconvolution/CFR2BlindDeconvolution.h"

#ifdef ROPTLIB_WITH_FFTW

/*Define the namespace*/
namespace ROPTLIB {

	CFR2BlindDeconvolution::CFR2BlindDeconvolution(realdp *iny, realdp *inB, integer innzmaxB, int *inirB, int *injcB, bool inisBsparse,
		realdp *inC, integer innzmaxC, int *inirC, int *injcC, bool inisCsparse, integer inL, integer inK, integer inN, integer inr, realdp inrho, realdp ind, realdp inmu)
	{
		y = iny;

		B = inB;
		nzmaxB = innzmaxB;
		irB = inirB;
		jcB = injcB;
		isBsparse = inisBsparse;

		C = inC;
		nzmaxC = innzmaxC;
		irC = inirC;
		jcC = injcC;
		isCsparse = inisCsparse;

		L = inL;
		K = inK;
		N = inN;
		r = inr;

		rho = inrho;
		d = ind;
		mu = inmu;

		flags = FFTW_ESTIMATE;
		p = fftw_plan_dft_1d(L, nullptr, nullptr, FFTW_FORWARD, flags);

		if (isBsparse)
		{
			sB = BLAS_zuscr_begin(L, K);
			BLAS_uscr_insert_entries(sB, nzmaxB, B, irB, jcB);
			BLAS_zuscr_end(sB);
		}
		if (isCsparse && C != nullptr)
		{
			sC = BLAS_zuscr_begin(L, N);
			BLAS_uscr_insert_entries(sC, nzmaxC, C, irC, jcC);
			BLAS_zuscr_end(sC);
		}
		//log2L = static_cast<int> (log(static_cast<float> (L)) / log(static_cast<float> (2)) + 0.5);
		if (C == nullptr && (static_cast<int> (pow(2.0, static_cast<int>(log(static_cast<float> (L)) / log(static_cast<float> (2)) + 0.5))) - L != 0))
			printf("Warning: L must be a power of 2!\n");

		//realdp *testarr = new realdp[1024 * 1024 * 2];
		//for (integer i = 0; i < 1024 * 1024 * 2; i++)
		//	testarr[i] = 1;
		//unsigned long starttime = getTickCount();
		//fftwrapper(1024 * 1024, 1, (fftw_complex *)testarr, (fftw_complex *)testarr, FFTW_FORWARD);
		//printf("computational time:%e\n", static_cast<realdp>(getTickCount() - starttime) / CLK_PS);//---
		//std::cout << "=========================" << std::endl;//---
		//delete[] testarr;
	};

	CFR2BlindDeconvolution::~CFR2BlindDeconvolution(void)
	{
		fftw_destroy_plan(p);
		if (isBsparse)
			BLAS_usds(sB);
		if (isCsparse && C != nullptr)
			BLAS_usds(sC);
	};

	void CFR2BlindDeconvolution::fftwrapper(integer inn, integer inr, fftw_complex *in, fftw_complex *out, int sign) const
	{
		for (integer i = 0; i < inr; i++)
		{
			p = fftw_plan_dft_1d(inn, in + inn * i, out + inn * i, sign, flags);
			fftw_execute(p);
		}
	};

	void CFR2BlindDeconvolution::BtimesU(realdp *U, realdp *result) const
	{
		if (isBsparse)
		{
			for (integer i = 0; i < r * 2 * L; i++)
				result[i] = 0;
			BLAS_usmm(blas_colmajor, blas_no_trans, r, &GLOBAL::ZONE, sB, (realdpcomplex*)U, K, (realdpcomplex*)result, L);
		}
		else
		{
			integer LL = L, rr = r, KK = K;
			gemm_(GLOBAL::N, GLOBAL::N, &LL, &rr, &KK, &GLOBAL::ZONE, (realdpcomplex*)B, &LL, (realdpcomplex*)U, &KK, &GLOBAL::ZZERO, (realdpcomplex *)result, &LL);
		}
	};

	void CFR2BlindDeconvolution::CtimesV(realdp *V, realdp *result) const
	{
		if (C == nullptr)
		{
			if (r != 1)
				printf("Warning: CFR2BlindDeconvolution::CtimesV: suppose r = 1!\n");
			for (integer i = 0; i < 2 * N; i++)
				result[i] = V[i];

			for (integer i = 2 * N; i < 2 * L; i++)
				result[i] = 0;
			haarFWT_1d_inverse(L, (realdpcomplex *)result);
			return;
		}
		if (isCsparse)
		{
			for (integer i = 0; i < r * 2 * L; i++)
				result[i] = 0;
			BLAS_usmm(blas_colmajor, blas_no_trans, r, &GLOBAL::ZONE, sC, (realdpcomplex*)V, N, (realdpcomplex*)result, L);
		}
		else
		{
			integer LL = L, rr = r, NN = N;
			gemm_(GLOBAL::N, GLOBAL::N, &LL, &rr, &NN, &GLOBAL::ZONE, (realdpcomplex*)C, &LL, (realdpcomplex*)V, &NN, &GLOBAL::ZZERO, (realdpcomplex*)result, &LL);
		}
	};

	void CFR2BlindDeconvolution::BHtimesM(realdp *M, realdp *result) const
	{
		if (isBsparse)
		{
			for (integer i = 0; i < 2 * r * K; i++)
				result[i] = 0;
			BLAS_usmm(blas_colmajor, blas_conj_trans, r, &GLOBAL::ZONE, sB, (realdpcomplex*)M, L, (realdpcomplex*)result, K);
		}
		else
		{
			integer LL = L, rr = r, KK = K;
			gemm_(GLOBAL::C, GLOBAL::N, &KK, &rr, &LL, &GLOBAL::ZONE, (realdpcomplex *)B, &LL, (realdpcomplex *)M, &LL, &GLOBAL::ZZERO, (realdpcomplex *)result, &KK);
		}
	};

	void CFR2BlindDeconvolution::CHtimesM(realdp *M, realdp *result) const
	{ /*Data in M will be destroy in this function*/
		if (C == nullptr)
		{
			if (r != 1)
				printf("Warning: CFR2BlindDeconvolution::CHtimesM: suppose r = 1!\n");

			haarFWT_1d(L, (realdpcomplex *)M);

			for (integer i = 0; i < 2 * N; i++)
			{
				result[i] = M[i];
			}

			return;
		}
		if (isCsparse)
		{
			for (integer i = 0; i < 2 * r * N; i++)
				result[i] = 0;
			BLAS_usmm(blas_colmajor, blas_conj_trans, r, &GLOBAL::ZONE, sC, (realdpcomplex*)M, L, (realdpcomplex*)result, N);
		}
		else
		{
			integer LL = L, rr = r, NN = N;
			gemm_(GLOBAL::C, GLOBAL::N, &NN, &rr, &LL, &GLOBAL::ZONE, (realdpcomplex *)C, &LL, (realdpcomplex *)M, &LL, &GLOBAL::ZZERO, (realdpcomplex *)result, &NN);
		}
	};

	realdp CFR2BlindDeconvolution::f(Variable *x) const
	{
		const realdp *xptr = x->ObtainReadData();
		const realdp *Uptr = xptr;
		const realdp *Vptr = xptr + 2 * K * r;
		SharedSpace *BUCV = new SharedSpace(1, L * r * 4);
		realdp *BU = BUCV->ObtainWriteEntireData();
		realdp *CV = BU + L * r * 2;
		integer LL2 = 2 * L, KK = K, NN = N, rr = r;
		BtimesU(const_cast<realdp *>(Uptr), BU);
		fftwrapper(L, r, (fftw_complex *)BU, (fftw_complex *)BU, FFTW_FORWARD);
		CtimesV(const_cast<realdp *> (Vptr), CV);
		fftwrapper(L, r, (fftw_complex *)CV, (fftw_complex *)CV, FFTW_BACKWARD);

		SharedSpace *ymdiagBXA = new SharedSpace(1, L * 2);
		realdp *tmpptr = ymdiagBXA->ObtainWriteEntireData();
		for (integer i = 0; i < L; i++) // this for loop can be improved
		{
			tmpptr[2 * i] = y[2 * i] - (BU[2 * i] * CV[2 * i] + BU[2 * i + 1] * CV[2 * i + 1]);
			tmpptr[2 * i + 1] = y[2 * i + 1] - (BU[2 * i + 1] * CV[2 * i] - BU[2 * i] * CV[2 * i + 1]);
		}

		realdp penalty = 0;
		if (rho != 0)
		{
			realdp tmp = static_cast<realdp> (L) / 8 / d / d / mu / mu;
			realdp tmp2 = 0;
			SharedSpace *rownorm2BU = new SharedSpace(1, L + 1);
			realdp *rownorm2BUptr = rownorm2BU->ObtainWriteEntireData();
			integer length = r;
			for (integer i = 0; i < L; i++)
			{
				rownorm2BUptr[i] = dot_(&length, BU + i * 2, &LL2, BU + i * 2, &LL2);
				rownorm2BUptr[i] += dot_(&length, BU + i * 2 + 1, &LL2, BU + i * 2 + 1, &LL2);

			}
			length = N * 2 * r;
			rownorm2BUptr[L] = dot_(&length, const_cast<realdp *> (Vptr), &GLOBAL::IONE, const_cast<realdp *> (Vptr), &GLOBAL::IONE);
			for (integer i = 0; i < L; i++)
			{
				tmp2 = ((tmp * rownorm2BUptr[i] * rownorm2BUptr[L] - 1 < 0) ? 0 : (tmp * rownorm2BUptr[i] * rownorm2BUptr[L] - 1));
				penalty += rho * tmp2 * tmp2;
			}
			x->AddToTempData("rownorm2BU", rownorm2BU);
		}
		realdp result = dot_(&LL2, tmpptr, &GLOBAL::IONE, tmpptr, &GLOBAL::IONE) + penalty;

		if (UseGrad)
		{
			x->AddToTempData("ymdiagBXA", ymdiagBXA);
			x->AddToTempData("BUCV", BUCV);
		}
		else
		{
			delete ymdiagBXA;
			delete BUCV;
		}
		return result;
	};

	void CFR2BlindDeconvolution::EucGrad(Variable *x, Vector *gf) const
	{
		const realdp *xptr = x->ObtainReadData();
		const realdp *Uptr = xptr;
		const realdp *Vptr = xptr + 2 * K * r;

		const SharedSpace *ymdiagBXA = x->ObtainReadTempData("ymdiagBXA");
		const realdp *tmpptr = ymdiagBXA->ObtainReadData();
		const SharedSpace *BUCV = x->ObtainReadTempData("BUCV");
		const realdp *BU = BUCV->ObtainReadData();
		const realdp *CV = BU + 2 * r * L;

		integer NN = N, LL = L, rr = r, KK = K, LL2 = L * 2;

		realdp *EGFV = new realdp[2 * K * r + 2 * N * r];
		realdp *EGFTU = EGFV + 2 * K * r;

		realdp *tmpp = new realdp[2 * r * L];
		integer length = 2 * r * L;

		/*Computing EFGV*/
		copy_(&length, const_cast<realdp *> (CV), &GLOBAL::IONE, tmpp, &GLOBAL::IONE);

		for (integer i = 0; i < L; i++)
			scal_(&rr, (realdpcomplex *)(tmpptr + 2 * i), (realdpcomplex *)(tmpp + 2 * i), &LL);
		fftwrapper(L, r, (fftw_complex *)tmpp, (fftw_complex *)tmpp, FFTW_BACKWARD);
		BHtimesM(tmpp, EGFV);

		/*Computing EFGTU*/
		copy_(&length, const_cast<realdp *> (BU), &GLOBAL::IONE, tmpp, &GLOBAL::IONE);
		scal_(&LL, &GLOBAL::DNONE, const_cast<realdp *> (tmpptr + 1), &GLOBAL::ITWO);

		for (integer i = 0; i < L; i++)
			scal_(&rr, (realdpcomplex *)(tmpptr + 2 * i), (realdpcomplex *)(tmpp + 2 * i), &LL);
		scal_(&LL, &GLOBAL::DNONE, const_cast<realdp *> (tmpptr + 1), &GLOBAL::ITWO);
		fftwrapper(L, r, (fftw_complex *)tmpp, (fftw_complex *)tmpp, FFTW_FORWARD);
		CHtimesM(tmpp, EGFTU);

		delete[] tmpp;

		length = 2 * r * KK;
		realdp ntwo = -2;
		scal_(&length, &ntwo, EGFV, &GLOBAL::IONE); /*gfh*/
		length = 2 * r * NN;
		scal_(&length, &ntwo, EGFTU, &GLOBAL::IONE); /*gfm*/

		if (rho != 0)
		{
			const SharedSpace *rownorm2BU = x->ObtainReadTempData("rownorm2BU");
			const realdp *rownorm2BUptr = rownorm2BU->ObtainReadData();

			realdp *gGh = new realdp[2 * K * r + 2 * L * r];
			realdp *DFBU = gGh + 2 * K * r;
			length = 2 * L * r;
			copy_(&length, const_cast<realdp *> (BU), &GLOBAL::IONE, DFBU, &GLOBAL::IONE);
			realdpcomplex coef = { 0, 0 };
			realdp tmp = static_cast<realdp> (L) / 8 / d / d / mu / mu;

			for (integer i = 0; i < L; i++)
			{
				coef.r = 2.0 * tmp * rho * 2 * ((tmp * rownorm2BUptr[i] * rownorm2BUptr[L] - 1 < 0) ? 0 : tmp * rownorm2BUptr[i] * rownorm2BUptr[L] - 1) * rownorm2BUptr[L];
				scal_(&rr, &coef, (realdpcomplex *)(DFBU + 2 * r * i), &LL);
			}
			fftwrapper(L, r, (fftw_complex *)DFBU, (fftw_complex *)DFBU, FFTW_BACKWARD);
			BHtimesM(DFBU, gGh);
			length = 2 * K * r;
			axpy_(&length, &GLOBAL::DONE, gGh, &GLOBAL::IONE, EGFV, &GLOBAL::IONE);
			delete[] gGh;

			realdp coef2 = 0;
			for (integer i = 0; i < L; i++)
			{
				coef2 += 2.0 * tmp * rho * rownorm2BUptr[i] * 2 * ((tmp * rownorm2BUptr[i] * rownorm2BUptr[L] - 1 < 0) ? 0 : tmp * rownorm2BUptr[i] * rownorm2BUptr[L] - 1);
			}
			length = 2 * r * NN;
			axpy_(&length, &coef2, const_cast<realdp *> (Vptr), &GLOBAL::IONE, EGFTU, &GLOBAL::IONE);
		}

		/*Compute (H^*H)^{-1} H^* eta_H */
		SharedSpace *invVVUU = new SharedSpace(1, 4 * r * r);
		realdpcomplex *VV = (realdpcomplex *)(invVVUU->ObtainWriteEntireData());
		realdpcomplex *UU = VV + r * r;

		Matrix MVV((realdp*)VV, r, r), MV(Vptr, N, r);
		// MVV <- MV^* MV
		Matrix::CGEMM(GLOBAL::ZONE, MV, true, MV, false, GLOBAL::ZZERO, MVV);
		Matrix MUU((realdp*)UU, r, r), MU(Uptr, K, r);
		// MUU <- MU^* MU
		Matrix::CGEMM(GLOBAL::ZONE, MU, true, MU, false, GLOBAL::ZZERO, MUU);

		realdp *gfh = gf->ObtainWriteEntireData();
		realdp *gfm = gfh + 2 * K * r;

		integer info = 0;
		potrf_(GLOBAL::L, &rr, VV, &rr, &info);
		potri_(GLOBAL::L, &rr, VV, &rr, &info);
		Matrix MEGFV(EGFV, K, r), Mgfh(gfh, K, r);
		Matrix::CGEMM(GLOBAL::ZONE, MEGFV, false, MVV, false, GLOBAL::ZZERO, Mgfh);
		if (info != 0)
		{
			printf("warning: potrs failed in CFR2BlindDeconvolution::EucGrad with info:%d!\n", info);
		}

		// solve for EGFTU (MUU)^{-1}
		potrf_(GLOBAL::L, &rr, UU, &rr, &info);
		potri_(GLOBAL::L, &rr, UU, &rr, &info);
		Matrix MEGFTU(EGFTU, N, r), Mgfm(gfm, N, r);
		Matrix::CGEMM(GLOBAL::ZONE, MEGFTU, false, MUU, false, GLOBAL::ZZERO, Mgfm);
		if (info != 0)
		{
			printf("warning: potrs failed in CFR2BlindDeconvolution::EucGrad with info:%d!\n", info);
		}
		delete[] EGFV;
		if (UseHess)
		{
			x->AddToTempData("invVVUU", invVVUU);
		}
		else
		{
			delete invVVUU;
		}
	};

	void CFR2BlindDeconvolution::EucHessianEta(Variable *x, Vector *etax, Vector *xix) const
	{ /*The Hessian of the penalty term is not considered here*/
		const realdp *xptr = x->ObtainReadData();
		const realdp *Uptr = xptr;
		const realdp *Vptr = xptr + 2 * K * r;
		const realdp *dUptr = etax->ObtainReadData();
		const realdp *dVptr = dUptr + 2 * K * r;

		const SharedSpace *BUCV = x->ObtainReadTempData("BUCV");
		const realdp *BU = BUCV->ObtainReadData();
		const realdp *CV = BU + 2 * r * L;

		realdp *BdU = new realdp[L * r * 4];
		realdp *CdV = BdU + L * r * 2;
		realdp *mdiagBetaxA = new realdp[L * 2];

		for (integer i = 0; i < 2 * L; i++)
		{
			mdiagBetaxA[i] = 0;
		}

		integer LL2 = 2 * L, KK = K, NN = N, rr = r, LL = L;
		BtimesU(const_cast<realdp *> (dUptr), BdU);
		fftwrapper(L, r, (fftw_complex *)BdU, (fftw_complex *)BdU, FFTW_FORWARD);
		for (integer i = 0; i < L; i++) // this for loop can be improved
		{
			mdiagBetaxA[2 * i] -= (BdU[2 * i] * CV[2 * i] + BdU[2 * i + 1] * CV[2 * i + 1]);
			mdiagBetaxA[2 * i + 1] -= (BdU[2 * i + 1] * CV[2 * i] - BdU[2 * i] * CV[2 * i + 1]);
		}

		CtimesV(const_cast<realdp *> (dVptr), CdV);
		fftwrapper(L, r, (fftw_complex *)CdV, (fftw_complex *)CdV, FFTW_BACKWARD);
		for (integer i = 0; i < L; i++) // this for loop can be improved
		{
			mdiagBetaxA[2 * i] -= (BU[2 * i] * CdV[2 * i] + BU[2 * i + 1] * CdV[2 * i + 1]);
			mdiagBetaxA[2 * i + 1] -= (BU[2 * i + 1] * CdV[2 * i] - BU[2 * i] * CdV[2 * i + 1]);
		}

		realdp *EHFV = new realdp[2 * K * r + 2 * N * r], *EHFTU = EHFV + 2 * K * r;
		realdp *tmpp = new realdp[2 * r * L];
		integer length = 2 * r * L;
		/*Computing EHFV*/
		copy_(&length, const_cast<realdp *> (CV), &GLOBAL::IONE, tmpp, &GLOBAL::IONE);

		for (integer i = 0; i < L; i++)
			scal_(&rr, (realdpcomplex *)(mdiagBetaxA + 2 * i), (realdpcomplex *)(tmpp + 2 * i), &LL);

		fftwrapper(L, r, (fftw_complex *)tmpp, (fftw_complex *)tmpp, FFTW_BACKWARD);
		BHtimesM(tmpp, EHFV);

		/*Computing EHFTU*/
		copy_(&length, const_cast<realdp *> (BU), &GLOBAL::IONE, tmpp, &GLOBAL::IONE);
		scal_(&LL, &GLOBAL::DNONE, const_cast<realdp *> (mdiagBetaxA + 1), &GLOBAL::ITWO);

		for (integer i = 0; i < L; i++)
			scal_(&rr, (realdpcomplex *)(mdiagBetaxA + 2 * i), (realdpcomplex *)(tmpp + 2 * i), &LL);

		scal_(&LL, &GLOBAL::DNONE, const_cast<realdp *> (mdiagBetaxA + 1), &GLOBAL::ITWO);
		fftwrapper(L, r, (fftw_complex *)tmpp, (fftw_complex *)tmpp, FFTW_FORWARD);
		CHtimesM(tmpp, EHFTU);

		delete[] tmpp;
		delete[] mdiagBetaxA;

		length = 2 * r * KK;
		realdp ntwo = -2;
		scal_(&length, &ntwo, EHFV, &GLOBAL::IONE);
		length = 2 * r * NN;
		scal_(&length, &ntwo, EHFTU, &GLOBAL::IONE);

		const SharedSpace *ymdiagBXA = x->ObtainReadTempData("ymdiagBXA");
		const realdp *tmpptr = ymdiagBXA->ObtainReadData();
		realdp *EGFdV = new realdp[2 * K * r], *EGFTdU = new realdp[2 * N * r];
		tmpp = new realdp[2 * r * L];
		length = 2 * r * L;
		/*Computing EGFdV*/
		copy_(&length, const_cast<realdp *> (CdV), &GLOBAL::IONE, tmpp, &GLOBAL::IONE);

		for (integer i = 0; i < L; i++)
			scal_(&rr, (realdpcomplex *)(tmpptr + 2 * i), (realdpcomplex *)(tmpp + 2 * i), &LL);
		fftwrapper(L, r, (fftw_complex *)tmpp, (fftw_complex *)tmpp, FFTW_BACKWARD);
		BHtimesM(tmpp, EGFdV);

		/*Computing EGFTdU*/
		copy_(&length, const_cast<realdp *> (BdU), &GLOBAL::IONE, tmpp, &GLOBAL::IONE);
		scal_(&LL, &GLOBAL::DNONE, const_cast<realdp *> (tmpptr + 1), &GLOBAL::ITWO);

		for (integer i = 0; i < L; i++)
			scal_(&rr, (realdpcomplex *)(tmpptr + 2 * i), (realdpcomplex *)(tmpp + 2 * i), &LL);

		scal_(&LL, &GLOBAL::DNONE, const_cast<realdp *> (tmpptr + 1), &GLOBAL::ITWO);
		fftwrapper(L, r, (fftw_complex *)tmpp, (fftw_complex *)tmpp, FFTW_FORWARD);
		CHtimesM(tmpp, EGFTdU);

		delete[] tmpp;

		ntwo = -2;
		length = 2 * r * KK;
		axpy_(&length, &ntwo, EGFdV, &GLOBAL::IONE, EHFV, &GLOBAL::IONE);
		length = 2 * r * NN;
		axpy_(&length, &ntwo, EGFTdU, &GLOBAL::IONE, EHFTU, &GLOBAL::IONE);

		delete[] EGFdV;
		delete[] EGFTdU;
		delete[] BdU;

		/*Compute (H^*H)^{-1} H^* eta_H */
		const SharedSpace *invVVUU = x->ObtainReadTempData("invVVUU");
		const realdpcomplex *VV = (realdpcomplex *)(invVVUU->ObtainReadData());
		const realdpcomplex *UU = VV + r * r;

		Matrix MVV((realdp*)VV, r, r), MUU((realdp*)UU, r, r);
		realdp *Hfh = xix->ObtainWriteEntireData();
		realdp *Hfm = Hfh + 2 * K * r;

		Matrix MEHFV(EHFV, K, r), MHfh(Hfh, K, r);
		Matrix::CGEMM(GLOBAL::ZONE, MEHFV, false, MVV, false, GLOBAL::ZZERO, MHfh);

		Matrix MEHFTU(EHFTU, N, r), MHfm(Hfm, N, r);
		Matrix::CGEMM(GLOBAL::ZONE, MEHFTU, false, MUU, false, GLOBAL::ZZERO, MHfm);

		delete[] EHFV;
	};

}; /*end of ROPTLIB namespace*/
#endif
