
#include "Problems/EucQuadratic/EucQuadratic.h"

/*Define the namespace*/
namespace ROPTLIB{

	EucQuadratic::EucQuadratic(realdp *M, integer dim)
	{
		Dim = dim;
		A = M;
	};

	EucQuadratic::~EucQuadratic(void)
	{
	};

	realdp EucQuadratic::f(Variable *x) const
	{
		const realdp *v = x->ObtainReadData();
		SharedSpace *Temp = new SharedSpace(1, Dim);
		realdp *temp = Temp->ObtainWriteEntireData();

		char *transn = const_cast<char *> ("n");
		realdp one = 1, zero = 0;
		integer inc = 1, N = Dim;
		// temp <- A * v, details: http://www.netlib.org/lapack/explore-html/dc/da8/gemv_8f.html
		gemv_(transn, &N, &N, &one, A, &N, const_cast<realdp *> (v), &inc, &zero, temp, &inc);

		x->AddToTempData("Ax", Temp);

		// output v^T temp, details: http://www.netlib.org/lapack/explore-html/d5/df6/dot_8f.html
		return dot_(&N, const_cast<realdp *> (v), &inc, temp, &inc);
	};

	void EucQuadratic::Grad(Variable *x, Vector *gf) const
	{
		realdp *gfTV = gf->ObtainWriteEntireData();
		const SharedSpace *Temp = x->ObtainReadTempData("Ax");
		const realdp *v = Temp->ObtainReadData();

		integer N = Dim, inc = 1;
		// gfTV <- v, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		copy_(&N, const_cast<realdp *> (v), &inc, gfTV, &inc);
		realdp two = 2;
		// gfTV <- 2 * gfTV, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		scal_(&N, &two, gfTV, &inc);
	};

	void EucQuadratic::HessianEta(Variable *x, Vector *etax, Vector *xix) const
	{
		const realdp *v = etax->ObtainReadData();
		realdp *xixTV = xix->ObtainWriteEntireData();

		char *transn = const_cast<char *> ("n");
		integer N = Dim, inc = 1;
		realdp two = 2, zero = 0;
		// xixTV <- 2 * A * v, details: http://www.netlib.org/lapack/explore-html/dc/da8/gemv_8f.html
		gemv_(transn, &N, &N, &two, A, &N, const_cast<realdp *> (v), &inc, &zero, xixTV, &inc);
	};
}; /*end of ROPTLIB namespace*/
