//
//  KarcherMean.cpp
//  Updated
//
//  Created by Yaqing You on 5/4/16.
//  Copyright © 2016 Yaqing You. All rights reserved.
//

#include "Problems/KarcherMean/KarcherMean.h"

/*Define the namespace*/
namespace ROPTLIB{
    KarcherMean::KarcherMean(realdp *Curves, integer innumP, integer indim, integer innumC, integer innumS)
    {
        Qs = Curves;
        numP = innumP;
        dim = indim;
        numC = innumC;
        numS = innumS;
    }

    KarcherMean::~KarcherMean(void)
    {
    };
    
    realdp LinesearchInput(integer iter, Variable *x1, Vector *eta1, realdp initialstepsize, realdp initialslope, const Problem *Prob, const Solvers *solver)
    {
        return 1;
    }

    void KarcherMean::ComputeGeodesic(const realdp *q1, const realdp *q2, realdp &distance, realdp *w) const
    {
        integer numofmanis = 3;    //what is this ?
        integer numofmani1 = 1;    //what is this ?
        integer numofmani2 = 1;    //what is this ?
        integer numofmani3 = 1;    //what is this ?
        
        L2Sphere *ll = new L2Sphere(numP);
        OrthGroup *OO = new OrthGroup(dim);
        Euclidean *mm = new Euclidean(1);
        
        L2SphereVariable *llv = new L2SphereVariable(numP);
        OrthGroupVariable *OOv = new OrthGroupVariable(dim);
        EucVariable *mmv = new EucVariable(1);
        
        ProductManifold *Domain = nullptr;
        Domain = new ProductManifold(numofmanis, ll, numofmani1, OO, numofmani2, mm, numofmani3);
        
        ProductElement *Xinitial = new ProductElement(numofmanis, llv, numofmani1, OOv, numofmani2, mmv, numofmani3);
        //Xinitial->RandInManifold();
        realdp *lOm = Xinitial->ObtainWriteEntireData();
        for (integer i = 0; i < numP; i++) {
            lOm[i] = 1.0;
        }
        for (integer i = 0; i < dim; i++) {
            for (integer j = 0; j < dim; j++) {
                if(i == j) lOm[numP+j+i*dim] = 1.0;
                else lOm[numP+j+i*dim] = 0.0;
            }
        }
        lOm[numP+dim*dim] = 0.0;
        
        
        //realdp *lOm = Xinitial->ObtainWriteEntireData();
        // ######### Read l ###########
        // ######### Read O ###########
        // ######### Read m ###########
        
        ShapePathStraighten Prob(const_cast<realdp *>(q1), const_cast<realdp *>(q2), numP, dim, numC);
        Prob.SetDomain(Domain);
        LRBFGS *LRBFGSsolver = new LRBFGS(&Prob, Xinitial);
        LRBFGSsolver->Debug = NOOUTPUT;
        LRBFGSsolver->Max_Iteration = 20;
        LRBFGSsolver->Stop_Criterion = FUN_REL;
        LRBFGSsolver->Tolerance = static_cast<realdp> (1e-3);
        //LRBFGSsolver->CheckParams();
        LRBFGSsolver->Run();
        
        const realdp *q2_new = LRBFGSsolver->GetXopt()->ObtainReadTempData("q2_new")->ObtainReadData();
        
        PSCVariable PSCV(numP, dim, numC);
        PSCV.Generate(const_cast<realdp *>(q1), const_cast<realdp *>(q2_new));
        //PSCV.Print();
        //ForDebug::Print("3rd", PSCV.ObtainReadData() + 3 * numP * dim, numP, dim); //----D
        PreShapeCurves PSCurves(numP, dim, numC);
        PreShapePathStraighten PreSPSprob(numP, dim, numC);
        PreSPSprob.SetDomain(&PSCurves);
        
         RSD *RSDsolver = new RSD(&PreSPSprob, &PSCV);
         //RSDsolver->LineSearch_LS = static_cast<LSAlgo> (i);
         RSDsolver->Debug= NOOUTPUT;
         RSDsolver->LineSearch_LS = INPUTFUN;
         RSDsolver->LinesearchInput = &LinesearchInput;
         RSDsolver->Max_Iteration = 10;
         RSDsolver->Stop_Criterion = GRAD_F;
         RSDsolver->Tolerance = static_cast<realdp> (1e-10);
         //RSDsolver->CheckParams();
         RSDsolver->Run();
         const realdp *Dalpha = RSDsolver->GetXopt()->ObtainReadTempData("Dalpha")->GetSharedElement()->ObtainReadData();
         //RSDsolver->GetXopt()->Print("OPT:", false);//---
         integer NXD = numP*dim;
         copy_(&NXD, const_cast<realdp *> (Dalpha + NXD*(numC-1)), &GLOBAL::IONE, w, &GLOBAL::IONE); //======
         realdp coeff = -1.0;//--
         //realdp coeff = -1.0/std::sqrt(PreShapePathStraighten::InnerProd_Q(w, w, numP, dim));
         scal_(&NXD, &coeff, w, &GLOBAL::IONE);
        
        
       //const realdp *eta = LRBFGSsolver->GetXopt()->ObtainReadTempData("eta")->ObtainReadData();
        distance = LRBFGSsolver->Getfinalfun();
        
        
        //copy_(&NXD, const_cast<realdp *> (eta), &GLOBAL::IONE, w, &GLOBAL::IONE);
        delete RSDsolver;
        delete LRBFGSsolver;

		delete ll;
		delete OO;
		delete mm;
		delete llv;
		delete OOv;
		delete mmv;
		delete Domain;
		delete Xinitial;
    }

    realdp KarcherMean::f(Variable *x) const
    {
        const realdp *Curve_x = x->ObtainReadData();
        //x->Print("fhjskdhashfkjasdhfklajhsdflkahskl");

        realdp distance = 0.0, dis_temp;
        integer NXD = numP*dim;
        realdp *w = new realdp[NXD];
        SharedSpace *SharedGrad = new SharedSpace(2, numP, dim);
        realdp *grad = SharedGrad->ObtainWriteEntireData();
        for (integer i = 0; i < NXD; i++) grad[i] = 0.0;  //initialize
        for (integer i = 0; i < numS; i++)
        {
            //ComputeGeodesic(Curve_x, Qs+i*numP*dim, dis_temp, w);
            ComputeGeodesic(Qs+i*numP*dim, Curve_x, dis_temp, w);    ////==========================gaigaigai=======
            distance = distance + dis_temp*dis_temp;
            axpy_(&NXD, &GLOBAL::DNONE, w, &GLOBAL::IONE, grad, &GLOBAL::IONE);
        }
        distance = distance/(2*numS);
        realdp coeff = static_cast<realdp> (1) / realdp(numS);
        scal_(&NXD, &coeff, grad, &GLOBAL::IONE);
        
        x->AddToTempData("grad", SharedGrad);
        
        delete[] w;
        
        return distance;
    }

    void KarcherMean::EucGrad(Variable *x, Vector *egf) const
    {
        const realdp *grad = x->ObtainReadTempData("grad")->ObtainReadData();
        realdp *egfGrad = egf->ObtainWriteEntireData();
        integer NXD = numP*dim;
        copy_(&NXD, const_cast<realdp *> (grad), &GLOBAL::IONE, egfGrad, &GLOBAL::IONE);
    }

}; /*end of ROPTLIB namespace*/
