
#include "Problems/MNCRModel/MNCRModel.h"

/*Define the namespace*/
namespace ROPTLIB{

	MNCRModel::MNCRModel(realdp *inc, realdp *inA, realdp inL, integer inm)
	{
		c = inc;
		A = inA;
		L = inL;
		m = inm;
	};

	MNCRModel::~MNCRModel(void)
	{
	};

	realdp MNCRModel::f(Variable *x) const
	{
		const realdp *xptr = x->ObtainReadData();
		realdp result = dot_(&m, const_cast<realdp *> (xptr), &GLOBAL::IONE, c, &GLOBAL::IONE);
		SharedSpace *SharedAx = new SharedSpace(1, m);
		realdp *tmpv = SharedAx->ObtainWriteEntireData();
		gemv_(GLOBAL::N, &m, &m, &GLOBAL::DONE, A, &m, const_cast<realdp *> (xptr), &GLOBAL::IONE, &GLOBAL::DZERO, tmpv, &GLOBAL::IONE);
		result += L * dot_(&m, const_cast<realdp *> (xptr), &GLOBAL::IONE, tmpv, &GLOBAL::IONE);
		x->AddToTempData("Ax", SharedAx);
		return result;
	};

	void MNCRModel::EucGrad(Variable *x, Vector *gf) const
	{
		const SharedSpace *SharedAx = x->ObtainReadTempData("Ax");
		const realdp *Ax = SharedAx->ObtainReadData();
		realdp *gfptr = gf->ObtainWriteEntireData();
		copy_(&m, c, &GLOBAL::IONE, gfptr, &GLOBAL::IONE);
		realdp coe = 2 * L;
		axpy_(&m, &coe, const_cast<realdp *> (Ax), &GLOBAL::IONE, gfptr, &GLOBAL::IONE);
	};

	void MNCRModel::EucHessianEta(Variable *x, Vector *etax, Vector *xix) const
	{
		//etax->CopyTo(xix);//--
		//return;
		const realdp *etaxptr = etax->ObtainReadData();
		realdp *xixptr = xix->ObtainWriteEntireData();
		gemv_(GLOBAL::N, &m, &m, &GLOBAL::DONE, A, &m, const_cast<realdp *> (etaxptr), &GLOBAL::IONE, &GLOBAL::DZERO, xixptr, &GLOBAL::IONE);
		realdp coe = 2 * L;
		scal_(&m, &coe, xixptr, &GLOBAL::IONE);
	};
}; /*end of ROPTLIB namespace*/
