
#include "Problems/NSOLyapunov/NSOLyapunov.h"

/*Define the namespace*/
namespace ROPTLIB {

	NSOLyapunov::NSOLyapunov(realdp *inA, integer innzmaxA, int *inirA, int *injcA, bool inisAsparse,
		realdp *inM, integer innzmaxM, int *inirM, int *injcM, bool inisMsparse,
		realdp *inC, integer inpC, integer inn, integer inp)
	{
		A = inA;
		nzmaxA = innzmaxA;
		irA = inirA;
		jcA = injcA;
		isAsparse = inisAsparse;

		M = inM;
		nzmaxM = innzmaxM;
		irM = inirM;
		jcM = injcM;
		isMsparse = inisMsparse;

		C = inC;
		pC = inpC;

		n = inn;
		p = inp;

		if (isAsparse)
		{
			sA = BLAS_duscr_begin(n, n);
			BLAS_uscr_insert_entries(sA, nzmaxA, A, irA, jcA);
			BLAS_duscr_end(sA);
		}

		if (isMsparse)
		{
			sM = BLAS_duscr_begin(n, n);
			BLAS_uscr_insert_entries(sM, nzmaxM, M, irM, jcM);
			BLAS_duscr_end(sM);
		}
	};

	void NSOLyapunov::AtimesX(realdp *X, integer num, realdp *result) const
	{

		if (isAsparse)
		{
			for (integer i = 0; i < n * num; i++)
				result[i] = 0;
			BLAS_usmm(blas_colmajor, blas_no_trans, num, GLOBAL::DONE, sA, X, n, result, n); /*result = 1 * A * X + result */
		}
		else
		{
			integer N = n, P = p;
			gemm_(GLOBAL::N, GLOBAL::N, &N, &num, &N, &GLOBAL::DONE, A, &N, X, &N, &GLOBAL::DZERO, result, &N);
		}
	};

	void NSOLyapunov::MtimesX(realdp *X, integer num, realdp *result) const
	{

		if (isMsparse)
		{
			for (integer i = 0; i < n * num; i++)
				result[i] = 0;
			BLAS_usmm(blas_colmajor, blas_no_trans, num, GLOBAL::DONE, sM, X, n, result, n); /*result = 1 * A * X + result */
		}
		else
		{
			integer N = n, P = p;
			gemm_(GLOBAL::N, GLOBAL::N, &N, &num, &N, &GLOBAL::DONE, M, &N, X, &N, &GLOBAL::DZERO, result, &N);
		}
};

	NSOLyapunov::~NSOLyapunov(void)
	{
	};

	realdp NSOLyapunov::f(Variable *x) const
	{
		const realdp *Y = x->ObtainReadData();
		SharedSpace *SharedAY = new SharedSpace(2, n, p);
		SharedSpace *SharedMY = new SharedSpace(2, n, p);
		SharedSpace *SharedCY = new SharedSpace(2, pC, p);
		SharedSpace *SharedYTAY = new SharedSpace(2, p, p);
		SharedSpace *SharedYTMY = new SharedSpace(2, p, p);

		realdp *AY = SharedAY->ObtainWriteEntireData();
		realdp *MY = SharedMY->ObtainWriteEntireData();
		realdp *CY = SharedCY->ObtainWriteEntireData();
		realdp *YTAY = SharedYTAY->ObtainWriteEntireData();
		realdp *YTMY = SharedYTMY->ObtainWriteEntireData();

		integer N = n, P = p, PC = pC;
		AtimesX(const_cast<realdp *> (Y), P, AY);
		/*AY = A * Y*/
//		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &N, &GLOBAL::DONE, A, &N, const_cast<realdp *>(Y), &N, &GLOBAL::DZERO, AY, &N);

		/*MY = M * Y*/
		MtimesX(const_cast<realdp *> (Y), P, MY);
//		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &N, &GLOBAL::DONE, M, &N, const_cast<realdp *>(Y), &N, &GLOBAL::DZERO, MY, &N);

		/*CY = Ct * Y*/
		gemm_(GLOBAL::T, GLOBAL::N, &PC, &P, &N, &GLOBAL::DONE, C, &N, const_cast<realdp *>(Y), &N, &GLOBAL::DZERO, CY, &PC);

		/*YTAY = Y^T * A * Y*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *>(Y), &N, AY, &N, &GLOBAL::DZERO, YTAY, &P);

		/*YTMY = Y^T * M * Y*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *>(Y), &N, MY, &N, &GLOBAL::DZERO, YTMY, &P);

		integer length = p * p, length2 = pC * p;

		realdp result = dot_(&length, YTAY, &GLOBAL::IONE, YTMY, &GLOBAL::IONE) - dot_(&length2, const_cast<realdp *> (CY), &GLOBAL::IONE, CY, &GLOBAL::IONE);

		x->AddToTempData("AY", SharedAY);
		x->AddToTempData("MY", SharedMY);
		x->AddToTempData("CY", SharedCY);
		x->AddToTempData("YTAY", SharedYTAY);
		x->AddToTempData("YTMY", SharedYTMY);

		return result;
	};

	void NSOLyapunov::EucGrad(Variable *x, Vector *gf) const
	{
		const SharedSpace *SharedAY = x->ObtainReadTempData("AY");
		const SharedSpace *SharedMY = x->ObtainReadTempData("MY");
		const SharedSpace *SharedCY = x->ObtainReadTempData("CY");
		const SharedSpace *SharedYTAY = x->ObtainReadTempData("YTAY");
		const SharedSpace *SharedYTMY = x->ObtainReadTempData("YTMY");

		const realdp *AY = SharedAY->ObtainReadData();
		const realdp *MY = SharedMY->ObtainReadData();
		const realdp *CY = SharedCY->ObtainReadData();
		const realdp *YTAY = SharedYTAY->ObtainReadData();
		const realdp *YTMY = SharedYTMY->ObtainReadData();


		realdp *gfv = gf->ObtainWriteEntireData();

		integer N = n, P = p, PC = pC;
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &PC, &GLOBAL::DONE, C, &N, const_cast<realdp *> (CY), &PC, &GLOBAL::DZERO, gfv, &N);

		realdp NTWO = -2;
		/*gfv = 2 * AY * YTMY - 2 * C * Ct * Y*/
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DTWO, const_cast<realdp *> (AY), &N, const_cast<realdp *> (YTMY), &P, &NTWO, gfv, &N);

		/*gfv = gfv + 2 * MY * YTAY*/
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DTWO, const_cast<realdp *>(MY), &N, const_cast<realdp *>(YTAY), &P, &GLOBAL::DONE, gfv, &N);
	};

	void NSOLyapunov::EucHessianEta(Variable *x, Vector *etax, Vector *exix) const
	{
		const SharedSpace *SharedAY = x->ObtainReadTempData("AY");
		const SharedSpace *SharedMY = x->ObtainReadTempData("MY");
		const SharedSpace *SharedYTAY = x->ObtainReadTempData("YTAY");
		const SharedSpace *SharedYTMY = x->ObtainReadTempData("YTMY");

		const realdp *Y = x->ObtainReadData();
		const realdp *AY = SharedAY->ObtainReadData();
		const realdp *MY = SharedMY->ObtainReadData();
		const realdp *YTAY = SharedYTAY->ObtainReadData();
		const realdp *YTMY = SharedYTMY->ObtainReadData();

		const realdp *etaTV = etax->ObtainReadData();
		realdp *exixTV = exix->ObtainWriteEntireData();

		realdp *Aeta = new realdp[2 * n * p + p * p * 2 + pC * p];
		realdp *Meta = Aeta + n * p;
		realdp *YTMeta = Meta + n * p;
		realdp *YTAeta = YTMeta + p * p;
		realdp *Cteta = YTAeta + p * p;
		integer N = n, P = p, PC = pC;

		/*Meta = M * eta*/
		MtimesX(const_cast<realdp *> (etaTV), P, Meta);
//		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &N, &GLOBAL::DONE, M, &N, const_cast<realdp *> (etaTV), &N, &GLOBAL::DZERO, Meta, &N);
		/*YTMeta = Y^T * M * eta*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (Y), &N, Meta, &N, &GLOBAL::DZERO, YTMeta, &P);
		/*Aeta = A * eta*/
		AtimesX(const_cast<realdp *> (etaTV), P, Aeta);
//		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &N, &GLOBAL::DONE, A, &N, const_cast<realdp *> (etaTV), &N, &GLOBAL::DZERO, Aeta, &N);
		/*YTAeta = Y^T * A * eta*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (Y), &N, Aeta, &N, &GLOBAL::DZERO, YTAeta, &P);

		/*exixTV = 2 * A * eta * Y^T * M * Y */
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DTWO, Aeta, &N, const_cast<realdp *> (YTMY), &P, &GLOBAL::DZERO, exixTV, &N);

		/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y*/
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DTWO, Meta, &N, const_cast<realdp *> (YTAY), &P, &GLOBAL::DONE, exixTV, &N);

		/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y + 2 * A * Y * eta^T * M * Y*/
		gemm_(GLOBAL::N, GLOBAL::T, &N, &P, &P, &GLOBAL::DTWO, const_cast<realdp *> (AY), &N, YTMeta, &P, &GLOBAL::DONE, exixTV, &N);

		/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y + 2 * A * Y * eta^T * M * Y + 2 * M * Y * eta^T * A * Y*/
		gemm_(GLOBAL::N, GLOBAL::T, &N, &P, &P, &GLOBAL::DTWO, const_cast<realdp *> (MY), &N, YTAeta, &P, &GLOBAL::DONE, exixTV, &N);

		/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y + 2 * A * Y * eta^T * M * Y + 2 * M * Y * eta^T * A * Y + 2 * A * Y * Y^T * M * \eta*/
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DTWO, const_cast<realdp *> (AY), &N, YTMeta, &P, &GLOBAL::DONE, exixTV, &N);

		/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y + 2 * A * Y * eta^T * M * Y + 2 * M * Y * eta^T * A * Y + 2 * A * Y * Y^T * M * \eta + 2 * M * Y * Y^T * A * eta*/
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DTWO, const_cast<realdp *> (MY), &N, YTAeta, &P, &GLOBAL::DONE, exixTV, &N);

		/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y + 2 * A * Y * eta^T * M * Y + 2 * M * Y * eta^T * A * Y + 2 * A * Y * Y^T * M * \eta + 2 * M * Y * Y^T * A * eta - 2 * C * Ct * \eta*/
		gemm_(GLOBAL::T, GLOBAL::N, &PC, &P, &N, &GLOBAL::DONE, C, &N, const_cast<realdp *> (etaTV), &N, &GLOBAL::DZERO, Cteta, &PC);
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &PC, &GLOBAL::DNTWO, C, &N, Cteta, &PC, &GLOBAL::DONE, exixTV, &N);

		delete[] Aeta;
	};
 
	void NSOLyapunov::PreConditioner(Variable *x, Vector *eta, Vector *result) const
	{
		eta->CopyTo(result);
		return;
		LinearCG(x, eta, result);
	};

	void NSOLyapunov::ActionEH(Variable *x, Vector *eta, Vector *result) const
	{
		NStQOrth *NSODomain = dynamic_cast<NStQOrth *> (Domain);

		Vector *Etreta = Domain->GetEMPTYEXTR()->ConstructEmpty();
		Vector *Etrresult = Domain->GetEMPTYEXTR()->ConstructEmpty();
		Domain->ObtainExtr(x, eta, Etreta);
		const realdp *etaTV = Etreta->ObtainReadData();
		realdp *resultTV = Etrresult->ObtainWriteEntireData();
		const realdp *Y = x->ObtainReadData();

		const SharedSpace *SharedAY = x->ObtainReadTempData("AY");
		const SharedSpace *SharedMY = x->ObtainReadTempData("MY");
		const SharedSpace *SharedYTAY = x->ObtainReadTempData("YTAY");
		const SharedSpace *SharedYTMY = x->ObtainReadTempData("YTMY");
		const realdp *AY = SharedAY->ObtainReadData();
		const realdp *MY = SharedMY->ObtainReadData();
		const realdp *YTAY = SharedYTAY->ObtainReadData();
		const realdp *YTMY = SharedYTMY->ObtainReadData();

		realdp *Aeta = new realdp[2 * n * p + p * p * 2];
		realdp *Meta = Aeta + n * p;
		realdp *YTMeta = Meta + n * p;
		realdp *YTAeta = YTMeta + p * p;
		integer N = n, P = p, PC = pC;

		/*Meta = M * eta*/
		MtimesX(const_cast<realdp *> (etaTV), P, Meta);

		/*YTMeta = Y^T * M * eta*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (Y), &N, Meta, &N, &GLOBAL::DZERO, YTMeta, &P);

		/*Aeta = A * eta*/
		AtimesX(const_cast<realdp *> (etaTV), P, Aeta);

		/*YTAeta = Y^T * A * eta*/
		gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, const_cast<realdp *> (Y), &N, Aeta, &N, &GLOBAL::DZERO, YTAeta, &P);

		/*exixTV = 2 * A * eta * Y^T * M * Y */
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DTWO, Aeta, &N, const_cast<realdp *> (YTMY), &P, &GLOBAL::DZERO, resultTV, &N);

		/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y*/
		gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &GLOBAL::DTWO, Meta, &N, const_cast<realdp *> (YTAY), &P, &GLOBAL::DONE, resultTV, &N);

		/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y + 2 * A * Y * eta^T * M * Y*/
		gemm_(GLOBAL::N, GLOBAL::T, &N, &P, &P, &GLOBAL::DTWO, const_cast<realdp *> (AY), &N, YTMeta, &P, &GLOBAL::DONE, resultTV, &N);

		/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y + 2 * A * Y * eta^T * M * Y + 2 * M * Y * eta^T * A * Y*/
		gemm_(GLOBAL::N, GLOBAL::T, &N, &P, &P, &GLOBAL::DTWO, const_cast<realdp *> (MY), &N, YTAeta, &P, &GLOBAL::DONE, resultTV, &N);

		Domain->EucGradToGrad(x, Etrresult, Etrresult, this);
		Domain->ObtainIntr(x, Etrresult, result);
		delete Etreta;
		delete Etrresult;
		delete []Aeta;
	};

	void NSOLyapunov::LinearCG(Variable *x, Vector *eta, Vector *result) const
	{
		integer length = eta->Getlength();
		Vector *y = eta->ConstructEmpty();
		Vector *p = eta->ConstructEmpty();
		Vector *r = eta->ConstructEmpty();
		Vector *Ap = eta->ConstructEmpty();
		Vector *Ay = eta->ConstructEmpty();
		realdp *yptr = y->ObtainWriteEntireData();
		//for (integer i = 0; i < length; i++) /*initial iterate y = 0*/
		//	yptr[i] = 0;
		eta->CopyTo(y);
		ActionEH(x, y, Ay);
		Domain->VectorMinusVector(x, Ay, eta, r); /*r = A y - b*/
		Domain->ScaleTimesVector(x, -1, r, p); /*p = -r*/
		integer k = 0, maxiter = length;
		realdp alpha = 0, beta = 0;
		realdp err = Domain->Metric(x, r, r);
		realdp err0 = err;
		while (err / err0 > 1e-3 && k < maxiter)
		{
			ActionEH(x, p, Ap);
			alpha = -Domain->Metric(x, r, p) / Domain->Metric(x, p, Ap);
			Domain->scalarVectorAddVector(x, alpha, p, y, y);
			ActionEH(x, y, Ay);
			Domain->VectorMinusVector(x, Ay, eta, r);
			beta = Domain->Metric(x, r, Ap) / Domain->Metric(x, p, Ap);
			Domain->VectorLinearCombination(x, -1, r, beta, p, p);
			err = Domain->Metric(x, r, r);
			k++;
		}
		printf("iter:%d, error in CG:%e\n", k, err);
		y->CopyTo(result);
		delete y;
		delete p;
		delete r;
		delete Ap;
		delete Ay;
	};
}; /*end of ROPTLIB namespace*/

		//Domain->ScaleTimesVector(x, 2, eta, result);//---

		//realdp *Seta = new realdp[p * p * 4 + n * p * 4 + p];
		//realdp *Zeta = Seta + p * p;
		//realdp *AV = Zeta + n * p;
		//realdp *VTAV = AV + n * p;
		//realdp *Q = VTAV + p * p;
		//realdp *Lambda = Q + p * p;
		//realdp *tV = Lambda + p;
		//realdp *tZeta = tV + n * p;
		//realdp *tSeta = Zeta + n * p;

		///*Compute Seta, Zeta*/
		//Vector *Exeta = Domain->GetEMPTYEXTR()->ConstructEmpty();
		//Domain->ObtainExtr(x, eta, Exeta);
		//const realdp *ExetaPtr = Exeta->ObtainReadData();

		//if (!x->TempDataExist("HHR"))
		//{
		//	NStQOrth::ComputeHHR(x);
		//}
		//const SharedSpace *HHR = x->ObtainReadTempData("HHR");
		//const SharedSpace *HHRTau = x->ObtainReadTempData("HHRTau");
		//const realdp *ptrHHR = HHR->ObtainReadData();
		//const realdp *ptrHHRTau = HHRTau->ObtainReadData();

		//realdp *V = new realdp[n * p];
		//for (integer i = 0; i < n *p; i++)
		//	V[i] = 0;
		//for (integer i = 0; i < p; i++)
		//	V[i + i * n] = 1;
		//integer N = n, P = p, info;
		//integer lwork = -1;
		//realdp lworkopt;
		//// compute the size of space required in the ormqr
		//ormqr_(GLOBAL::L, GLOBAL::T, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
		//	(const_cast<realdp *> (ptrHHRTau)), V, &N, &lworkopt, &lwork, &info);
		//lwork = static_cast<integer> (lworkopt);
		//realdp *work = new realdp[lwork];
		//// tempspace <- Q^T * tempspace, where Q is the orthogonal matrix defined as the product elementary reflectors defined by ptrHHR and ptrHHRTau,
		//// details: http://www.netlib.org/lapack/explore-html/da/d82/ormqr_8f.html
		//ormqr_(GLOBAL::L, GLOBAL::T, &N, &P, &P, (const_cast<realdp *> (ptrHHR)), &N,
		//	(const_cast<realdp *> (ptrHHRTau)), V, &N, work, &lwork, &info);
		//delete[] work;

		//realdp half = 0.5, nhalf = -0.5;
		//gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &half, V, &N, const_cast<realdp *> (ExetaPtr), &N, &GLOBAL::DZERO, Seta, &P);

		//gemm_(GLOBAL::N, GLOBAL::N, &N, &P, &P, &nhalf, V, &N, Seta, &P, &GLOBAL::DONE, const_cast<realdp *> (ExetaPtr), &N);

		//delete[] V;


		//delete[] Seta;
		//delete Exeta;
		///*Compute Q, Lambda*/
		//AtimesX(V, P, AV);
		//gemm_(GLOBAL::T, GLOBAL::N, &P, &P, &N, &GLOBAL::DONE, V, &N, AV, &N, &GLOBAL::DZERO, VTAV, &P);
		//Matrix M_VTAV(VTAV, p, p), M_Q(Q, p, p), M_Lambda(Lambda, p, 1);;
		//Matrix::EigenSymmetricM(GLOBAL::U, M_VTAV, M_Lambda, M_Q);

		/*Compute tV, tZeta, tSeta*/

		/*Compute Ki, Vi*/

		/*Compute R*/

		/*Solve linear system for tSeps*/

		/*Compute tZeps*/

		/*Form the result*/
//
//void NSOLyapunov::subproblem1(realdp lambdai, realdp *tV, realdp *B, integer num, realdp *soln) const
//{
//};
//
//void NSOLyapunov::subproblem2(realdp *Ki, realdp *R, realdp *soln) const
//{
//};
