/*
This file defines a problem for finding a low rank approximation for the Lyapunov equation:
A X M + M X A = C C^T,
where A, M, and C are symmetric matrices.
The optimization problem is:
min_{Y \in R_*^{n \times p} / O_p} \tr(Y Y^T A Y Y^T M) - tr(Y Y^T C C^T).

Problem --> NSOLyapunov

---- WH
*/

#ifndef NSOLYAPUNOV_H
#define NSOLYAPUNOV_H

#include "Manifolds/NStQOrth/NStQOrth.h"
#include "Manifolds/NStQOrth/NSOVariable.h"
#include "Manifolds/NStQOrth/NSOVector.h"
#include "Problems/Problem.h"
#include "Manifolds/SharedSpace.h"
#include "Others/def.h"
#include "Others/SparseBLAS/blas_sparse.h"

/*Define the namespace*/
namespace ROPTLIB{

	class NSOLyapunov : public Problem{
	public:
		/*define A, M and C in the cost function \tr(Y Y^T A Y Y^T M) - tr(Y Y^T C)*/
		NSOLyapunov(realdp *inA, integer innzmaxA, int *inirA, int *injcA, bool inisAsparse, 
			realdp *inM, integer innzmaxM, int *inirM, int *injcM, bool inisMsparse,
			realdp *inC, integer inpC, integer inn, integer inp);

		virtual ~NSOLyapunov();
		virtual realdp f(Variable *x) const;

		virtual void EucGrad(Variable *x, Vector *gf) const;

		virtual void EucHessianEta(Variable *x, Vector *etax, Vector *exix) const;

		virtual void PreConditioner(Variable *x, Vector *eta, Vector *result) const;

		void AtimesX(realdp *X, integer num, realdp *result) const;

		void MtimesX(realdp *X, integer num, realdp *result) const;

		void ActionEH(Variable *x, Vector *xi, Vector *result) const;

		void LinearCG(Variable *x, Vector *eta, Vector *result) const;

		realdp *A;
		integer nzmaxA;
		int *irA;
		int *jcA;
		bool isAsparse;
		blas_sparse_matrix sA;

		realdp *M;
		integer nzmaxM;
		int *irM;
		int *jcM;
		bool isMsparse;
		blas_sparse_matrix sM;

		realdp *C;
		integer pC;

		integer n;
		integer p;
	};
}; /*end of ROPTLIB namespace*/

#endif //NSOLYAPUNOV_H
