
#include "Problems/ObliqueQCor/ObliqueQCor.h"

/*Define the namespace*/
namespace ROPTLIB{

	ObliqueQCor::ObliqueQCor(realdp *inA, integer inn, integer innum)
	{
		A = inA;
		n = inn;
		num = innum;
		integer length = num * num;
		C = dot_(&length, A, &GLOBAL::IONE, A, &GLOBAL::IONE) * 0.5;
	};

	ObliqueQCor::~ObliqueQCor(void)
	{
	};

	realdp ObliqueQCor::f(Variable *x) const
	{
		integer np1 = n - 1;
		const realdp *xptr = x->ObtainReadData();

		SharedSpace *SharedAXt = new SharedSpace(2, num, np1);
		realdp *AXt = SharedAXt->ObtainWriteEntireData();
		gemm_(GLOBAL::N, GLOBAL::T, &num, &np1, &num, &GLOBAL::DONE, A, &num, const_cast<realdp *> (xptr), &n, &GLOBAL::DZERO, AXt, &num);

		SharedSpace *SharedXtXXt = new SharedSpace(2, num, np1);
		realdp *XtXXt = SharedXtXXt->ObtainWriteEntireData();
		realdp *tmp = new realdp[np1 * np1];
		gemm_(GLOBAL::N, GLOBAL::T, &np1, &np1, &num, &GLOBAL::DONE, const_cast<realdp *> (xptr), &n, const_cast<realdp *> (xptr), &n, &GLOBAL::DZERO, tmp, &np1);
		gemm_(GLOBAL::T, GLOBAL::N, &num, &np1, &np1, &GLOBAL::DONE, const_cast<realdp *> (xptr), &n, tmp, &np1, &GLOBAL::DZERO, XtXXt, &num);
		delete[] tmp;

		SharedSpace *ShareddiagyyXt = new SharedSpace(2, num, np1);
		realdp *diagyyXt = ShareddiagyyXt->ObtainWriteEntireData();
		for (integer i = 0; i < num; i++)
		{
			double coef = xptr[np1 + n * i];
			for (integer j = 0; j < np1; j++)
			{
				diagyyXt[i + j * num] = coef * coef * xptr[j + i * n];
			}
		}

		realdp *Xt = new realdp[num * np1];
		for (integer i = 0; i < num; i++)
			for (integer j = 0; j < np1; j++)
				Xt[i + j * num] = xptr[j + i * n];

		realdp result = 0;

		integer length = np1 * num;
		result = -dot_(&length, Xt, &GLOBAL::IONE, AXt, &GLOBAL::IONE) + 0.5 * ddot_(&length, Xt, &GLOBAL::IONE, XtXXt, &GLOBAL::IONE) 
			+ ddot_(&length, Xt, &GLOBAL::IONE, diagyyXt, &GLOBAL::IONE);
		for (integer i = 0; i < num; i++)
			result += -A[i + i * num] * xptr[np1 + i * n] * xptr[np1 + i * n] + 0.5 * xptr[np1 + i * n] * xptr[np1 + i * n] * xptr[np1 + i * n] * xptr[np1 + i * n];

		delete[] Xt;
		
		x->AddToTempData("AXt", SharedAXt);
		x->AddToTempData("XtXXt", SharedXtXXt);
		x->AddToTempData("diagyyX", ShareddiagyyXt);
		result += C;
		return result;
	};

	void ObliqueQCor::EucGrad(Variable *x, Vector *egf) const
	{
		integer np1 = n - 1;
		const realdp *xptr = x->ObtainReadData();
		const SharedSpace *SharedAXt = x->ObtainReadTempData("AXt");
		const SharedSpace *SharedXtXXt = x->ObtainReadTempData("XtXXt");
		const SharedSpace *ShareddiagyyXt = x->ObtainReadTempData("diagyyX");
		const realdp *AXt = SharedAXt->ObtainReadData();
		const realdp *XtXXt = SharedXtXXt->ObtainReadData();
		const realdp *diagyyXt = ShareddiagyyXt->ObtainReadData();

		realdp *egfptr = egf->ObtainWriteEntireData();
		for (integer i = 0; i < np1; i++)
		{
			for (integer j = 0; j < num; j++)
			{
				egfptr[i + j * n] = -2.0 * AXt[j + i * num] + 2.0 * XtXXt[j + i * num] + 2.0 * diagyyXt[j + i * num];
			}
		}
		for (integer i = 0; i < num; i++)
			egfptr[np1 + i * n] = 2.0 * xptr[np1 + n * i] - 2.0 * xptr[np1 + n * i] * A[i + i * num];
	};

	void ObliqueQCor::EucHessianEta(Variable *x, Vector *etax, Vector *exix) const
	{
		printf("Warning: The action of the Hessian has not been implemented!\n");
		etax->CopyTo(exix);
	};
}; /*end of ROPTLIB namespace*/
