
#include "Problems/ObliqueTestSparsePCA/ObliqueTestSparsePCA.h"

/*Define the namespace*/
namespace ROPTLIB{

	ObliqueTestSparsePCA::ObliqueTestSparsePCA(realdp *inB, realdp *inDsq, realdp inmu, realdp ineps, integer inp, integer inn, integer inr)
	{
		B = inB;
		Dsq = inDsq;
		p = inp;
		n = inn;
		r = inr;
		mu = inmu;
		epsilon = ineps;
	};

	ObliqueTestSparsePCA::~ObliqueTestSparsePCA(void)
	{
	};

	realdp ObliqueTestSparsePCA::f(Variable *x) const
	{
		const realdp *xptr = x->ObtainReadData();
		realdp epsilonsq = epsilon * epsilon;

		realdp result = 0;
		for (integer i = 0; i < p * r; i++)
		{
			result += sqrt(xptr[i] * xptr[i] + epsilonsq) - epsilon;
		}
		realdp *BtX = new realdp[n * r];

		char *transn = const_cast<char *> ("n"), *transt = const_cast<char *> ("t");
		integer N = n, P = p, R = r;
		realdp one = 1.0, zero = 0.0;
		// BtX <- B^T * xptr, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transt, transn, &N, &R, &P, &one, B, &P, const_cast<realdp *> (xptr), &P, &zero, BtX, &N);
		SharedSpace *SharedBBtX = new SharedSpace(2, p, r);
		realdp *BBtX = SharedBBtX->ObtainWriteEntireData();
		// BBtX = B * BtX, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transn, transn, &P, &R, &N, &one, B, &P, BtX, &N, &zero, BBtX, &P);

		SharedSpace *SharedXtBBtX = new SharedSpace(2, r, r);
		realdp *XtBBtX = SharedXtBBtX->ObtainWriteEntireData();
		// XtBBtX = xptr^T * BBtX, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transt, transn, &R, &R, &P, &one, const_cast<realdp *> (xptr), &P, BBtX, &P, &zero, XtBBtX, &R);

		for (integer i = 0; i < r; i++)
		{
			result += mu * (XtBBtX[i + i * r] - Dsq[i]) * (XtBBtX[i + i * r] - Dsq[i]);
			for (integer j = i + 1; j < r; j++)
			{
				result += mu * 2 * XtBBtX[i + j * r] * XtBBtX[i + j * r];
			}
		}
		//integer rsqr = r * r, inc = 1;
		//result += mu * dot_(&rsqr, XtBBtX, &inc, Dsq, &inc);

		x->AddToTempData("BBtX", SharedBBtX);
		x->AddToTempData("XtBBtX", SharedXtBBtX);
		delete[] BtX;
		return result;
	};

	void ObliqueTestSparsePCA::EucGrad(Variable *x, Vector *egf) const
	{
		const realdp *xptr = x->ObtainReadData();
		realdp *egfptr = egf->ObtainWriteEntireData();
		realdp epsilonsq = epsilon * epsilon;

		for (integer i = 0; i < p * r; i++)
		{
			egfptr[i] = xptr[i] / sqrt(xptr[i] * xptr[i] + epsilonsq);
		}

		const SharedSpace *SharedXtBBtX = x->ObtainReadTempData("XtBBtX");
		const realdp *XtBBtX = SharedXtBBtX->ObtainReadData();

		SharedSpace *SharedXtBBtXmDsq = new SharedSpace(2, r, r);
		realdp *XtBBtXmDsq = SharedXtBBtXmDsq->ObtainWriteEntireData();
		integer rsq = r * r, inc = 1;
		// XtBBtXmDsq <- XtBBtX, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		copy_(&rsq, const_cast<realdp *> (XtBBtX), &inc, XtBBtXmDsq, &inc);
		for (integer i = 0; i < r; i++)
		{
			XtBBtXmDsq[i + i * r] -= Dsq[i];
		}

		const SharedSpace *SharedBBtX = x->ObtainReadTempData("BBtX");
		const realdp *BBtX = SharedBBtX->ObtainReadData();
		integer P = p, R = r;
		char *transn = const_cast<char *> ("n");
		realdp one = 1.0, fourmu = 4 * mu;
		// egfptr = egfptr + 4 * mu * BBtX * XtBBtXmDsq, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transn, transn, &P, &R, &R, &fourmu, const_cast<realdp *> (BBtX), &P, XtBBtXmDsq, &R, &one, egfptr, &P);
		x->AddToTempData("XtBBtXmDsq", SharedXtBBtXmDsq);
	};

	void ObliqueTestSparsePCA::EucHessianEta(Variable *x, Vector *etax, Vector *exix) const
	{
		//etax->CopyTo(exix);
		realdp epsilonsq = epsilon * epsilon;
		const realdp *etaxTV = etax->ObtainReadData();
		const realdp *xptr = x->ObtainReadData();
		realdp *exixTV = exix->ObtainWriteEntireData();
		for (integer i = 0; i < p * r; i++)
		{
			exixTV[i] = etaxTV[i] * epsilonsq / pow(xptr[i] * xptr[i] + epsilonsq, static_cast<realdp> (1.5));
		}

		const SharedSpace *SharedXtBBtXmDsq = x->ObtainReadTempData("XtBBtXmDsq");
		const realdp *XtBBtXmDsq = SharedXtBBtXmDsq->ObtainReadData();

		realdp *etaXtBBtXmDsq = new realdp[p * r + n * r];
		realdp *BtetaXtBBtXmDsq = etaXtBBtXmDsq + p * r;

		char *transn = const_cast<char *> ("n"), *transt = const_cast<char *> ("t");
		integer N = n, P = p, R = r, inc = 1;
		realdp one = 1.0, zero = 0.0;
		// etaXtBBtXmDsq = etaxTV * XtBBtXmDsq, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transn, transn, &P, &R, &R, &one, const_cast<realdp *> (etaxTV), &P, const_cast<realdp *> (XtBBtXmDsq), &R, &zero, etaXtBBtXmDsq, &P);
		// BtetaXtBBtXmDsq = B^T * etaXtBBtXmDsq, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transt, transn, &N, &R, &P, &one, B, &P, etaXtBBtXmDsq, &P, &zero, BtetaXtBBtXmDsq, &N);
		realdp fourmu = 4 * mu;
		// exixTV = exixTV + 4 * mu * B * BtetaXtBBtXmDsq, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transn, transn, &P, &R, &N, &fourmu, B, &P, BtetaXtBBtXmDsq, &N, &one, exixTV, &P);
		delete[] etaXtBBtXmDsq;

		const SharedSpace *SharedBBtX = x->ObtainReadTempData("BBtX");
		const realdp *BBtX = SharedBBtX->ObtainReadData();
		realdp *temp = new realdp[r * r];

		// temp = etaxTV^T * BBtX, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transt, transn, &R, &R, &P, &one, const_cast<realdp *> (etaxTV), &P, const_cast<realdp *> (BBtX), &P, &zero, temp, &R);
		for (integer i = 0; i < r; i++)
		{
			temp[i + i * r] *= 2.0;
			for (integer j = i + 1; j < r; j++)
			{
				temp[i + j * r] += temp[j + i * r];
				temp[j + i * r] = temp[i + j * r];
			}
		}
		// exixTV = exixTV + 4 * mu * BBtX * temp, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transn, transn, &P, &R, &R, &fourmu, const_cast<realdp *> (BBtX), &P, temp, &R, &one, exixTV, &P);
		delete[] temp;
	};
}; /*end of ROPTLIB namespace*/
