﻿
#include "Problems/SPDMean/SPDMean.h"

/*Define the namespace*/
namespace ROPTLIB{

	SPDMean::SPDMean(realdp *inLs, integer inn, integer innum)
	{
		Ls = inLs;
		n = inn;
		num = innum;
	};

	SPDMean::~SPDMean(void)
	{
	};

	realdp SPDMean::f(Variable *x) const
	{
		SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
		if (!x->TempDataExist("L"))
		{
			Mani->CholeskyRepresentation(x);
		}
		const SharedSpace *SharedL = x->ObtainReadTempData("L");
		Variable *LElement = SharedL->GetSharedElement();
		const realdp *L = LElement->ObtainReadData();

		SharedSpace *SharedlogLXL = new SharedSpace(3, n, n, num);
		realdp *logLXL = SharedlogLXL->ObtainWriteEntireData();
		realdp *Ltmp = new realdp[n * n];
		integer length = n * n, N = n, info;
		for (integer i = 0; i < num; i++)
		{
			copy_(&length, const_cast<realdp *> (L), &GLOBAL::IONE, Ltmp, &GLOBAL::IONE);
			/*Solve the linear system Li X = Lx, i.e., X = Li^{-1} Lx. The solution X is stored in LiiLx.
			Note that Li is a lower triangular matrix.
			Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
			trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Ls + n * n * i, &N, Ltmp, &N, &info);
			if (info != 0)
			{
				printf("Solving a linear system in SPDMean::f failed with info:%d!\n", info);
			}
			gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, Ltmp, &N, Ltmp, &N, &GLOBAL::DZERO, logLXL + n * n * i, &N);
			Matrix MMt(logLXL + n * n * i, n, n);
			Matrix::LogSymmetricM(GLOBAL::L, MMt, MMt);
		}
		length = n * n * num;

		delete[] Ltmp;
		x->AddToTempData("logLXL", SharedlogLXL);

		realdp result = nrm2_(&length, logLXL, &GLOBAL::IONE);
		result *= result;
		result /= static_cast<realdp> (2) * num;
		return result;
	};

	void SPDMean::EucGrad(Variable *x, Vector *gf) const
	{
		const SharedSpace *SharedL = x->ObtainReadTempData("L");
		Variable *LElement = SharedL->GetSharedElement();
		const realdp *L = LElement->ObtainReadData();


		const SharedSpace *SharedlogLXL = x->ObtainReadTempData("logLXL");
		const realdp *logLXL = SharedlogLXL->ObtainReadData();
		realdp *gfVT = gf->ObtainWriteEntireData();
		for (integer i = 0; i < n * n; i++)
			gfVT[i] = 0;
		const realdp *xM = x->ObtainReadData();

		realdp *tmp = new realdp[n * n];
		integer N = n, length = n * n, info;

		for (integer i = 0; i < num; i++)
		{
			/*tmp <-- log(Li^{-1} X Li^{-T}) Li^T */
			gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *> (logLXL + n * n * i), &N,
				Ls + n * n * i, &N, &GLOBAL::DZERO, tmp, &N);

			/*Solve the linear system Li^T X = tmp, i.e., X = Li^{-T} log(Li^{-1} X Li^{-T}) Li^T. The solution X is stored in tmp.
			Note that Li is a lower triangular matrix.
			Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
			trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, Ls + n * n * i, &N, tmp, &N, &info);
			if (info != 0)
			{
				printf("The cholesky decompsotion in SPDEucMean::RieGrad failed with info:%d!\n", info);
			}
			axpy_(&length, &GLOBAL::DONE, tmp, &GLOBAL::IONE, gfVT, &GLOBAL::IONE);
		}
		for (integer i = 0; i < n; i++)
			for (integer j = 0; j < n; j++)
				tmp[i + j * n] = gfVT[j + i * n];

		trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<realdp *> (L), &N, tmp, &N, &info);
		trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, const_cast<realdp *> (L), &N, tmp, &N, &info);

		for (integer i = 0; i < n; i++)
			for (integer j = 0; j < n; j++)
				gfVT[i + j * n] = tmp[j + i * n];

		realdp scalar = static_cast<realdp> (1) / num;
		scal_(&length, &scalar, gfVT, &GLOBAL::IONE);
		delete[] tmp;
	};

	//void SPDMean::RieGrad(Variable *x, Vector *gf) const
	//{
	//	const SharedSpace *SharedlogLXL = x->ObtainReadTempData("logLXL");
	//	const realdp *logLXL = SharedlogLXL->ObtainReadData();
	//	realdp *gfVT = gf->ObtainWriteEntireData();
	//	for (integer i = 0; i < n * n; i++)
	//		gfVT[i] = 0;
	//	const realdp *xM = x->ObtainReadData();

	//	realdp *tmp = new realdp[n * n];
	//	integer N = n, info;
	//	for (integer i = 0; i < num; i++)
	//	{
	//		/*tmp <-- log(Li^{-1} X Li^{-T}) Li^T */
	//		gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *> (logLXL + n * n * i), &N,
	//			Ls + n * n * i, &N, &GLOBAL::DZERO, tmp, &N);

	//		/*Solve the linear system Li^T X = tmp, i.e., X = Li^{-T} log(Li^{-1} X Li^{-T}) Li^T. The solution X is stored in tmp.
	//		Note that Li is a lower triangular matrix.
	//		Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
	//		trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, Ls + n * n * i, &N, tmp, &N, &info);
	//		if (info != 0)
	//		{
	//			printf("Solving a linear system in SPDMean::RieGrad failed with info:%d!\n", info);
	//		}
	//		gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *> (xM), &N, tmp, &N, &GLOBAL::DONE, gfVT, &N);
	//	}
	//	realdp scalar = 1.0 / num;
	//	integer length = n * n;
	//	scal_(&length, &scalar, gfVT, &GLOBAL::IONE);
	//	delete[] tmp;
	//};

	void SPDMean::EucHessianEta(Variable *x, Vector *etax, Vector *xix) const
	{
		printf("warning: SPDMean::EucHessianEta has not been implemented!\n");
		etax->CopyTo(xix);
	};
}; /*end of ROPTLIB namespace*/
