
#include "SPDMeanL1.h"

/*Define the namespace*/
namespace ROPTLIB{

	SPDMeanL1::SPDMeanL1(realdp *inLs, integer inn, integer innum)
	{
		Ls = inLs;
		n = inn;
		num = innum;
	};

	SPDMeanL1::~SPDMeanL1(void)
	{
	};

	realdp SPDMeanL1::f(Variable *x) const
	{
		SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
		if (!x->TempDataExist("L"))
		{
			Mani->CholeskyRepresentation(x);
		}
		const SharedSpace *SharedL = x->ObtainReadTempData("L");
		Variable *LElement = SharedL->GetSharedElement();
		const realdp *L = LElement->ObtainReadData();

		SharedSpace *SharedlogLXL = new SharedSpace(3, n, n, num);
		realdp *logLXL = SharedlogLXL->ObtainWriteEntireData();
		
		integer length = n * n, N = n, info;
        realdp a, result = 0.0;
        
        realdp *Ltmp = new realdp[n * n];
        realdp *eigenvalues = new realdp[n * num];
        realdp *eigenvectors = new realdp[2 * n * n];
        realdp *eigenvectorsD = eigenvectors + n * n;
        
        SharedSpace *Sharedcoef = new SharedSpace(3, num, 1, 1);
        realdp *coef = Sharedcoef->ObtainWriteEntireData();
        
		for (integer i = 0; i < num; i++)
		{
			copy_(&length, const_cast<realdp *> (L), &GLOBAL::IONE, Ltmp, &GLOBAL::IONE);
			/*Solve the linear system Li X = Lx, i.e., X = Li^{-1} Lx. The solution X is stored in LiiLx.
			Note that Li is a lower triangular matrix.
			Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
			trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Ls + n * n * i, &N, Ltmp, &N, &info);
			if (info != 0)
			{
				std::cout << "The cholesky decompsotion in SPDMean::f failed with info:" << info << "!" << std::endl;
			}
			gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, Ltmp, &N, Ltmp, &N, &GLOBAL::DZERO, logLXL + n * n * i, &N);
            
             Matrix E(eigenvalues + n * i, n, 1), V(eigenvectors, n, n), VD(eigenvectorsD, n, n);
             Matrix MMt(logLXL + n * n * i, n, n);
             Matrix::EigenSymmetricM(GLOBAL::L, MMt, E, V);

             copy_(&length, eigenvectors, &GLOBAL::IONE, VD.matrix, &GLOBAL::IONE);
             for (integer j = 0; j < n; j++)
             {
                 a = log(eigenvalues[j + n * i]);
                 scal_(&N, &a, eigenvectors + j * n, &GLOBAL::IONE);
             }
             
             
             Matrix MlogLXL(logLXL + n * n * i, n, n);
             Matrix::DGEMM(GLOBAL::DONE, V, false, VD, true, GLOBAL::DZERO, MlogLXL);

             
             coef[i] = 0.0;
             for(integer j = 0; j < N; j++) {coef[i] = coef[i] + log(eigenvalues[j + i * N]) * log(eigenvalues[j + i * N]);}
             coef[i] = std::sqrt(coef[i]);
             result = result + coef[i];
		}
        
        x->AddToTempData("logLXL", SharedlogLXL);
        x->AddToTempData("coef", Sharedcoef);

		delete[] Ltmp;
        delete[] eigenvalues;
        delete[] eigenvectors;
    

		return result;
	};
    
    

	void SPDMeanL1::RieGrad(Variable *x, Vector *gf) const
	{
		const SharedSpace *SharedlogLXL = x->ObtainReadTempData("logLXL");
		const realdp *logLXL = SharedlogLXL->ObtainReadData();
        const realdp *xM = x->ObtainReadData();
        
		realdp *gfVT = gf->ObtainWriteEntireData();
		for (integer i = 0; i < n * n; i++)
			gfVT[i] = 0;
		

		realdp *tmp = new realdp[n * n];
		integer N = n, info;
        
        
        const SharedSpace *Sharedcoef = x->ObtainReadTempData("coef");
        const realdp *coef  = Sharedcoef->ObtainReadData();
        
        realdp coefa, minval;
        
        
        minval = coef[0];
        integer idx = 0;
        for (integer i = 1; i < num; i++)
        {
            if(coef[i] < minval)
            {
                minval = coef[i];
                idx = i;
            }
        }
        
        
        
        if (minval > 1e-16)
        {
		    for (integer i = 0; i < num; i++)
		    {
			    /*tmp <-- log(Li^{-1} X Li^{-T}) Li^T */
			    gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *> (logLXL + n * n * i), &N,
				    Ls + n * n * i, &N, &GLOBAL::DZERO, tmp, &N);

			    /*Solve the linear system Li^T X = tmp, i.e., X = Li^{-T} log(Li^{-1} X Li^{-T}) Li^T. The solution X is stored in tmp.
			    Note that Li is a lower triangular matrix.
			    Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
			    trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, Ls + n * n * i, &N, tmp, &N, &info);
			    if (info != 0)
			    {
				    std::cout << "The cholesky decompsotion in SPDMean::RieGrad failed with info:" << info << "!" << std::endl;
			    }
                coefa = static_cast<realdp> (1)/coef[i];
                gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &coefa, const_cast<realdp *> (xM), &N, tmp, &N, &GLOBAL::DONE, gfVT, &N);
		    }
		 }
		else
		 {
		    gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *> (logLXL + n * n * idx), &N, 
		        Ls + n * n * idx, &N, &GLOBAL::DZERO, tmp, &N);
            
//          Li^{-T} * tmp -> tmp, i.e.,  Li^{-T} * logLXL * Li^T -> tmp
            trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, Ls + n * n * idx, &N, tmp, &N, &info);
            
            if (info != 0)
            {
                printf("The cholesky decompsotion in SPDMean::RieGrad failed with info:%d!\n", info);
            }
//          X * tmp -> eta
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *> (xM), &N, tmp, &N, &GLOBAL::DZERO, gfVT, &N);
		    
		 }
		
		
		delete[] tmp;
	};
    
    
    
    
    
//     realdp SPDMeanL1::dis(Variable *x) const
//     {
//          SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
//         // dis(x, y) = |log(Lx^{-1} y Lx^{-T})|_F
//         if (!x->TempDataExist("L"))
// 		{
// 			Mani->CholeskyRepresentation(x);
// 		}
// 		const SharedSpace *SharedLx = x->ObtainReadTempData("L");
// 		Variable *LElementx = SharedLx->GetSharedElement();
// 		const realdp *Lx = LElementx->ObtainReadData();
//          
//         integer N = n, info, length = n * n;
//         realdp *y = new realdp[n * n];
//         copy_(&length, Xtrue, &GLOBAL::IONE, y, &GLOBAL::IONE);
//          
//         potrf_(GLOBAL::L, &N, y, &N, &info);
//         for (integer j = 0; j < n; j++)
//         {
//             for (integer k = j + 1; k < n; k++)
//             {
//                 y[j + k * n] = 0; //manually make the upper triangle 0
//             }
//         }
// 
// 		realdp *LxLy = new realdp[n * n];
// //         
// 		copy_(&length, y, &GLOBAL::IONE, LxLy, &GLOBAL::IONE);
// // 		/*Solve the linear system Lx X = Ly, i.e., X = Lx^{-1}Ly. The solution X is stored in LxLy
// // 		Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
//         trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<realdp *> (Lx), &N, LxLy, &N, &info);
// //         
// //         /* compute temp = LxLy*(LxLy)^T  */
//         realdp *temp = new realdp[n * n];
//         gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, LxLy, &N, LxLy, &N, &GLOBAL::DZERO, temp, &N);
// //         
//         realdp *eigenvalues = new realdp[n + n * n];
// 	    realdp *eigenvectors = eigenvalues + n;
//         Matrix E(eigenvalues, n, 1), V(eigenvectors, n, n);
//         Matrix MMt(temp, n, n);
// 	    Matrix::EigenSymmetricM(GLOBAL::L, MMt, E, V);
//         
//         realdp result = 0.0;
//         for (integer i = 0; i < N; i++) 
//         {
//             result = result + log(eigenvalues[i])*log(eigenvalues[i]);
//         }
// 
//         result = std::sqrt(result);
// //         realdp result = 1.0;
//         
//         delete[] eigenvalues;   
//         delete[] temp;
//         delete[] y;
//         delete[] LxLy;
//         
//         return result;
//        
//     };
//     
//    
//     
//     
//     
//     realdp SPDMeanL1::distwo(Variable *x, Variable *y) const
//     {
//         SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
//         // dis(x, y) = |log(Lx^{-1} y Lx^{-T})|_F
//         if (!x->TempDataExist("L"))
// 		{
// 			Mani->CholeskyRepresentation(x);
// 		}
// 		const SharedSpace *SharedLx = x->ObtainReadTempData("L");
// 		Variable *LElementx = SharedLx->GetSharedElement();
// 		const realdp *Lx = LElementx->ObtainReadData();
//         
//         if (!y->TempDataExist("L"))
// 		{
// 			Mani->CholeskyRepresentation(y);
// 		}
// 		const SharedSpace *SharedLy = y->ObtainReadTempData("L");
// 		Variable *LElementy = SharedLy->GetSharedElement();
// 		const realdp *Ly = LElementy->ObtainReadData();
//         
//         integer N = n, info;
// 		realdp *LxLy = new realdp[N * N];
//         
//         integer length = n * n;
// 		copy_(&length, const_cast<realdp *> (Ly), &GLOBAL::IONE, LxLy, &GLOBAL::IONE);
// 		/*Solve the linear system Lx X = Ly, i.e., X = Lx^{-1}Ly. The solution X is stored in LxLy
// 		Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
//         trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<realdp *> (Lx), &N, LxLy, &N, &info);
//         
//         /* compute temp = LxLy*(LxLy)^T  */
//         realdp *temp = new realdp[N * N];
//         gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, LxLy, &N, LxLy, &N, &GLOBAL::DZERO, temp, &N);
//         
//         realdp *eigenvalues = new realdp[n + n * n];
// 	    realdp *eigenvectors = eigenvalues + n;
//         Matrix E(eigenvalues, n, 1), V(eigenvectors, n, n);
//         Matrix MMt(temp, n, n);
// 	    Matrix::EigenSymmetricM(GLOBAL::L, MMt, E, V);
//         
//         realdp result = 0.0;
//         for (integer i = 0; i < N; i++) 
//         {
//             result = result + log(eigenvalues[i])*log(eigenvalues[i]);
//         }
// 
//         result = sqrt(result);
//         return result;
//         
//         delete[] eigenvalues;   
//         delete[] temp;
//     }
//     
    

	void SPDMeanL1::RieHessianEta(Variable *x, Vector *etax, Vector *xix) const
	{
		std::cout << "warning: SPDMean::RieHessianEta has not been implemented!" << std::endl;
		etax->CopyTo(xix);
	};
}; /*end of ROPTLIB namespace*/
