
#include "SPDMeanLDOneParam.h"

/*Define the namespace*/
namespace ROPTLIB{

	SPDMeanLDOneParam::SPDMeanLDOneParam(realdp *inLs, integer inn, integer innum, realdp inalpha)
	{
		Ls = inLs;
		n = inn;
		num = innum;
        alpha = inalpha;
	};

	SPDMeanLDOneParam::~SPDMeanLDOneParam(void)
	{
	};


    realdp SPDMeanLDOneParam::f(Variable *x) const
    {
        SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
        if (!x->TempDataExist("L"))
        {
            Mani->CholeskyRepresentation(x);
        }
        const SharedSpace *SharedL = x->ObtainReadTempData("L");
        Variable *LElement = SharedL->GetSharedElement();
        const realdp *L = LElement->ObtainReadData();
        
        const realdp *xxM = x->ObtainReadData();
        
        integer N = n, length = n * n, Nnum = n * n * num, info;
        realdp onehalf = 0.5;
        realdp a1 = static_cast<realdp> ((1.0 - alpha)/2.0), a2 = static_cast<realdp> ((1.0 + alpha)/2.0);
        
        realdp *I = new realdp[n * n]; // make identity matrix
        for (integer i = 0; i < length; i++) I[i] = 0.0;
        for (integer i = 0; i < n; i++) I[i * n + i] = 1.0;
        
        
        // CLplusX is used to store a1 * Ai + a2 * x        
        SharedSpace *SharedCLplusX = new SharedSpace(3, n, n, num);
        realdp *CLplusX = SharedCLplusX->ObtainWriteEntireData();
        
        realdp *coef1 = new realdp[num];
        
       
        copy_(&Nnum, Ls, &GLOBAL::IONE, CLplusX, &GLOBAL::IONE);
        
        for (integer i = 0; i < num; i++)
        {
            // compute a1 * Ai + a2 * x
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &a2, const_cast<realdp *>(xxM), &N, I, &N, &a1, CLplusX + n * n * i, &N);
            
            // Cholesky factorization of CLplusX
            potrf_(GLOBAL::L, &N, CLplusX + n * n * i, &N, &info);
            for (integer j = 0; j < n; j++)
                for (integer k = j + 1; k < n; k++)
                    CLplusX[j + k * n + i * n * n] = 0;
            
            // compute the determinant 
            coef1[i] = 1.0;
            for (integer j = 0; j < n; j++)
            {
                coef1[i] = coef1[i] * CLplusX[j * n + j + i * n * n];
            }
            coef1[i] = log(coef1[i] * coef1[i]);            
        }
        x->AddToTempData("CLplusX", SharedCLplusX);
        
        
        // compute the determinant of X: product of diagonal elements
        realdp detX = 1.0;
        for (integer j = 0; j < n; j++)
        {
            detX = detX * const_cast<realdp *>(L)[j * n + j];
        }
        detX = log(detX * detX);
        detX = a2 * num * detX;

        
        realdp result = 0.0;
        for (integer i = 0; i < num; i++)
        {
            result = result + coef1[i];
        }
        result = result - detX;
        result = static_cast<realdp> ((result * 4.0)/(1 - alpha * alpha));
        
        
        delete[] I;
        delete[] coef1;

        return result;
    };
    
    
    
    void SPDMeanLDOneParam::RieGrad(Variable *x, Vector *gf) const
    {        
        const SharedSpace *SharedCLplusX = x->ObtainReadTempData("CLplusX");
        const realdp *CLplusX = SharedCLplusX->ObtainReadData();
        
        SharedSpace *SharediLplusX = new SharedSpace(3, n, n, num); // used to store the inverse of L
        realdp *iLplusX = SharediLplusX->ObtainWriteEntireData();
        
        
        const realdp *xM = x->ObtainReadData();
        realdp *gfVT = gf->ObtainWriteEntireData();
        
        for (integer i = 0; i < n * n; i++) gfVT[i] = 0.0;
        
        
        realdp *tmp = new realdp[n * n];
        realdp *LLplusX = new realdp[n * n * num];
        
        integer length = n * n, N = n, Nnum = n * n * num, info;
        realdp negone = -1.0;
        realdp *I = new realdp[n * n]; // make identity matrix
        for (integer i = 0; i < length; i++) I[i] = 0.0;
        for (integer i = 0; i < n; i++) I[i * n + i] = 1.0;

        
        copy_(&Nnum, const_cast<realdp *>(CLplusX), &GLOBAL::IONE, LLplusX, &GLOBAL::IONE);
        
        for (integer i = 0; i < num; i++)
        {
            // Compute inv(LplusX) -> iLplusX
            copy_(&length, I, &GLOBAL::IONE, iLplusX + n * n * i, &GLOBAL::IONE);
            trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, LLplusX + n * n * i, &N, iLplusX + n * n * i, &N, &info);
            
            // compute inv(LplusX) = inv(LplusX)^T * inv(LplusX)
            copy_(&length, iLplusX + n * n * i, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
            gemm_(GLOBAL::T, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, tmp, &N, tmp, &N, &GLOBAL::DZERO, iLplusX + n * n * i, &N);
            
            // add each iLplusX to gfVT
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, iLplusX + n * n * i, &N, I, &N, &GLOBAL::DONE, gfVT, &N);
        }
        
        
         // compute X * gfVT -> tmp
         gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *>(xM), &N, gfVT, &N, &GLOBAL::DZERO, tmp, &N);
            
        // compute tmp * X -> gfVT
        gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, tmp, &N, const_cast<realdp *>(xM), &N, &GLOBAL::DZERO, gfVT, &N);

        // compute gfVT - num * X -> tmp
        copy_(&length, gfVT, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        copy_(&length, const_cast<realdp *>(xM), &GLOBAL::IONE, gfVT, &GLOBAL::IONE);
        realdp negnum = static_cast<realdp> (-1.0) * num;
        gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, tmp, &N, I, &N, &negnum, gfVT, &N);

        
        realdp scalar = static_cast<realdp> (2.0/(1.0 - alpha));
        scal_(&length, &scalar, gfVT, &GLOBAL::IONE);
        
        x -> AddToTempData("iLplusX", SharediLplusX);
        
        delete[] I;
        delete[] tmp;
        delete[] LLplusX;
    };
    
    
    
    
    
    
    void SPDMeanLDOneParam::RieHessianEta(Variable *x, Vector *etax, Vector *xix) const
    {
        
        if (!x -> TempDataExist("iLplusX"))
        {
            const realdp *xM = x->ObtainReadData();
            
            const SharedSpace *Temp = x->ObtainReadTempData("CLplusX");
            const realdp *LplusX = Temp->ObtainReadData();
            
            SharedSpace *iLplusX = new SharedSpace(3, n, n, num);
            realdp *iLplusXptr = iLplusX->ObtainWriteEntireData();
            
            integer N = n, length = N * N, numlength = N * N * num, info;
            
            realdp *I = new realdp[n * n]; // make identity matrix
            for (integer i = 0; i < length; i++) I[i] = 0.0;
            for (integer i = 0; i < n; i++) I[i * n + i] = 1.0;
            
            realdp *LLplusX = new realdp[n * n * num];
            realdp *tmp = new realdp[n * n];
            
            copy_(&numlength, const_cast<realdp *>(LplusX), &GLOBAL::IONE, LLplusX, &GLOBAL::IONE);
            
            
            for (integer i = 0; i < num; i++)
            {
                // Compute inv(LplusX) -> iLplusX
                copy_(&length, I, &GLOBAL::IONE, iLplusXptr + n * n * i, &GLOBAL::IONE);
                trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, LLplusX + n * n * i, &N, iLplusXptr + n * n * i, &N, &info);
                
                // compute inv(LplusX) = inv(LplusX)^T * inv(LplusX)
                copy_(&length, iLplusXptr + n * n * i, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
                gemm_(GLOBAL::T, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, tmp, &N, tmp, &N, &GLOBAL::DZERO, iLplusXptr + n * n * i, &N);
            }
            x->AddToTempData("iLplusX", iLplusX);
            
            delete[] I;
            delete[] LLplusX;
            delete[] tmp;
        }
        
        integer N = n, length = N * N, numlength = num * length;
        
        const SharedSpace *SharediLplusX = x->ObtainReadTempData("iLplusX");
        const realdp *iLplusX = SharediLplusX->ObtainReadData();
        const realdp *xM = x->ObtainReadData();
        const realdp *etaxTV = etax->ObtainReadData();
        realdp *xixTV = xix->ObtainWriteEntireData();
        
        if (xixTV == etaxTV)
        {
            std::cout << "Error in RieHessianEta!" << std::endl;
            exit(0);
        }
        
        realdp *tmp1 = new realdp[n * n];
        realdp *tmp2 = new realdp[n * n];
        realdp *tmp3 = new realdp[n * n];
        realdp *rightX = new realdp[n * n];
        realdp *leftX = new realdp[n * n];
        
        
        for (integer i = 0; i < n * n; i++)
            xixTV[i] = 0.0;
        
        for (integer i = 0; i < num; i++)
        {
            // iL * X -> rightX
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *>(iLplusX + n * n * i), &N, const_cast<realdp *>(xM), &N, &GLOBAL::DZERO, rightX, &N);
            
            // X * iL -> leftX
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *>(xM), &N, const_cast<realdp *>(iLplusX + n * n * i), &N, &GLOBAL::DZERO, leftX, &N);
            
            // xi * iL * X -> tmp1
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *>(etaxTV), &N, rightX, &N, &GLOBAL::DZERO, tmp1, &N);
            
            // X * iL * xi -> tmp2
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, leftX, &N, const_cast<realdp *>(etaxTV), &N, &GLOBAL::DZERO, tmp2, &N);
            
            // tmp3
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, tmp2, &N, rightX, &N, &GLOBAL::DZERO, tmp3, &N);
            
            for(integer j = 0; j < n * n; j++) xixTV[j] = xixTV[j] + tmp1[j] + tmp2[j] - (static_cast<realdp> (1) + alpha)*tmp3[j];
            
        }
        
        
        realdp scalar = static_cast<realdp> (1.0/(1.0 - alpha));
        scal_(&length, &scalar, xixTV, &GLOBAL::IONE);
        
        delete[] tmp1;
        delete[] tmp2;
        delete[] tmp3;
        delete[] rightX;
        delete[] leftX;
    };
  
}; /*end of ROPTLIB namespace*/
