
#include "SPDMeanLDOneParamLinfty.h"

/*Define the namespace*/
namespace ROPTLIB{
    
    SPDMeanLDOneParamLinfty::SPDMeanLDOneParamLinfty(realdp *inLs, realdp *inLDAs, integer inn, integer innum, realdp inalpha)
    {
        Ls = inLs;
        LDAs = inLDAs;
        n = inn;
        num = innum;
        alpha = inalpha;
    };
    
    SPDMeanLDOneParamLinfty::~SPDMeanLDOneParamLinfty(void)
    {
    };
    
    realdp SPDMeanLDOneParamLinfty::f(Variable *x) const
    {
        
        SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
        if (!x->TempDataExist("L"))
        {
            Mani->CholeskyRepresentation(x);
        }
        const SharedSpace *SharedL = x->ObtainReadTempData("L");
        Variable *LElement = SharedL->GetSharedElement();
        const realdp *L = LElement->ObtainReadData();
        
        const realdp *xxM = x->ObtainReadData();
        
        integer N = n, length = n * n, Nnum = n * n * num, info;
        realdp onehalf = 0.5;
        realdp a1 = static_cast<realdp> ((1.0 - alpha)/2.0), a2 = static_cast<realdp> ((1.0 + alpha)/2.0);
        
        realdp *I = new realdp[n * n]; // make identity matrix
        for (integer i = 0; i < length; i++) I[i] = 0.0;
        for (integer i = 0; i < n; i++) I[i * n + i] = 1.0;
        
        
        // CLplusX is used to store a1 * Ai + a2 * x
        SharedSpace *SharedCLplusX = new SharedSpace(3, n, n, num);
        realdp *CLplusX = SharedCLplusX->ObtainWriteEntireData();
        
        realdp *coef1 = new realdp[num];
        
        copy_(&Nnum, Ls, &GLOBAL::IONE, CLplusX, &GLOBAL::IONE);
        
        for (integer i = 0; i < num; i++)
        {
            // compute a1 * Ai + a2 * x
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &a2, const_cast<realdp *>(xxM), &N, I, &N, &a1, CLplusX + n * n * i, &N);
            
            // Cholesky factorization of CLplusX
            potrf_(GLOBAL::L, &N, CLplusX + n * n * i, &N, &info);
            for (integer j = 0; j < n; j++)
                for (integer k = j + 1; k < n; k++)
                    CLplusX[j + k * n + i * n * n] = 0;
            
            // compute the determinant
            coef1[i] = 1.0;
            for (integer j = 0; j < n; j++)
            {
                coef1[i] = coef1[i] * CLplusX[j * n + j + i * n * n];
            }
            coef1[i] = log(coef1[i] * coef1[i]);
        }
        x->AddToTempData("CLplusX", SharedCLplusX);
        
        
        // compute the determinant of X: product of diagonal elements
        realdp detX = 1.0;
        for (integer j = 0; j < n; j++)
        {
            detX = detX * const_cast<realdp *>(L)[j * n + j];
        }
        detX = log(detX * detX);
        
        
        SharedSpace *Sharedmaxval = new SharedSpace(1, 2);
        realdp *maxval = Sharedmaxval->ObtainWriteEntireData();
        
        
        for (integer i = 0; i < num; i++)
        {
            coef1[i] = coef1[i] - a1 * LDAs[i] - a2 * detX;
        }
        
        
        realdp result = coef1[0];
        maxval[1] = 0;
        for (integer i = 1; i < num; i++)
        {
            if(coef1[i] > result)
            {
                result = coef1[i];
                maxval[1] = static_cast<realdp>(i);
            }
        }
        maxval[0] = result;
        
        result = static_cast<realdp> ((result * 4.0)/(1 - alpha * alpha));
        x->AddToTempData("maxval", Sharedmaxval);
        
        
        delete[] I;
        delete[] coef1;
        
        return result;
        
    };
    
    
    
    void SPDMeanLDOneParamLinfty::RieGrad(Variable *x, Vector *gf) const
    {
        const SharedSpace *Sharedmaxval = x->ObtainReadTempData("maxval");
        const realdp *maxval  = Sharedmaxval->ObtainReadData();
        
        const SharedSpace *SharedCLplusX = x->ObtainReadTempData("CLplusX");
        const realdp *CLplusX = SharedCLplusX->ObtainReadData();
        
        const realdp *xM = x->ObtainReadData();
        realdp *gfVT = gf->ObtainWriteEntireData();
        
        for (integer i = 0; i < n * n; i++) gfVT[i] = 0.0;
        
        realdp *tmp = new realdp[n * n];
        realdp *iLplusX = new realdp[n * n];
        
        integer length = n * n, N = n, Nnum = n * n * num, info;
        realdp negone = -1.0;
        
        realdp *I = new realdp[n * n]; // make identity matrix
        for (integer i = 0; i < length; i++) I[i] = 0.0;
        for (integer i = 0; i < n; i++) I[i * n + i] = 1.0;
        
        
        integer idx = static_cast<integer>(maxval[1]);
         
        copy_(&length, const_cast<realdp *>(CLplusX + n * n * idx), &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        
        // Compute inv(LplusX) -> iLplusX
        copy_(&length, I, &GLOBAL::IONE, iLplusX, &GLOBAL::IONE);
        trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, tmp, &N, iLplusX, &N, &info);
        
        // compute inv(LplusX) = inv(LplusX)^T * inv(LplusX) -> gfVT
        gemm_(GLOBAL::T, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, iLplusX, &N, iLplusX, &N, &GLOBAL::DZERO, gfVT, &N);
        
        // compute X * gfVT -> tmp
        gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *>(xM), &N, gfVT, &N, &GLOBAL::DZERO, tmp, &N);
        
        // compute tmp * X -> gfVT
        gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, tmp, &N, const_cast<realdp *>(xM), &N, &GLOBAL::DZERO, gfVT, &N);
        
        // compute gfVT - X -> tmp
        copy_(&length, gfVT, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        copy_(&length, const_cast<realdp *>(xM), &GLOBAL::IONE, gfVT, &GLOBAL::IONE);
        gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, tmp, &N, I, &N, &negone, gfVT, &N);
        
        realdp scalar = static_cast<realdp> (2.0/(1.0 - alpha));
        scal_(&length, &scalar, gfVT, &GLOBAL::IONE);
        
        
        delete[] I;
        delete[] tmp;
        delete[] iLplusX;
        
        
    };
    
    
    
    void SPDMeanLDOneParamLinfty::RieHessianEta(Variable *x, Vector *etax, Vector *xix) const
    {
        std::cout << "warning: SPDMeanLDOneParam::RieHessianEta has not been implemented!" << std::endl;
        etax->CopyTo(xix);
    };
}; /*end of ROPTLIB namespace*/
