
#include "SPDMeanLinfty.h"

/*Define the namespace*/
namespace ROPTLIB{

	SPDMeanLinfty::SPDMeanLinfty(realdp *inLs, integer inn, integer innum)
	{
		Ls = inLs;
		n = inn;
		num = innum;
	};

	SPDMeanLinfty::~SPDMeanLinfty(void)
	{
	};

	realdp SPDMeanLinfty::f(Variable *x) const
	{
		SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
		if (!x->TempDataExist("L"))
		{
			Mani->CholeskyRepresentation(x);
		}
		const SharedSpace *SharedL = x->ObtainReadTempData("L");
		Variable *LElement = SharedL->GetSharedElement();
		const realdp *L = LElement->ObtainReadData();

		SharedSpace *SharedlogLXL = new SharedSpace(3, n, n, num);
		realdp *logLXL = SharedlogLXL->ObtainWriteEntireData();
        
		realdp *Ltmp = new realdp[n * n];
		integer length = n * n, N = n, info;
        
        
        realdp *coef = new realdp[num];
        
        SharedSpace *Sharedmaxval = new SharedSpace(1, 2); 
        realdp *maxval = Sharedmaxval->ObtainWriteEntireData();

        
        
		for (integer i = 0; i < num; i++)
		{
			copy_(&length, const_cast<realdp *> (L), &GLOBAL::IONE, Ltmp, &GLOBAL::IONE);
			/*Solve the linear system Li X = Lx, i.e., X = Li^{-1} Lx. The solution X is stored in LiiLx.
			Note that Li is a lower triangular matrix.
			Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
			trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Ls + n * n * i, &N, Ltmp, &N, &info);
			if (info != 0)
			{
				printf("The cholesky decompsotion in SPDMean::f failed with info:%d!\n", info);
			}
			gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, Ltmp, &N, Ltmp, &N, &GLOBAL::DZERO, logLXL + n * n * i, &N);
			Matrix MMt(logLXL + n * n * i, n, n);
			Matrix::LogSymmetricM(GLOBAL::L, MMt, MMt);
            
            coef[i] = nrm2_(&length, logLXL + n * n * i, &GLOBAL::IONE);
		}
        
        
        realdp result = coef[0];
        maxval[1] = 0;
        for (integer i = 1; i < num; i++)
        {
            if(coef[i] > result)
            {
                result = coef[i];
                maxval[1] = static_cast<realdp>(i);
//                 printf("I am the true index: %d, I am not: %d\n", i, static_cast<integer>(maxval[1]));
            }
        }
        
        
        
        maxval[0] = result;
//         printf("I am the true result: %.2e, I am not: %.2e\n", result, maxval[0]);

        delete[] Ltmp;
        delete[] coef;
		x->AddToTempData("logLXL", SharedlogLXL);
        x->AddToTempData("maxval", Sharedmaxval);

		return result;
  
// -------------------------------------------------------------------------
//         SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
// 		if (!x->TempDataExist("L"))
// 		{
// 			Mani->CholeskyRepresentation(x);
// 		}
// 		const SharedSpace *SharedL = x->ObtainReadTempData("L");
// 		Variable *LElement = SharedL->GetSharedElement();
// 		const realdp *L = LElement->ObtainReadData();
// 
// 		SharedSpace *SharedlogLXL = new SharedSpace(3, n, n, num);
// 		realdp *logLXL = SharedlogLXL->ObtainWriteEntireData();
// 		realdp *Ltmp = new realdp[n * n];
// 		integer length = n * n, N = n, info;
//         
//         
//         realdp *eigenvalues = new realdp[n * num];
//         realdp *eigenvectors = new realdp[2 * n * n];
//         realdp *eigenvectorsD = eigenvectors + n * n;
//         
//         realdp a, result;
//         SharedSpace *Sharedcoef = new SharedSpace(3, num, 1, 1); 
//         realdp *coef = Sharedcoef->ObtainWriteEntireData();
//         
//         
// 		for (integer i = 0; i < num; i++)
// 		{
// 			copy_(&length, const_cast<realdp *> (L), &GLOBAL::IONE, Ltmp, &GLOBAL::IONE);
// 			/*Solve the linear system Li X = Lx, i.e., X = Li^{-1} Lx. The solution X is stored in LiiLx.
// 			Note that Li is a lower triangular matrix.
// 			Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
// 			trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Ls + n * n * i, &N, Ltmp, &N, &info);
// 			if (info != 0)
// 			{
// 				std::cout << "The cholesky decompsotion in SPDMean::f failed with info:" << info << "!" << std::endl;
// 			}
// 			gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, Ltmp, &N, Ltmp, &N, &GLOBAL::DZERO, logLXL + n * n * i, &N);
//             
//              Matrix E(eigenvalues + n * i, n, 1), V(eigenvectors, n, n), VD(eigenvectorsD, n, n);
//              Matrix MMt(logLXL + n * n * i, n, n);
//              Matrix::EigenSymmetricM(GLOBAL::L, MMt, E, V);
// 
//              copy_(&length, eigenvectors, &GLOBAL::IONE, VD.matrix, &GLOBAL::IONE);
//              for (integer j = 0; j < n; j++)
//              {
//                  a = log(eigenvalues[j + n * i]);
//                  scal_(&N, &a, eigenvectors + j * n, &GLOBAL::IONE);
//              }
//              
//              Matrix MlogLXL(logLXL + n * n * i, n, n);
//              Matrix::DGEMM(GLOBAL::DONE, V, false, VD, true, GLOBAL::DZERO, MlogLXL);
//              
//              coef[i] = 0.0;
//              for(integer j = 0; j < N; j++) {coef[i] = coef[i] + log(eigenvalues[j + i * N]) * log(eigenvalues[j + i * N]);}
//              coef[i] = std::sqrt(coef[i]);
// 		}
//         
//         result = coef[0];
//         for (integer i = 1; i < num; i++)
//         {
//             if (coef[i] > result)
//             {
//                 result = coef[i];
//             }
//         }
//         
// 
// 		delete[] Ltmp;
//         delete[] eigenvalues;
//         delete[] eigenvectors;
// 		x->AddToTempData("logLXL", SharedlogLXL);
//         x->AddToTempData("coef", Sharedcoef);        
// 
// 		return result;
	};
    
    

	void SPDMeanLinfty::RieGrad(Variable *x, Vector *gf) const
	{
		const SharedSpace *SharedlogLXL = x->ObtainReadTempData("logLXL");
		const realdp *logLXL = SharedlogLXL->ObtainReadData();
		realdp *gfVT = gf->ObtainWriteEntireData();
		for (integer i = 0; i < n * n; i++)
			gfVT[i] = 0;
		const realdp *xM = x->ObtainReadData();

		integer N = n, info;

        const SharedSpace *Sharedmaxval = x->ObtainReadTempData("maxval");
        const realdp *maxval  = Sharedmaxval->ObtainReadData();
        
        realdp *tmp = new realdp[n * n];

        integer idx = static_cast<integer>(maxval[1]);
        
        gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *> (logLXL + n * n * idx), &N,
                Ls + n * n * idx, &N, &GLOBAL::DZERO, tmp, &N);
        
        /*Solve the linear system Li^T X = tmp, i.e., X = Li^{-T} log(Li^{-1} X Li^{-T}) Li^T. The solution X is stored in tmp.
         * Note that Li is a lower triangular matrix.
         * Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
        trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, Ls + n * n * idx, &N, tmp, &N, &info);
        if (info != 0)
        {
            printf("The cholesky decompsotion in SPDMean::RieGrad failed with info:%d!\n", info);
        }
        gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *> (xM), &N, tmp, &N, &GLOBAL::DONE, gfVT, &N);

        
        realdp coefa = static_cast<realdp> (1)/maxval[0];
        integer length = n * n;
        scal_(&length, &coefa, gfVT, &GLOBAL::IONE);
        delete[] tmp;
    };
    
    

	void SPDMeanLinfty::RieHessianEta(Variable *x, Vector *etax, Vector *xix) const
	{
		std::cout << "warning: SPDMean::RieHessianEta has not been implemented!" << std::endl;
		etax->CopyTo(xix);
	};
}; /*end of ROPTLIB namespace*/
