/*
This file defines the class for the problem: 
Computing the karcher mean of SPD manifold, i.e.,
min_{X \in SPD} sum dist^2(X, A_i)
where A_i \in SPD and dist(A, B) = \|logm(A^{-1/2} B A^{-1/2})\|_F.

Problem --> SPDMean

---- WH
*/

#ifndef SPDMEANSYMMLDONEPARAML1_H
#define SPDMEANSYMMLDONEPARAML1_H

#include "Manifolds/SPDManifold/SPDManifold.h"
#include "Manifolds/SPDManifold/SPDVariable.h"
#include "Manifolds/SPDManifold/SPDVector.h"
#include "Problems/Problem.h"
#include "Manifolds/SharedSpace.h"
#include "Others/def.h"
#include "Others/MyMatrix.h"

/*Define the namespace*/
namespace ROPTLIB{

	class SPDMeanSymmLDOneParamL1 : public Problem{
	public:
		/*The sample SPD matrices Ai are stored as their Cholesky matrix, i.e., Ai = Li Li^T.
		inLs is a n by n by num array of realdp numbers.*/
		SPDMeanSymmLDOneParamL1(realdp *inLs, realdp *inLDAs, integer inn, integer innum, realdp inalpha);
		virtual ~SPDMeanSymmLDOneParamL1();
		virtual realdp f(Variable *x) const;

		virtual void RieGrad(Variable *x, Vector *gf) const;

		/*Riemannian action of the Hessian has not been done yet*/
		virtual void RieHessianEta(Variable *x, Vector *etax, Vector *xix) const;
        
//         virtual void RieHessianEig(Variable *x, realdp *EigVal) const;
        
//         mutable realdp *xinru;

		/*The Euclidean gradient and Hes sian are not done.*/
		//virtual void EucGrad(Variable *x, Vector *egf) const;
		//virtual void EucHessianEta(Variable *x, Vector *etax, Vector *exix) const;

		realdp *Ls;
        realdp *LDAs;
		integer n;
		integer num;
        realdp alpha;
	};
}; /*end of ROPTLIB namespace*/
#endif // end of STIEBROCKETT_H
