
#include "Problems/StieBrockett/StieBrockett.h"

/*Define the namespace*/
namespace ROPTLIB{

	StieBrockett::StieBrockett(realdp *inB, realdp *inD, integer inn, integer inp)
	{
		B = inB;
		D = inD;
		n = inn;
		p = inp;
	};

	StieBrockett::~StieBrockett(void)
	{
	};

	realdp StieBrockett::f(Variable *x) const
	{
		const realdp *xxM = x->ObtainReadData();
		Vector *BxD = x->ConstructEmpty();
		SharedSpace *Temp = new SharedSpace(BxD);
		realdp *temp = BxD->ObtainWriteEntireData();
		realdp result = 0;

		Matrix MB(B, n, n), MxxM(xxM, n, p), Mtemp(temp, n, p);
		// temp = B * xxM, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		Matrix::DGEMM(1, MB, false, MxxM, false, 0, Mtemp);

		for (integer i = 0; i < p; i++)
		{
			// temp(i * n : i * n + N - 1) <- D[i] * temp(i * n : i * n + N - 1),
			// details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
			scal_(const_cast<integer *> (&n), &D[i], temp + i * n, &GLOBAL::IONE);
		}
		integer length = n * p;
		// output temp(:)^T * xxM(:), details: http://www.netlib.org/lapack/explore-html/d5/df6/dot_8f.html
		result = dot_(&length, temp, &GLOBAL::IONE, const_cast<realdp *> (xxM), &GLOBAL::IONE);
		if (UseGrad)
		{
			x->AddToTempData("BxD", Temp);
		}
		else
		{
			delete Temp;
		}
		return result;
	};

	//void StieBrockett::RieGrad(Variable *x, Vector *gf) const
	//{
	//	const realdp *xxM = x->ObtainReadData();
	//	const SharedSpace *SharedTemp = x->ObtainReadTempData("BxD");
	//	Vector *Temp = SharedTemp->GetSharedElement();
	//	const realdp *xBxD = Temp->ObtainReadData();
	//	realdp *gfTV = gf->ObtainWriteEntireData();
	//	SharedSpace *symxtBxN = new SharedSpace(2, p, p);
	//	realdp *symxtBxNptr = symxtBxN->ObtainWriteEntireData();
	//
	//	char *transn = "n", *transt = "t";
	//	realdp one = 1, zero = 0;
	//	integer inc = 1, N = n, P = p, Length = N * P;
	//	gemm_(transt, transn, &P, &P, &N, &one, const_cast<realdp *> (xxM), &N, const_cast<realdp *> (xBxD), &N, &zero, symxtBxNptr, &P);
	//	for (integer i = 0; i < p; i++)
	//	{
	//		for (integer j = i + 1; j < p; j++)
	//		{
	//			symxtBxNptr[i + j * p] += symxtBxNptr[j + i * p];
	//			symxtBxNptr[i + j * p] /= 2.0;
	//			symxtBxNptr[j + i * p] = symxtBxNptr[i + j * p];
	//		}
	//	}
	//	copy_(&Length, const_cast<realdp *> (xBxD), &inc, gfTV, &inc);
	//	realdp negone = -1;
	//	gemm_(transn, transn, &N, &P, &P, &negone, const_cast<realdp *> (xxM), &N, symxtBxNptr, &P, &one, gfTV, &N);
	//	realdp two = 2;
	//	scal_(&Length, &two, gfTV, &inc);
	//	if (UseHess)
	//	{
	//		x->AddToTempData("symxtBxD", symxtBxN);
	//	}
	//	else
	//	{
	//		delete symxtBxN;
	//	}
	//};
	//
	//void StieBrockett::RieHessianEta(Variable *x, Vector *etax, Vector *xix) const
	//{
	//	if (!x->TempDataExist("symxtBxD"))
	//	{
	//		const realdp *xxM = x->ObtainReadData();
	//		const SharedSpace *Temp = x->ObtainReadTempData("BxD");
	//		const realdp *xBxD = Temp->ObtainReadData();
	//		SharedSpace *symxtBxN = new SharedSpace(2, p, p);
	//		realdp *symxtBxNptr = symxtBxN->ObtainWriteEntireData();
	//
	//		char *transn = "n", *transt = "t";
	//		realdp one = 1, zero = 0;
	//		integer inc = 1, N = n, P = p, Length = N * P;
	//		gemm_(transt, transn, &P, &P, &N, &one, const_cast<realdp *> (xxM), &N, const_cast<realdp *> (xBxD), &N, &zero, symxtBxNptr, &P);
	//		for (integer i = 0; i < p; i++)
	//		{
	//			for (integer j = i + 1; j < p; j++)
	//			{
	//				symxtBxNptr[i + j * p] += symxtBxNptr[j + i * p];
	//				symxtBxNptr[i + j * p] /= 2.0;
	//				symxtBxNptr[j + i * p] = symxtBxNptr[i + j * p];
	//			}
	//		}
	//		x->AddToTempData("symxtBxD", symxtBxN);
	//	}
	//
	//	const SharedSpace *symxtBxD = x->ObtainReadTempData("symxtBxD");
	//	const realdp *DsymxtBxD = symxtBxD->ObtainReadData();
	//	const realdp *xM = x->ObtainReadData();
	//	const realdp *etaxTV = etax->ObtainReadData();
	//	realdp *xixTV = xix->ObtainWriteEntireData();
	//
	//	if (xixTV == etaxTV)
	//	{
	//		printf("Error in RieHessianEta!\n");
	//		exit(0);
	//	}
	//	char *transn = "n";
	//	integer N = n, P = p, inc = 1, Length = N * P;
	//	realdp one = 1, zero = 0, negone = -1, two = 2;
	//	gemm_(transn, transn, &N, &P, &N, &one, B, &N, const_cast<realdp *> (etaxTV), &N, &zero, xixTV, &N);
	//	for (integer i = 0; i < p; i++)
	//	{
	//		scal_(&N, &D[i], xixTV + i * n, &inc);
	//	}
	//
	//	gemm_(transn, transn, &N, &P, &P, &negone, const_cast<realdp *> (etaxTV), &N, const_cast<realdp *> (DsymxtBxD), &P, &one, xixTV, &N);
	//	scal_(&Length, &two, xixTV, &inc);
	//	Stiefel *StieDomain = dynamic_cast<Stiefel *> (Domain);
	//	StieDomain->ExtrProjection(x, xix, xix);
	//};

	void StieBrockett::EucGrad(Variable *x, Vector *egf) const
	{
		const SharedSpace *Temp = x->ObtainReadTempData("BxD");
		Vector *BxD = Temp->GetSharedElement();
		Domain->ScaleTimesVector(x, 2.0, BxD, egf);
	};

	void StieBrockett::EucHessianEta(Variable *x, Vector *etax, Vector *exix) const
	{
		const realdp *etaxTV = etax->ObtainReadData();
		realdp *exixTV = exix->ObtainWriteEntireData();

		char *transn = const_cast<char *> ("n");
		integer N = n, P = p, inc = 1, Length = N * P;
		realdp one = 1, zero = 0, two = 2;
		// exxiTV <- B * etaxTV, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transn, transn, &N, &P, &N, &one, B, &N, const_cast<realdp *> (etaxTV), &N, &zero, exixTV, &N);
		for (integer i = 0; i < p; i++)
		{
			// exixTV(i * n : i * n + N - 1) <- D[i] * exixTV(i * n : i * n + N - 1),
			// details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
			scal_(&N, &D[i], exixTV + i * n, &inc);
		}
		Domain->ScaleTimesVector(x, 2.0, exix, exix);
	};
}; /*end of ROPTLIB namespace*/
