
#include "Problems/StieSPCA/StieSPCA.h"

/*Define the namespace*/
namespace ROPTLIB{

	StieSPCA::StieSPCA(realdp *inB, realdp inlambda, integer inn, integer inm, integer inp)
	{
		B = inB;
		n = inn;
        m = inm;
		p = inp;
        lambda = inlambda;
        colnormB = new realdp[n];
        for(integer i = 0; i < n; i++)
        {
            colnormB[i] = dot_(&m, B + i * m, &GLOBAL::IONE, B + i * m, &GLOBAL::IONE);
        }
        
        
        realdp *U = new realdp[m * m + m + m * n * 2];
        realdp *S = U + m * m;
        realdp *Vt = S + m;
        realdp *Bc = Vt + m * n;
        /*B is an n by n matrix*/
        for (integer i = 0; i < n * m; i++)
        {
            Bc[i] = B[i];
        }
        Matrix MBc(Bc, m, n), MU(U, m, m), MS(S, m, 1), MVt(Vt, m, n);
        Matrix::SVDDecom(MBc, MU, MS, MVt);
        L = S[0] * S[0] * 2;
        delete [] U;
	};

	StieSPCA::~StieSPCA(void)
	{
        delete [] colnormB;
	};

	realdp StieSPCA::f(Variable *x) const
	{
		const realdp *xxM = x->ObtainReadData();
		SharedSpace *Temp = new SharedSpace(2, m, p);
		realdp *temp = Temp->ObtainWriteEntireData();
		realdp result = 0;

		Matrix MB(B, m, n), MxxM(xxM, n, p), Mtemp(temp, m, p);
		// temp = B * xxM, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		Matrix::DGEMM(1, MB, false, MxxM, false, 0, Mtemp);

		integer length = m * p;
		// output temp(:)^T * xxM(:), details: http://www.netlib.org/lapack/explore-html/d5/df6/dot_8f.html
		result = dot_(&length, temp, &GLOBAL::IONE, temp, &GLOBAL::IONE);
		if (UseGrad)
		{
			x->AddToTempData("Bx", Temp);
		}
		else
		{
			delete Temp;
		}
        result = -result;
        for(integer i = 0; i < n * p; i++)
            result += lambda * std::fabs(xxM[i]);
		return result;
	};

	void StieSPCA::EucGrad(Variable *x, Vector *egf) const
	{
		const SharedSpace *Temp = x->ObtainReadTempData("Bx");
        const realdp *Bxptr = Temp->ObtainReadData();
        realdp *egfptr = egf->ObtainWriteEntireData();
        Matrix MB(B, m, n), MBx(Bxptr, m, p), Megf(egfptr, n, p);
        Matrix::DGEMM(1, MB, true, MBx, false, 0, Megf);
		Domain->ScaleTimesVector(x, -2.0, egf, egf);
	};

	void StieSPCA::EucHessianEta(Variable *x, Vector *etax, Vector *exix) const
	{
		const realdp *etaxTV = etax->ObtainReadData();
		realdp *exixTV = exix->ObtainWriteEntireData();
        realdp *tmp = new realdp[m * p];
		integer N = n, P = p, M = m;
		// exxiTV <- B * etaxTV, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(GLOBAL::N, GLOBAL::N, &M, &P, &N, &GLOBAL::DONE, B, &M, const_cast<realdp *> (etaxTV), &N, &GLOBAL::DZERO, tmp, &M);
        gemm_(GLOBAL::T, GLOBAL::N, &N, &P, &M, &GLOBAL::DONE, B, &M, tmp, &M, &GLOBAL::DZERO, exixTV, &N);
		Domain->ScaleTimesVector(x, -2.0, exix, exix);
        delete [] tmp;
	};

    void StieSPCA::PreConditioner(Variable *x, Vector *eta, Vector *result) const
    {
        if(result->Getlength() == x->Getlength())
        {
            const SharedSpace *Temp = x->ObtainReadTempData("Bx");
            const realdp *Bxptr = Temp->ObtainReadData();
            realdp *w1 = new realdp[p];
            for (integer i = 0; i < p; i++)
            {
                w1[i] = dot_(&m, const_cast<realdp *>(Bxptr + i * m), &GLOBAL::IONE, const_cast<realdp *>(Bxptr + i * m), &GLOBAL::IONE);
            }
            realdp *resultptr = result->ObtainWriteEntireData();
            for(integer i = 0; i < n; i++)
                for(integer j = 0; j < p; j++)
                    resultptr[i + j * n] = ((w1[j] - colnormB[i]) * 2.0 > 0.1) ? (w1[j] - colnormB[i]) * 2.0 : 0.1;
            delete [] w1;
        } else
        if(result->Getlength() == p)
        {
            const SharedSpace *Temp = x->ObtainReadTempData("Bx");
            const realdp *Bxptr = Temp->ObtainReadData();
            realdp *resultptr = result->ObtainWriteEntireData();
            for (integer i = 0; i < p; i++)
            {
                resultptr[i] = dot_(&m, const_cast<realdp *>(Bxptr + i * m), &GLOBAL::IONE, const_cast<realdp *>(Bxptr + i * m), &GLOBAL::IONE);
                resultptr[i] = ((resultptr[i] - 1.0) * 2.0 > 0.1) ? (resultptr[i] - 1.0) * 2.0 : 0.1;
            }
        } else
        if(result->Getlength() == 1)
        {
            realdp *resultptr = result->ObtainWriteEntireData();
            resultptr[0] = L;
        } else
        {
            printf("Warning: StieSPCA::PreConditioner: the size of weighting matrix is not supported.\n");
            result->SetToZeros();
        }
    };

    void StieSPCA::ProxW(Variable *x, Vector *Weight, Vector *result) const
    { //output = min(0, X + mu ./ W) + max(0, X - mu ./ W)
        integer length = x->Getlength();
        realdp *resultptr = result->ObtainWriteEntireData();
        const realdp *xptr = x->ObtainReadData();
        const realdp *Wptr = Weight->ObtainReadData();
        
        
        integer nblock = Weight->Getlength();
        integer blocksize = x->Getlength() / nblock;
        integer idx = 0;
        
        for(integer i = 0; i < nblock; i++)
        {
            for(integer j = 0; j < blocksize; j++)
            {
                idx = j + i * blocksize;
                resultptr[idx] = ((xptr[idx] + lambda / Wptr[i] < 0.0) ? xptr[idx] + lambda / Wptr[i] : 0.0) + ((xptr[idx] - lambda / Wptr[i] > 0.0) ? xptr[idx] - lambda / Wptr[i] : 0.0);
            }
        }
        
//        if(Weight->Getlength() != 1)
//        {
//            for(integer i = 0; i < length; i++)
//            {
//                resultptr[i] = ((xptr[i] + lambda / Wptr[i] < 0.0) ? xptr[i] + lambda / Wptr[i] : 0.0) + ((xptr[i] - lambda / Wptr[i] > 0.0) ? xptr[i] - lambda / Wptr[i] : 0.0);
//            }
//        } else
//        {
//
//            for(integer i = 0; i < length; i++)
//            {
//                resultptr[i] = ((xptr[i] + lambda / Wptr[0] < 0.0) ? xptr[i] + lambda / Wptr[0] : 0.0) + ((xptr[i] - lambda / Wptr[0] > 0.0) ? xptr[i] - lambda / Wptr[0] : 0.0);
//            }
//        }
    };

    void StieSPCA::CalJW(Variable *x, Vector *eta, Vector *Weight, Vector *result) const
    { // output = (abs(x) .* W > lambda) .* eta;
        integer length = x->Getlength();
        const realdp *xptr = x->ObtainReadData();
        const realdp *etaptr = eta->ObtainReadData();
        realdp *resultptr = result->ObtainWriteEntireData();
        const realdp *Wptr = Weight->ObtainReadData();
        
        integer nblock = Weight->Getlength();
        integer blocksize = x->Getlength() / nblock;
        integer idx = 0;
        
        for(integer i = 0; i < nblock; i++)
        {
            for(integer j = 0; j < blocksize; j++)
            {
                idx = j + i * blocksize;
                resultptr[idx] = (std::abs(xptr[idx]) * Wptr[i] > lambda) ? etaptr[idx] : 0;
            }
        }
        
//        if(Weight->Getlength() != 1)
//        {
//            for(integer i = 0; i < length; i++)
//            {
//                resultptr[i] = (std::abs(xptr[i]) * Wptr[i] > lambda) ? etaptr[i] : 0;
//            }
//        } else
//        {
//            for(integer i = 0; i < length; i++)
//            {
//                resultptr[i] = (std::abs(xptr[i]) * Wptr[0] > lambda) ? etaptr[i] : 0;
//            }
//        }
//        std::cout << "L:" << L << std::endl;//---
    };
}; /*end of ROPTLIB namespace*/
