/*
This file defines the class for a sparse PCA model on the Stiefel manifold
min_X - tr(X^T B^T B X) + \lambda \|X\|_1, where B is a m-by-n matrix, and X \in St(p, n).
Typically, n > m > p.

Problem --> SPCA

---- WH
*/

#ifndef STIESPCA_H
#define STIESPCA_H

#include "Manifolds/Stiefel/Stiefel.h"
#include "Manifolds/Stiefel/StieVariable.h"
#include "Manifolds/Stiefel/StieVector.h"
#include "Problems/Problem.h"
#include "Manifolds/SharedSpace.h"
#include "Solvers/RPG.h"
#include "Others/def.h"
#include "Others/MyMatrix.h"

/*Define the namespace*/
namespace ROPTLIB{

	class StieSPCA : public Problem{
	public:
		StieSPCA(realdp *inB, realdp inlambda, integer inn, integer inm, integer inp);
		virtual ~StieSPCA();
		virtual realdp f(Variable *x) const;

		virtual void EucGrad(Variable *x, Vector *egf) const;
		virtual void EucHessianEta(Variable *x, Vector *etax, Vector *exix) const;
        virtual void ProxW(Variable *x, Vector *Weight, Vector *result) const;
        virtual void CalJW(Variable *x, Vector *eta, Vector *Weight, Vector *result) const;
        
        /*The create the weight matrix*/
        virtual void PreConditioner(Variable *x, Vector *eta, Vector *result) const;

		realdp *B;
        realdp lambda;
		mutable integer n;
        mutable integer m;
		mutable integer p;
        realdp *colnormB;
        realdp L;
	};
}; /*end of ROPTLIB namespace*/
#endif // end of GRASSRQ_H
