/*
This file defines the proximal gradient method with acceleration on manifold with or without preconditioner.


Solvers --> QuasiNewton --> SolversLS --> RPG --> ARPG

---- WH
*/

#ifndef ARPG_H
#define ARPG_H

#include "Solvers/RPG.h"
#include "Others/def.h"

/*Define the namespace*/
namespace ROPTLIB{

	class ARPG : public RPG{
	public:
		/*The contructor of ARPG method. It calls the function Solvers::Initialization.
		INPUT : prob is the problem which defines the cost function, gradient and possible the action of Hessian
		and specifies the manifold of domain.
		initialx is the initial iterate.
		insoln is the true solution. It is not required and only used for research.*/
		ARPG(const Problem *prob, const Variable *initialx, const Variable *insoln = nullptr);

        ~ARPG();
        
		/*Call Solvers::SetProbX function and indicate RCG does not need action of Hessian.
		INPUT:	prob is the problem which defines the cost function, gradient and possible the action of Hessian
		and specifies the manifold of domain.
		initialx is the initial iterate.
		insoln is the true solution. It is not required and only used for research.*/
		virtual void SetProbX(const Problem *prob, const Variable *initialx, const Variable *insoln);

		/*Setting parameters (member variables) to be default values */
		virtual void SetDefaultParams();
        
        /*Check whether the parameters in RPG are legal or not.*/
        virtual void CheckParams(void);
        
        /*Print general information, which is not specific to an algorithm.*/
        virtual void PrintGenInfo(void);
        
        /*Print information specific to an algorithm*/
        virtual void PrintInfo(void);

        /*Run the proximal gradient algorithm with acceleration*/
        void Run(void);
        
        /*PARAMSMAP is defined in "def.h" and it is a map from string to realdp, i.e., std::map<std::string, realdp> .
        This function is used to set the parameters by the mapping*/
        virtual void SetParams(PARAMSMAP params);
        
        integer SGIterGap;
	protected:
        /*Print last information in an algorithm*/
        virtual void PrintFinalInfo(void);

        /*The variable x1 and x2 declear in Solvers.h are used for safeguard*/
        /*The variable z1 and z2 are iterates*/
        Variable *z1;
        Variable *z2;
        realdp fz1;
        realdp fz2;
        Vector *gfz1;
        Vector *gfz2;
        /*The variable y1 and y2 are auxiliary iterates*/
        Variable *y1;
        Variable *y2;
        realdp fy1;
        realdp fy2;
        Vector *gfy1;
        Vector *gfy2;
        
        realdp s1;
        realdp s2;
        
        realdp *initDy;
	};
}; /*end of ROPTLIB namespace*/
#endif // end of RPG_H
