
#include "Solvers/RMBB.h"

/*Define the namespace*/
namespace ROPTLIB{

	RMBB::RMBB(const Problem *prob, const Variable *initialx, const Variable *insoln)
	{
		Initialization(prob, initialx, insoln);
	};
	
	RMBB::~RMBB(void)
	{
		delete s;
		delete y;
		delete Py;
		delete initscalings;
		DeleteVectors(S, LengthSY);
		DeleteVectors(Y, LengthSY);
		if (RHO != nullptr)
			delete[] RHO;
	};

	void RMBB::SetProbX(const Problem *prob, const Variable *initialx, const Variable *insoln)
	{
		SolversLS::SetProbX(prob, initialx, insoln);
		const Vector *EMPTYETA;
		if (prob->GetDomain()->GetIsIntrinsic())
			EMPTYETA = prob->GetDomain()->GetEMPTYINTR();
		else
			EMPTYETA = prob->GetDomain()->GetEMPTYEXTR();
		initscalings = EMPTYETA->ConstructEmpty();
		s = EMPTYETA->ConstructEmpty();
		y = EMPTYETA->ConstructEmpty();
		Py = EMPTYETA->ConstructEmpty();
		prob->SetUseGrad(true);
		prob->SetUseHess(false);
	};

	void RMBB::Run(void)
	{
		DeleteVectors(S, LengthSY);
		NewVectors(S, LengthSY);
		DeleteVectors(Y, LengthSY);
		NewVectors(Y, LengthSY);
		if (RHO != nullptr)
			delete[] RHO;
		RHO = new realdp[LengthSY];
		SolversLS::Run();
	};

	void RMBB::CheckParams(void)
	{
		SolversLS::CheckParams();
		char YES[] = "YES";
		char NO[] = "NO";
		char *status;

		printf("RMBB METHOD PARAMETERS:\n");
		status = (numBlocks >= 0 && numBlocks <= 1000000000000) ? YES : NO;
		printf("numBlocks     :%15d[%s],\n", numBlocks, status);
	};

	void RMBB::SetDefaultParams()
	{
		SolversLS::SetDefaultParams();
		isconvex = false;
		BBratio = 1;
		Num_pre_BB = 0;
		nu = static_cast<realdp> (1e-4);
		mu = 1;
		LengthSY = 4;
		S = nullptr;
		Y = nullptr;
		Currentlength = 0;
		beginidx = 0;
		RHO = nullptr;
		gamma = 1;
		InitSteptype = ONESTEP;
		SolverName.assign("RMBB");
		numBlocks = 1;
	};

	void RMBB::GetSearchDir(void)
	{
		Mani->ScaleTimesVector(x1, -1, gf1, eta1);

		if (iter == 0)
			stepsize = Initstepsize / ngf1;
		else
		{
			realdp *isptr = initscalings->ObtainWriteEntireData();

			for (integer i = 0; i < s->Getlength(); i++)
			{
				isptr[i] = 0;
				realdp rsry = 0, ryry = 0;
				integer idx;
				for (integer j = 0; j < Currentlength; j++)
				{
					idx = (beginidx + j) % LengthSY;
					rsry += S[idx]->ObtainReadData()[i] * Y[idx]->ObtainReadData()[i];
					ryry += Y[idx]->ObtainReadData()[i] * Y[idx]->ObtainReadData()[i];
				}
				if (std::fabs(ryry) < std::numeric_limits<realdp>::epsilon())
					isptr[i] = 0;
				else
					isptr[i] = std::fabs(rsry / ryry);
			}
			realdp *eta1ptr = eta1->ObtainWritePartialData();
			for (integer i = 0; i < eta1->Getlength(); i++)
				eta1ptr[i] *= isptr[i];

			stepsize = 1;

			//Vector *s = eta2->ConstructEmpty();
			//Vector *y = eta2->ConstructEmpty();
			///*Since x1 and x2 are swapped, we have the following formula.*/
			//Mani->VectorTransport(x2, eta2, x1, eta2, s);
			///*Since gf1 and gf2 are swapped, we have the following formula.*/
			//Mani->VectorTransport(x2, eta2, x1, gf2, y);
			//Mani->VectorMinusVector(x2, gf1, y, y);

			//integer *BlockIdx = new integer[numBlocks + 1];
			//if (s->Getlength() < numBlocks)
			//{
			//	printf("Warning: numBlocks is less than the length of vector. Reset numBlocks!\n");
			//	numBlocks = s->Getlength();
			//}
			//integer step = s->Getlength() / numBlocks;
			//BlockIdx[0] = 0;
			//for (integer i = 1; i < numBlocks; i++)
			//{
			//	BlockIdx[i] = i * step;
			//}
			//BlockIdx[numBlocks] = s->Getlength();

			//const realdp *sptr = s->ObtainReadData();
			//const realdp *yptr = y->ObtainReadData();
			//realdp *isptr = initscalings->ObtainWriteEntireData();
			//realdp BB = Mani->Metric(x2, s, y) / Mani->Metric(x2, y, y);
			//for (integer i = 0; i < numBlocks; i++)
			//{
			//	integer length = BlockIdx[i + 1] - BlockIdx[i];
			//	realdp sy = dot_(&length, const_cast<realdp *> (sptr + BlockIdx[i]), &GLOBAL::IONE, const_cast<realdp *> (yptr + BlockIdx[i]), &GLOBAL::IONE);
			//	realdp yy = dot_(&length, const_cast<realdp *> (yptr + BlockIdx[i]), &GLOBAL::IONE, const_cast<realdp *> (yptr + BlockIdx[i]), &GLOBAL::IONE);
			//	realdp BBi = 0;
			//	if (yy != 0)
			//		BBi = sy / yy;
			//	else
			//		BBi = 0;
			//	std::cout << "i:" << i << ", sy:" << sy << ", yy:" << yy << ", BBi:" << BBi << ", ";//--- << std::endl;//---
			//	for (integer j = BlockIdx[i]; j < BlockIdx[i + 1]; j++)
			//	{
			//		isptr[j] = (BBi > 0) ? BBi : 0;//-- BB;
			//	}
			//}

			//std::cout << "BB2:" << Mani->Metric(x2, s, y) / Mani->Metric(x2, y, y) << std::endl;//---

			////initscalings->Print("IS:");//---
			//delete[] BlockIdx;

			//realdp *eta1ptr = eta1->ObtainWritePartialData();
			//for (integer i = 0; i < eta1->Getlength(); i++)
			//	eta1ptr[i] *= isptr[i];

			//stepsize = 1;
			//delete s;
			//delete y;
		}
	};

	void RMBB::UpdateData(void)
	{
		UpdateDataLRBFGS(s, y, Py, S, Y, RHO, Currentlength, betay, inpss, inpsy, inpyy,
                         rho, gamma, beginidx, isupdated);
	};

	void RMBB::SetParams(PARAMSMAP params)
	{
		SolversLS::SetParams(params);
		PARAMSMAP::iterator iter;
		for (iter = params.begin(); iter != params.end(); iter++)
		{
			if (iter->first == static_cast<std::string> ("numBlocks"))
			{
				numBlocks = static_cast<integer> (iter->second);
			}
		}
	};
}; /*end of ROPTLIB namespace*/
