/*
This file defines the class of the Riemannian Multiple block BB method. This method is still under investigation.
The current version does not work well.
The idea is to scale the steepest descent direction in blocks using blocked BB step sizes.

Solvers --> QuasiNewton --> SolversLS --> RMBB

---- WH
*/

#ifndef RMBB_H
#define RMBB_H

#include "Solvers/SolversLS.h"
#include "Others/def.h"

/*Define the namespace*/
namespace ROPTLIB{

	class RMBB : public SolversLS{
	public:
		/*The contructor of RSD method. It calls the function Solvers::Initialization.
		INPUT : prob is the problem which defines the cost function, gradient and possible the action of Hessian
		and specifies the manifold of domain.
		initialx is the initial iterate.
		insoln is the true solution. It is not required and only used for research.*/
		RMBB(const Problem *prob, const Variable *initialx, const Variable *insoln = nullptr);

		/*Destructor. Delete the vector initscalings*/
		~RMBB(void);

		/*Call Solvers::SetProbX function and indicate RCG does not need action of Hessian.
		INPUT:	prob is the problem which defines the cost function, gradient and possible the action of Hessian
		and specifies the manifold of domain.
		initialx is the initial iterate.
		insoln is the true solution. It is not required and only used for research.*/
		virtual void SetProbX(const Problem *prob, const Variable *initialx, const Variable *insoln);
		
		/*New or delete S and Y if needed*/
		virtual void Run(void);

		/*Setting parameters (member variables) to be default values */
		virtual void SetDefaultParams();

		/*PARAMSMAP is defined in "def.h" and it is a map from string to realdp, i.e., std::map<std::string, realdp> .
		This function is used to set the parameters by the mapping*/
		virtual void SetParams(PARAMSMAP params);

		/*Check whether the parameters about RMBB are legal or not.*/
		virtual void CheckParams(void);

		integer numBlocks;
	protected:
		/*Set the search direction to be negative gradient*/
		virtual void GetSearchDir();

		/*Update S and Y*/
		virtual void UpdateData(void);

		Vector *initscalings;
        
        bool isupdated; /*Mark whether the (inverse) Hessian approximation is updated*/
        realdp betay, inpsy, inpss, inpyy;  /*betay: \|\xi\| / \|\mathcal{T}_{R_\xi} \xi\| in the locking condition;
                                                phic: the coefficient (1-phic) BFGS + phi DFP in Broyden family method
                                                inpsy: g(s, y); inpss: g(s, s); inpyy: g(y, y); */
        Vector *s, *y;/*the s, y, and u of current step*/
        Vector *Py; /*the preconditioned y.*/
        Vector **S, **Y; /*The stored pairs of s and y*/
        realdp *RHO; /*the sequence of 1 / g(s_k, y_k), where Hessian approximation k-th iteration is updated*/
        realdp rho, gamma; /*rho: 1/g(s, y) at current iteration, gamma: g(s, y) / g(y, y) for LRBFGS and gamma: g(y, y) / g(s, y) for RTRSR1*/
        integer Currentlength; /*The current length of array S, Y and RHO*/
        integer beginidx; /*The starting index of S, Y and RHO at current iteration*/
	};
}; /*end of ROPTLIB namespace*/
#endif // end of RMBB_H
