
#include "test/DriverMexProb.h"

#ifdef MATLAB_MEX_FILE

using namespace ROPTLIB;

std::map<integer *, integer> *CheckMemoryDeleted;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	CheckMemoryDeleted = new std::map<integer *, integer>;

	DriverMexProb(nlhs, plhs, nrhs, prhs);
	// check memory
	std::map<integer *, integer>::iterator iter = CheckMemoryDeleted->begin();
	for (iter = CheckMemoryDeleted->begin(); iter != CheckMemoryDeleted->end(); iter++)
	{
		if (iter->second != 1)
			printf("Global address: %p, sharedtimes: %d\n", iter->first, iter->second);
	}
	delete CheckMemoryDeleted;
	return;
}

void DriverMexProb(int &nlhs, mxArray ** &plhs, int &nrhs, const mxArray ** &prhs)
{
	if (nrhs < 6)
	{
		mexErrMsgTxt("The number of arguments should be at least five.\n");
	}
	// Argument Checking:
	// First to third arguments should be function handles
	if (!mxIsClass(prhs[0], "function_handle") || !mxIsClass(prhs[1], "function_handle") || !mxIsClass(prhs[2], "function_handle") || !mxIsClass(prhs[3], "function_handle"))
	{
		mexErrMsgTxt("At least one of first to fourth input arguments is not a function handle.");
	}
	// fifth to sixth arguments are structures
	// The fifth one is SolverParams
	// The sixth one is ManiParams
	if (!mxIsStruct(prhs[4]) || !mxIsStruct(prhs[5]))
	{
		mexErrMsgTxt("At least one of fifth to sixth input arguments is not a structure.");
	}

	// Obtain manifold and iterate structure
	Manifold *domain = nullptr, **manifolds = nullptr;
	Variable *initialX = nullptr, *soln = nullptr;
	Element **elements;
	integer *powsinterval, numoftype, numoftotal;

	if (!ParseManiParams(prhs[5], manifolds, elements, powsinterval, numoftype, numoftotal))
	{
		mexErrMsgTxt("Parsing ManiParams fails.");
	}
	if (numoftotal > 1)
		domain = new ProductManifold(manifolds, numoftype, powsinterval, numoftotal);
	else
		domain = manifolds[0];
	mxArray *tmp = mexProblem::GetFieldbyName(prhs[5], 0, "IsCheckParams");
	if (tmp != nullptr)
	{
		if (fabs(mxGetScalar(tmp)) > std::numeric_limits<realdp>::epsilon()) // if the value is nonzero
		{
			domain->CheckParams();
		}
	}
	bool HasHHR = false;
	if (nrhs >= 7)
	{
		if (!mxIsDouble(prhs[6]))
		{
			mexErrMsgTxt("Seventh input argument is not a scalar.");
		}

		HasHHR = (static_cast<integer> (mxGetScalar(prhs[6])) != 0);
	}
	domain->SetHasHHR(HasHHR);
	if (numoftotal > 1)
		initialX = new ProductElement(elements, numoftotal, powsinterval, numoftype);
	else
		initialX = elements[0]->ConstructEmpty();
	//initialX->Print("initialX");

	// initialize the initial iterate
	if (nrhs >= 8)
	{
		if (!mxIsStruct(prhs[7]))
		{
			mexErrMsgTxt("Eighth input argument is not a structure.");
		}
		mexProblem::ObtainElementFromMxArray(initialX, prhs[7]);
	}
	else
	{
		initialX->RandInManifold();
	}

	// 	initialX->Print("initialX", false);

	if (nrhs >= 9)
	{
		if (numoftotal > 1)
			soln = new ProductElement(elements, numoftotal, powsinterval, numoftype);
		else
			soln = elements[0]->ConstructEmpty();
		if (!mxIsStruct(prhs[8]))
		{
			mexErrMsgTxt("Eighth input argument is not a structure.");
		}
		mexProblem::ObtainElementFromMxArray(soln, prhs[8]);
	}
	//mexProblem::ObtainMxArrayFromElement(plhs[0], initialX);

	// Define the problem
	Problem *Prob = new mexProblem(prhs[0], prhs[1], prhs[2], prhs[3]);
	Prob->SetDomain(domain);
	//	Vector *egf = initialX->ConstructEmpty();
	//	Prob->EucGrad(initialX, egf);
	//	delete egf;
	// solve the optimization problem
	ParseSolverParamsAndOptimizing(prhs[4], Prob, initialX, soln, plhs);

	delete Prob;
	if (numoftotal > 1)
		delete domain;
	delete initialX;
	delete soln;

	for (integer i = 0; i < numoftype; i++)
	{
		delete manifolds[i];
		delete elements[i];
	}
	delete[] manifolds;
	delete[] elements;
	delete[] powsinterval;
};

#endif // end of MATLAB_MEX_FILE
