
#include "test/TestEucQuadratic.h"

using namespace ROPTLIB;

void testEucQuadratic(void)
{
	// size of the domain
	integer dim = 10;

	// Generate the matrices in the Euclidean Quadratic problem.
	// Use blas to obtain a positive definite matrix by M = Temp * Temp^T
	realdp *M = new realdp[dim * dim];
	realdp *Temp = new realdp[dim * dim];
	for (integer i = 0; i < dim * dim; i++)
		Temp[i] = genrandnormal();
	char *transn = const_cast<char *> ("n"), *transt = const_cast<char *> ("t");
	realdp one = 1, zero = 0;
	integer N = dim;
	// M = temp * temp^T, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
	gemm_(transn, transt, &N, &N, &N, &one, Temp, &N, Temp, &N, &zero, M, &N);

	delete[] Temp;

	testEucQuadratic(M, dim);
	delete[] M;

}

void testEucQuadraticMore()
{
	// size of the domain
	integer dim = 4;

	// Generate the matrices in the Euclidean Quadratic problem.
	// Use blas to obtain a positive definite matrix by M = Temp * Temp^T
	realdp *M = new realdp[dim * dim];
	realdp *Temp = new realdp[dim * dim];
	for (integer i = 0; i < dim * dim; i++)
		Temp[i] = genrandnormal();
	char *transn = const_cast<char *> ("n"), *transt = const_cast<char *> ("t");
	realdp one = 1, zero = 0;
	integer N = dim;
	// M = temp * temp^T, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
	gemm_(transn, transt, &N, &N, &N, &one, Temp, &N, Temp, &N, &zero, M, &N);

	//for (integer i = 0; i < dim * dim; i++)
	//	M[i] = 0;
	//for (integer i = 0; i < dim; i++)
	//	M[i + i * dim] = i;

	delete[] Temp;

	// Obtain an initial iterate
	EucVariable EucX(dim, 1);
	EucX.RandInManifold();

	// Define the manifold
	Euclidean Domain(dim);

	// Define the problem
	EucQuadratic Prob(M, dim);
	Prob.SetDomain(&Domain);

	RSD *RSDsolver = new RSD(&Prob, &EucX);
	RSDsolver->Debug = ITERRESULT;//--- FINALRESULT;
	RSDsolver->InitSteptype = BBSTEP;
	RSDsolver->Max_Iteration = 100;
	RSDsolver->CheckParams();
	RSDsolver->Run();
	delete RSDsolver;

	RMBB *RMBBsolver = new RMBB(&Prob, &EucX);
	RMBBsolver->numBlocks = 2;
	RMBBsolver->Debug = ITERRESULT;//--- FINALRESULT;
	RMBBsolver->InitSteptype = ONESTEP;
	RMBBsolver->Max_Iteration = 100;
	RMBBsolver->CheckParams();
	RMBBsolver->Run();
	delete RMBBsolver;

	delete[] M;
};

void testEucQuadratic(realdp *M, integer dim, realdp *X, realdp *Xopt)
{
	// Obtain an initial iterate
	EucVariable EucX(dim, 1);
	if (X == nullptr)
	{
		EucX.RandInManifold();
	}
	else
	{
		realdp *EucXptr = EucX.ObtainWriteEntireData();
		for (integer i = 0; i < dim; i++)
			EucXptr[i] = X[i];
	}

	// Define the manifold
	Euclidean Domain(dim);

	// Define the problem
	EucQuadratic Prob(M, dim);
	Prob.SetDomain(&Domain);

	// test RSD
	printf("********************************Check all line search algorithm in RSD*****************************************\n");
	for (integer i = 0; i < 1; i++)//--INPUTFUN
	{
		RSD *RSDsolver = new RSD(&Prob, &EucX);
		RSDsolver->LineSearch_LS = static_cast<LSAlgo> (i);
        RSDsolver->Debug = FINALRESULT;//-- ITERRESULT;
//		RSDsolver->CheckParams();
		RSDsolver->Run();
		delete RSDsolver;
	}
	//// test RNewton
	//printf("********************************Check all line search algorithm in RNewton*************************************\n");
	//for (integer i = 0; i < INPUTFUN; i++)
	//{
	//	RNewton *RNewtonsolver = new RNewton(&Prob, &EucX);
	//	RNewtonsolver->LineSearch_LS = static_cast<LSAlgo> (i);
	//	RNewtonsolver->Debug = FINALRESULT;
	//	RNewtonsolver->CheckParams();
	//	RNewtonsolver->Run();
	//	delete RNewtonsolver;
	//}

	//// test RCG
	//printf("********************************Check all Formulas in RCG*************************************\n");
	//for (integer i = 0; i < RCGMETHODSLENGTH; i++)
	//{
	//	RCG *RCGsolver = new RCG(&Prob, &EucX);
	//	RCGsolver->RCGmethod = static_cast<RCGmethods> (i);
	//	RCGsolver->LineSearch_LS = STRONGWOLFE;
	//	RCGsolver->LS_beta = 0.1;
	//	RCGsolver->Debug = FINALRESULT;
	//	RCGsolver->CheckParams();
	//	RCGsolver->Run();
	//	delete RCGsolver;
	//}

	//// test RBroydenFamily
	//printf("********************************Check all line search algorithm in RBroydenFamily*************************************\n");
	//for (integer i = 0; i < INPUTFUN; i++)
	//{
	//	RBroydenFamily *RBroydenFamilysolver = new RBroydenFamily(&Prob, &EucX);
	//	RBroydenFamilysolver->LineSearch_LS = static_cast<LSAlgo> (i);
	//	RBroydenFamilysolver->Debug = FINALRESULT;
	//	RBroydenFamilysolver->CheckParams();
	//	RBroydenFamilysolver->Run();
	//	delete RBroydenFamilysolver;
	//}

	//// test RWRBFGS
	//printf("********************************Check all line search algorithm in RWRBFGS*************************************\n");
	//for (integer i = 0; i < INPUTFUN; i++)
	//{
	//	RWRBFGS *RWRBFGSsolver = new RWRBFGS(&Prob, &EucX);
	//	RWRBFGSsolver->LineSearch_LS = static_cast<LSAlgo> (i);
	//	RWRBFGSsolver->Debug = FINALRESULT;
	//	RWRBFGSsolver->CheckParams();
	//	RWRBFGSsolver->Run();
	//	delete RWRBFGSsolver;
	//}

	//// test RBFGS
	//printf("********************************Check all line search algorithm in RBFGS*************************************\n");
	//for (integer i = 0; i < INPUTFUN; i++)
	//{
	//	RBFGS *RBFGSsolver = new RBFGS(&Prob, &EucX);
	//	RBFGSsolver->LineSearch_LS = static_cast<LSAlgo> (i);
	//	RBFGSsolver->Debug = FINALRESULT;
	//	RBFGSsolver->CheckParams();
	//	RBFGSsolver->Run();
	//	delete RBFGSsolver;
	//}

// 	// test LRBFGS
// 	//printf("********************************Check all line search algorithm in LRBFGS*************************************\n");
// 	for (integer i = 0; i < 1; i++)//INPUTFUN
// 	{
// 		LRBFGS *LRBFGSsolver = new LRBFGS(&Prob, &EucX);
// 		LRBFGSsolver->LineSearch_LS = static_cast<LSAlgo> (i);
// 		LRBFGSsolver->Debug = FINALRESULT;
// 		//LRBFGSsolver->CheckParams();
// 		LRBFGSsolver->Max_Iteration = 200;
// 		LRBFGSsolver->Run();
// 		if (LRBFGSsolver->Getnormgfgf0() < 1e-6)
// 			printf("SUCCESS!\n");
// 		else
// 			printf("FAIL!\n");
// 		delete LRBFGSsolver;
// 	}

	//printf("********************************Check RTRSD*************************************\n");
	//RTRSD RTRSDsolver(&Prob, &EucX);
	//printf("\n");
	//RTRSDsolver.Debug = FINALRESULT;
	//RTRSDsolver.CheckParams();
	//RTRSDsolver.Run();

	//printf("********************************Check RTRNewton*************************************\n");
	//RTRNewton RTRNewtonsolver(&Prob, &EucX);
	//printf("\n");
	//RTRNewtonsolver.Debug = FINALRESULT;
	//RTRNewtonsolver.CheckParams();
	//RTRNewtonsolver.Run();

	//printf("********************************Check RTRSR1*************************************\n");
	//RTRSR1 RTRSR1solver(&Prob, &EucX);
	//printf("\n");
	//RTRSR1solver.Debug = FINALRESULT;
	//RTRSR1solver.CheckParams();
	//RTRSR1solver.Run();

	//printf("********************************Check LRTRSR1*************************************\n");
	//LRTRSR1 LRTRSR1solver(&Prob, &EucX);
	//printf("\n");
	//LRTRSR1solver.Debug = FINALRESULT;
	//LRTRSR1solver.CheckParams();
	//LRTRSR1solver.Run();

	//// Check gradient and Hessian
	//Prob.CheckGradHessian(&EucX);
	//const Variable *xopt = RTRNewtonsolver.GetXopt();
	//Prob.CheckGradHessian(xopt);

	//if (Xopt != nullptr)
	//{
	//	const realdp *xoptptr = xopt->ObtainReadData();
	//	for (integer i = 0; i < dim; i++)
	//		Xopt[i] = xoptptr[i];
	//}
};

#ifdef MATLAB_MEX_FILE

std::map<integer *, integer> *CheckMemoryDeleted;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    if(nrhs < 2)
    {
        mexErrMsgTxt("The number of arguments should be at least two.\n");
    }
    
	realdp *M, *X, *Xopt;
	M = mxGetPr(prhs[0]);
	X = mxGetPr(prhs[1]);
	/* dimensions of input matrices */
	integer dim;
	dim = mxGetM(prhs[0]);
    if(mxGetN(prhs[0]) != dim)
    {
        mexErrMsgTxt("The size of matrix is not correct.\n");
    }
    if(mxGetM(prhs[1]) != dim || mxGetN(prhs[1]) != 1)
    {
        mexErrMsgTxt("The size of the initial X is not correct!\n");
    }
    
	printf("dim:%d\n", dim);

	/*create output matrix*/
	plhs[0] = mxCreateDoubleMatrix(dim, 1, mxREAL);
	Xopt = mxGetPr(plhs[0]);

	genrandseed(0);

	CheckMemoryDeleted = new std::map<integer *, integer>;
	testEucQuadratic(M, dim, X, Xopt);
	std::map<integer *, integer>::iterator iter = CheckMemoryDeleted->begin();
	for (iter = CheckMemoryDeleted->begin(); iter != CheckMemoryDeleted->end(); iter++)
	{
		if (iter->second != 1)
			printf("Global address: %p, sharedtimes: %d\n", iter->first, iter->second);
	}
	delete CheckMemoryDeleted;
	return;
}

#endif
