﻿
#include "test/TestMNCRModel.h"

using namespace ROPTLIB;

void testMNCRModel(void)
{
	// size of the samples and number of samples
	integer m = 4;
	realdp *c = new realdp[m + m * m];
	realdp *A = c + m;
	realdp L = 1;

	//for (integer i = 0; i < m; i++)
	//	c[i] = - 2.0 / m;
	//for (integer i = 0; i < m * m; i++)
	//	A[i] = 0;
	//for (integer i = 0; i < m; i++)
	//	A[i + i * m] = 1;

	for (integer i = 0; i < m; i++)
		c[i] = genrandnormal();
	for (integer i = 0; i < m; i++)
	{
		for (integer j = i; j < m; j++)
		{
			A[i + j * m] = genrandnormal();
			A[j + i * m] = A[i + j * m];
		}
	}

	testMNCRModel(c, A, L, m);
	delete[] c;
}

void testMNCRModel(realdp *c, realdp *A, realdp L, integer m, realdp *X, realdp *Xopt)
{
	// Obtain an initial iterate
	MNVariable MNX(1, m);
	if (X == nullptr)
	{
		MNX.RandInManifold();
	}
	else
	{
		realdp *MNXptr = MNX.ObtainWriteEntireData();
		for (integer i = 0; i < m; i++)
			MNXptr[i] = X[i];
	}

	// Define the manifold
	Multinomial Domain(1, m);

	// Define the problem
	MNCRModel Prob(c, A, L, m);
	Prob.SetDomain(&Domain);

	//Domain.CheckRetraction(&MNX);
	//Domain.CheckDiffRetraction(&MNX, true);
	//Domain.CheckLockingCondition(&MNX);
	//Domain.CheckcoTangentVector(&MNX);
	//Domain.CheckIsometryofVectorTransport(&MNX);
	//Domain.CheckIsometryofInvVectorTransport(&MNX);
	//Domain.CheckVecTranComposeInverseVecTran(&MNX);
	//Domain.CheckTranHInvTran(&MNX);
	//Domain.CheckHaddScaledRank1OPE(&MNX);

	//Prob.CheckGradHessian(&MNX);//----
	
	// test RSD
	for (integer i = 0; i < 1; i++)
	{
		RSD *RSDsolver = new RSD(&Prob, &MNX);
		RSDsolver->LineSearch_LS = static_cast<LSAlgo> (i);
		RSDsolver->Debug = FINALRESULT;
		RSDsolver->Max_Iteration = 100;
		RSDsolver->Num_pre_funs = 5;
		RSDsolver->PreFunsAccuracy = 1e-6;
		RSDsolver->Tolerance = 1e-10;
		//RSDsolver->CheckParams();
		RSDsolver->Run();
		delete RSDsolver;
	}
	//// test RNewton
	//for (integer i = 0; i < 1; i++)
	//{
	//	RNewton *RNewtonsolver = new RNewton(&Prob, &MNX);
	//	RNewtonsolver->LineSearch_LS = static_cast<LSAlgo> (i);
	//	RNewtonsolver->Debug = FINALRESULT;
	//	RNewtonsolver->Num_pre_funs = 5;
	//	RNewtonsolver->PreFunsAccuracy = 1e-6;
	//	RNewtonsolver->Tolerance = 1e-10;
	//	//RNewtonsolver->CheckParams();
	//	RNewtonsolver->Run();
	//	delete RNewtonsolver;
	//}

	//// test RCG
	//for (integer i = RCGMETHODSLENGTH - 1; i < RCGMETHODSLENGTH; i++) //RCGMETHODSLENGTH
	//{
	//	RCG *RCGsolver = new RCG(&Prob, &MNX);
	//	RCGsolver->RCGmethod = static_cast<RCGmethods> (i);
	//	RCGsolver->LineSearch_LS = ARMIJO;
	//	RCGsolver->LS_beta = 0.1;
	//	RCGsolver->Num_pre_funs = 5;
	//	RCGsolver->PreFunsAccuracy = 1e-6;
	//	RCGsolver->Tolerance = 1e-10;
	//	RCGsolver->Debug = ITERRESULT;
	//	RCGsolver->CheckParams();
	//	RCGsolver->Run();
	//	delete RCGsolver;
	//}

	//// test RBroydenFamily
	//for (integer i = 0; i < 1; i++)
	//{
	//	RBroydenFamily *RBroydenFamilysolver = new RBroydenFamily(&Prob, &MNX);
	//	RBroydenFamilysolver->LineSearch_LS = static_cast<LSAlgo> (i);
	//	RBroydenFamilysolver->Debug = FINALRESULT;
	//	RBroydenFamilysolver->Num_pre_funs = 5;
	//	RBroydenFamilysolver->PreFunsAccuracy = 1e-6;
	//	RBroydenFamilysolver->Tolerance = 1e-10;
	//	//RBroydenFamilysolver->CheckParams();
	//	RBroydenFamilysolver->Run();
	//	delete RBroydenFamilysolver;
	//}

	//// test RWRBFGS
	//for (integer i = 0; i < 1; i++)
	//{
	//	RWRBFGS *RWRBFGSsolver = new RWRBFGS(&Prob, &MNX);
	//	RWRBFGSsolver->LineSearch_LS = static_cast<LSAlgo> (i);
	//	RWRBFGSsolver->Debug = FINALRESULT;
	//	RWRBFGSsolver->Num_pre_funs = 5;
	//	RWRBFGSsolver->PreFunsAccuracy = 1e-6;
	//	RWRBFGSsolver->Tolerance = 1e-10;
	//	//RWRBFGSsolver->CheckParams();
	//	RWRBFGSsolver->Run();
	//	delete RWRBFGSsolver;
	//}

	//// test RBFGS
	//for (integer i = 0; i < 1; i++)
	//{
	//	RBFGS *RBFGSsolver = new RBFGS(&Prob, &MNX);
	//	RBFGSsolver->LineSearch_LS = static_cast<LSAlgo> (i);
	//	RBFGSsolver->Debug = FINALRESULT;
	//	RBFGSsolver->Num_pre_funs = 5;
	//	RBFGSsolver->PreFunsAccuracy = 1e-6;
	//	RBFGSsolver->Tolerance = 1e-10;
	//	//RBFGSsolver->CheckParams();
	//	RBFGSsolver->Run();
	//	delete RBFGSsolver;
	//}

	//// test LRBFGS
	//for (integer i = 0; i < 1; i++) // INPUTFUN
	//{
	//	LRBFGS *LRBFGSsolver = new LRBFGS(&Prob, &MNX);
	//	LRBFGSsolver->LineSearch_LS = static_cast<LSAlgo> (i);
	//	LRBFGSsolver->Debug = ITERRESULT;
	//	LRBFGSsolver->CheckParams();
	//	LRBFGSsolver->Num_pre_funs = 5;
	//	LRBFGSsolver->PreFunsAccuracy = 1e-7;
	//	LRBFGSsolver->Tolerance = 1e-10;
	//	LRBFGSsolver->Max_Iteration = 50;
	//	LRBFGSsolver->LengthSY = 4;
	//	LRBFGSsolver->Run();
	//	LRBFGSsolver->GetXopt()->Print("xopt:");//--
	//	delete LRBFGSsolver;
	//}

	//RTRSD RTRSDsolver(&Prob, &MNX);
	//RTRSDsolver.Debug = FINALRESULT;
	////RTRSDsolver.CheckParams();
	//RTRSDsolver.Run();

	//RTRNewton RTRNewtonsolver(&Prob, &MNX);
	//RTRNewtonsolver.Debug = FINALRESULT;
	////RTRNewtonsolver.Max_Iteration = 1000;
	////RTRNewtonsolver.OutputGap = 10;
	////RTRNewtonsolver.CheckParams();
	//RTRNewtonsolver.Run();

	//RTRSR1 RTRSR1solver(&Prob, &MNX);
	//RTRSR1solver.Debug = FINALRESULT;
	////RTRSR1solver.CheckParams();
	//RTRSR1solver.Run();

	//LRTRSR1 LRTRSR1solver(&Prob, &MNX);
	//LRTRSR1solver.Debug = FINALRESULT;
	////LRTRSR1solver.CheckParams();
	//LRTRSR1solver.Run();

	//// Check gradient and Hessian
	//Prob.CheckGradHessian(&MNX);
	//const Variable *xopt = RTRSDsolver.GetXopt();
	//Prob.CheckGradHessian(xopt);

	//if (Xopt != nullptr)
	//{
	//	const realdp *xoptptr = xopt->ObtainReadData();
	//	for (integer i = 0; i < dim; i++)
	//		Xopt[i] = xoptptr[i];
	//}
};

#ifdef MATLAB_MEX_FILE

std::map<integer *, integer> *CheckMemoryDeleted;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs < 5)
	{
		mexErrMsgTxt("The number of arguments should be at least five.\n");
	}
	realdp *c, *A, *x, *Xopt;
	c = mxGetPr(prhs[0]);
	A = mxGetPr(prhs[1]);
	x = mxGetPr(prhs[2]);
	realdp L = mxGetScalar(prhs[3]);
	/* dimensions of input matrices */
	integer m;
	m = mxGetM(prhs[0]);

	/*Check the correctness of the inputs*/
	if (mxGetN(prhs[1]) != m || mxGetM(prhs[1]) != m || mxGetM(prhs[2]) != m)
	{
		mexErrMsgTxt("The sizes of c, A and x are not correct.\n");
	}
	
	/*create output matrix*/
	plhs[0] = mxCreateDoubleMatrix(m, 1, mxREAL);
	Xopt = mxGetPr(plhs[0]);

	genrandseed(0);

	CheckMemoryDeleted = new std::map<integer *, integer>;
	//	testStieBrockett(B, D, n, p, X, Xopt);

	// Obtain an initial iterate by taking the Q factor of qr decomposition
	MNVariable MNX(1, m);
	realdp *MNXptr = MNX.ObtainWriteEntireData();
	for (integer i = 0; i < m; i++)
		MNXptr[i] = x[i];

	// Define the manifold
	Multinomial Domain(1, m);
	
	// Define the Brockett problem
	MNCRModel Prob(c, A, L, m);
	Prob.SetDomain(&Domain);
	// Call the function defined in DriverMexProb.h
	ParseSolverParamsAndOptimizing(prhs[4], &Prob, &MNX, nullptr, plhs);

	std::map<integer *, integer>::iterator iter = CheckMemoryDeleted->begin();
	for (iter = CheckMemoryDeleted->begin(); iter != CheckMemoryDeleted->end(); iter++)
	{
		if (iter->second != 1)
			printf("Global address: %p, sharedtimes: %d\n", iter->first, iter->second);
	}
	delete CheckMemoryDeleted;
	return;
}

#endif
