
#include "test/TestObliqueQCor.h"

using namespace ROPTLIB;

void testObliqueQCor()
{
	integer n = 4, num = 20;
	// Obtain an initial iterate by taking the Q factor of qr decomposition
	// lapack is used
	ObliqueQVariable OQV(n, num);

	//realdp *ptr = SCOX.ObtainWriteEntireData();
	//realdp a = 1.0 / sqrt(8) * 100000;
	//for (integer i = 0; i < 8; i++)
	//	ptr[i] = a;
	OQV.RandInManifold();

	ObliqueQVector OQVV(n, num);
	OQVV.RandGaussian();

	// Define the manifold
	ObliqueQ Domain(n, num);
	Domain.SetHasHHR(true);

	Domain.Projection(&OQV, &OQVV, &OQVV);

	realdp *A = new realdp[num * num];
	for (integer i = 0; i < num; i++)
	{
		for (integer j = 0; j <= i; j++)
		{
			A[j + i * num] = genrandnormal();
			A[i + j * num] = A[j + i * num];
		}
	}


	ObliqueQCor Prob(A, n, num);
	/*The domain of the problem is a Stiefel manifold*/
	Prob.SetDomain(&Domain);

	Domain.CheckParams();

	//Prob.f(&OQV);
	//Prob.EucGrad(&OQV, &OQVV);

	//Domain.CheckIntrExtr(&OQV);
	//Domain.CheckRetraction(&OQV);
	//Domain.CheckDiffRetraction(&OQV);
	//Domain.CheckIsometryofVectorTransport(&OQV);
	//Domain.CheckIsometryofInvVectorTransport(&OQV);
	//Domain.CheckLockingCondition(&OQV);
	//Domain.CheckcoTangentVector(&OQV);
	//Domain.CheckVecTranComposeInverseVecTran(&OQV);
	//Domain.CheckTranHInvTran(&OQV);
	//Domain.CheckHaddScaledRank1OPE(&OQV);

	 //Prob.CheckGradHessian(&OQV);

	RSD *RSDsolver = new RSD(&Prob, &OQV);
	RSDsolver->Debug = ITERRESULT;//--- FINALRESULT;
	RSDsolver->InitSteptype = BBSTEP;
	RSDsolver->Max_Iteration = 2000;
	//RSDsolver->Tolerance = 1e-9;
	RSDsolver->CheckParams();
	RSDsolver->Run();
	delete RSDsolver;

	delete[] A;
}

#ifdef MATLAB_MEX_FILE

/*Help to check the memory leakage problem. No necesary any more.*/
std::map<integer *, integer> *CheckMemoryDeleted;

/*This function checks the number and formats of input parameters.
nlhs: the number of output in mxArray format
plhs: the output objects in mxArray format
nrhs: the number of input in mxArray format
prhs: the input objects in mxArray format */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs < 5)
	{
		mexErrMsgTxt("The number of arguments should be at least four.\n");
	}
	realdp *A, *X, *Xopt, *soln;
	A = mxGetPr(prhs[0]);
	X = mxGetPr(prhs[1]);
	/* dimensions of input matrices */
	integer num, n, HasHHR, ParamSet;
	num = mxGetM(prhs[0]);
	n = mxGetM(prhs[1]);

	/*Check the correctness of the inputs*/
	if (mxGetN(prhs[0]) != num)
	{
		mexErrMsgTxt("The size of matrix A is not correct.\n");
	}
	if (mxGetN(prhs[1]) != num)
	{
		mexErrMsgTxt("The size of the initial X is not correct!\n");
	}

	HasHHR = static_cast<integer> (mxGetScalar(prhs[2]));
	ParamSet = static_cast<integer> (mxGetScalar(prhs[4]));

	printf("(n, num):%d,%d\n", n, num);

	/*create output matrix*/
	plhs[0] = mxCreateDoubleMatrix(n, num, mxREAL);
	Xopt = mxGetPr(plhs[0]);

	genrandseed(0);

	CheckMemoryDeleted = new std::map<integer *, integer>;

	// Obtain an initial iterate by taking the Q factor of qr decomposition
	ObliqueQVariable OQVX(n, num);
	realdp *OQVXptr = OQVX.ObtainWriteEntireData();
	for (integer i = 0; i < n * num; i++)
		OQVXptr[i] = X[i];

	ObliqueQVariable *OQVsoln = nullptr;
	if (nrhs >= 6)
	{
		realdp *soln = mxGetPr(prhs[5]); /*soln: n by num*/
		OQVsoln = new ObliqueQVariable(n, num);
		realdp *OQVsolnptr = OQVsoln->ObtainWriteEntireData();
		for (integer i = 0; i < n * num; i++)
		{
			OQVsolnptr[i] = soln[i];
		}
	}

	// Define the manifold
	ObliqueQ Domain(n, num);

	if (ParamSet == 2)
		Domain.ChooseObliqueQParamsSet2();
	else
		Domain.ChooseObliqueQParamsSet1();

	// Define the problem
	ObliqueQCor Prob(A, n, num);
	Prob.SetDomain(&Domain);

	Domain.SetHasHHR(HasHHR != 0);
	//Domain.CheckParams();

	// Call the function defined in DriverMexProb.h
	ParseSolverParamsAndOptimizing(prhs[3], &Prob, &OQVX, OQVsoln, plhs);

	std::map<integer *, integer>::iterator iter = CheckMemoryDeleted->begin();
	for (iter = CheckMemoryDeleted->begin(); iter != CheckMemoryDeleted->end(); iter++)
	{
		if (iter->second != 1)
			printf("Global address: %p, sharedtimes: %d\n", iter->first, iter->second);
	}
	delete CheckMemoryDeleted;
	delete OQVsoln;
	return;
}

#endif
