
#include "test/TestSPDMean.h"

using namespace ROPTLIB;

void testSPDMean(void)
{
	/*Randomly generate a point on the SPD manifold*/
	integer n = 10, num = 4;
	SPDVariable SPDX(n);
	realdp *initialX = SPDX.ObtainWriteEntireData();
	for (integer i = 0; i < n; i++)
	{
		for (integer j = 0; j < n; j++)
		{
			initialX[i + j * n] = 0;
		}
		initialX[i + i * n] = 1;
	}

	// Define the manifold
	SPDManifold Domain(n);
	Domain.SetHasHHR(true); /*set whether the manifold uses the idea in [HGA2015, Section 4.3] or not*/

	realdp *Ls = new realdp[n * n * num + n * n];
	realdp *tmp = Ls + n * n * num;
	integer info;
	for (integer i = 0; i < num; i++)
	{
		for (integer j = 0; j < n * n; j++)
			tmp[j] = genrandnormal();

		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Ls + i * n * n, &n);


		potrf_(GLOBAL::L, &n, Ls + i * n * n, &n, &info);
		if (info != 0)
		{
			printf("Warning: TestSPDMean Cholesky decomposition fails with info:%d!\n", info);
		}
		for (integer j = 0; j < n; j++)
			for (integer k = j + 1; k < n; k++)
				Ls[j + k * n + i * n * n] = 0;
	}

	// Define the problem
	SPDMean Prob(Ls, n, num);
	/*The domain of the problem is a SPD manifold*/
	Prob.SetDomain(&Domain);

	//Prob.CheckGradHessian(&SPDX);

	/*Output the parameters of the domain manifold*/
	//Domain.CheckParams();

	/*Check the correctness of the manifold operations*/
	//Domain.CheckIntrExtr(&SPDX);
	//Domain.CheckRetraction(&SPDX);
	//Domain.CheckDiffRetraction(&SPDX);
	//Domain.CheckLockingCondition(&SPDX);
	//Domain.CheckcoTangentVector(&SPDX);
	//Domain.CheckIsometryofVectorTransport(&SPDX);
	//Domain.CheckIsometryofInvVectorTransport(&SPDX);
	//Domain.CheckVecTranComposeInverseVecTran(&SPDX);
	//Domain.CheckTranHInvTran(&SPDX);
	//Domain.CheckHaddScaledRank1OPE(&SPDX);

	// test LRBFGS
	//printf("********************************Test Geometric mean in LRBFGS*************************************\n");
	LRBFGS *LRBFGSsolver = new LRBFGS(&Prob, &SPDX);
	LRBFGSsolver->LineSearch_LS = ARMIJO;
	LRBFGSsolver->Debug = FINALRESULT; //ITERRESULT;// 
	LRBFGSsolver->Max_Iteration = 50;
	LRBFGSsolver->Tolerance = static_cast<realdp> (1e-6);
	LRBFGSsolver->Accuracy = static_cast<realdp> (1e-4);
	LRBFGSsolver->Finalstepsize = 1;
	//LRBFGSsolver->CheckParams();
	LRBFGSsolver->Run();
	if (LRBFGSsolver->Getnormgfgf0() < 1e-6)
		printf("SUCCESS!\n");
	else
		printf("FAIL!\n");
	//// Check gradient and Hessian
	//Prob.CheckGradHessian(&SPDX);
	//const Variable *xopt = LRBFGSsolver->GetXopt();
	//Prob.CheckGradHessian(xopt);

	//LRBFGS *LRBFGSsolver2 = new LRBFGS(&Prob, &SPDX, LRBFGSsolver->GetXopt());
	//LRBFGSsolver2->LineSearch_LS = ARMIJO;
	//LRBFGSsolver2->Debug = ITERRESULT; //ITERRESULT;// 
	//LRBFGSsolver2->Max_Iteration = 500;
	//LRBFGSsolver2->Tolerance = 1e-5;
	//LRBFGSsolver2->Accuracy = 1e-4;
	//LRBFGSsolver2->Finalstepsize = 1;
	//LRBFGSsolver2->CheckParams();
	//LRBFGSsolver2->Run();
	//realdp *dists = LRBFGSsolver2->GetdistSeries();
	//ForDebug::Print("Dist:", dists, LRBFGSsolver2->GetlengthSeries());
	//delete LRBFGSsolver2;
	delete LRBFGSsolver;

	delete[] Ls;
};

void testSPDMeanMore(void)
{
	/*Randomly generate a point on the SPD manifold*/
	integer n = 3, num = 3;


	// Define the manifold
	SPDManifold Domain(n);
	Domain.SetHasHHR(false); /*set whether the manifold uses the idea in [HGA2015, Section 4.3] or not*/
	//Domain.ChooseSPDParamsSet2();

	realdp *Ls = new realdp[n * n * num + n * n];
	realdp *tmp = Ls + n * n * num;
	integer info;

	SPDVariable SPDX(n);
	realdp *initialX = SPDX.ObtainWriteEntireData();
	for (integer j = 0; j < n * n; j++)
		tmp[j] = genrandnormal();
	gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, initialX, &n);

	for (integer i = 0; i < num; i++)
	{
		for (integer j = 0; j < n * n; j++)
			tmp[j] = genrandnormal();

		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Ls + i * n * n, &n);


		potrf_(GLOBAL::L, &n, Ls + i * n * n, &n, &info);
		if (info != 0)
		{
			printf("Warning: TestSPDMean Cholesky decomposition fails with info:%d!\n", info);
		}
		for (integer j = 0; j < n; j++)
			for (integer k = j + 1; k < n; k++)
				Ls[j + k * n + i * n * n] = 0;
	}

	//MajMinAlgo(Ls, n, num, initialX, nullptr);

	// Define the problem
	SPDMean Prob(Ls, n, num);
	/*The domain of the problem is a SPD manifold*/
	Prob.SetDomain(&Domain);
	
	//Prob.CheckGradHessian(&SPDX);

	/*Output the parameters of the domain manifold*/
	Domain.CheckParams();

	/*Check the correctness of the manifold operations*/
	//Domain.CheckIntrExtr(&SPDX);
	//Domain.CheckRetraction(&SPDX);
	//Domain.CheckDiffRetraction(&SPDX);
	//Domain.CheckLockingCondition(&SPDX);
	//Domain.CheckcoTangentVector(&SPDX);
	//Domain.CheckIsometryofVectorTransport(&SPDX);
	//Domain.CheckIsometryofInvVectorTransport(&SPDX);
	//Domain.CheckVecTranComposeInverseVecTran(&SPDX);
	//Domain.CheckTranHInvTran(&SPDX);
	//Domain.CheckHaddScaledRank1OPE(&SPDX);

	//// test LRTRSR1
	//printf("********************************Test Geometric mean in LRTRSR1*************************************\n");
	//LRTRSR1 *LRTRSR1solver = new LRTRSR1(&Prob, &SPDX);
	//LRTRSR1solver->Debug = ITERRESULT;// 
	//LRTRSR1solver->Max_Iteration = 40;
	//LRTRSR1solver->Tolerance = 1e-10;
	//LRTRSR1solver->OutputGap = 100;
	//LRTRSR1solver->Shrinked_tau = 1.0 / 16;
	//LRTRSR1solver->Magnified_tau = 4;
	//LRTRSR1solver->LengthSY = 8;
	//LRTRSR1solver->Accuracy = 1e-4;
	////LRTRSR1solver->Shrinked_tau = 0.5;
	////LRTRSR1solver->CheckParams();
	//LRTRSR1solver->Run();
	//delete LRTRSR1solver;

	// test LRBFGS
	printf("********************************Test Geometric mean in LRBFGS*************************************\n");
	LRBFGS *LRBFGSsolver = new LRBFGS(&Prob, &SPDX);
	LRBFGSsolver->LineSearch_LS = ARMIJO;
	LRBFGSsolver->Debug = ITERRESULT;// 
	LRBFGSsolver->Max_Iteration = 40;
	LRBFGSsolver->OutputGap = 1;
	LRBFGSsolver->Tolerance = static_cast<realdp> (1e-10);
	//LRBFGSsolver->Accuracy = 1e-4;
	LRBFGSsolver->Finalstepsize = 1;
	LRBFGSsolver->Num_pre_funs = 10;
	LRBFGSsolver->PreFunsAccuracy = static_cast<realdp> (1e-3);
	LRBFGSsolver->LengthSY = 4;
	LRBFGSsolver->InitSteptype = ONESTEP;
	LRBFGSsolver->CheckParams();
	LRBFGSsolver->Run();

	//LRBFGSsolver->GetXopt()->Print("Xopt:");//---
	
	//// Check gradient and Hessian
	//Prob.CheckGradHessian(&SPDX);
	//const Variable *xopt = LRBFGSsolver->GetXopt();
	//Prob.CheckGradHessian(xopt);

	//LRBFGS *LRBFGSsolver2 = new LRBFGS(&Prob, &SPDX, LRBFGSsolver->GetXopt());
	//LRBFGSsolver2->LineSearch_LS = ARMIJO;
	//LRBFGSsolver2->Debug = ITERRESULT; //ITERRESULT;// 
	//LRBFGSsolver2->Max_Iteration = 500;
	//LRBFGSsolver2->Tolerance = 1e-5;
	//LRBFGSsolver2->Accuracy = 1e-4;
	//LRBFGSsolver2->Finalstepsize = 1;
	//LRBFGSsolver2->CheckParams();
	//LRBFGSsolver2->Run();
	//realdp *dists = LRBFGSsolver2->GetdistSeries();
	//ForDebug::Print("Dist:", dists, LRBFGSsolver2->GetlengthSeries());
	//delete LRBFGSsolver2;
	delete LRBFGSsolver;

	//// test LRTRSR1
	//printf("********************************Test Geometric mean in LRTRSR1*************************************\n");
	//RTRSR1 *RTRSR1solver = new RTRSR1(&Prob, &SPDX);
	//RTRSR1solver->Debug = ITERRESULT;// 
	//RTRSR1solver->Max_Iteration = 40;
	//RTRSR1solver->Tolerance = 1e-10;
	//RTRSR1solver->OutputGap = 100;
	//RTRSR1solver->Shrinked_tau = 1.0 / 16;
	//RTRSR1solver->Magnified_tau = 4;
	//RTRSR1solver->LengthSY = 8;
	//RTRSR1solver->Accuracy = 1e-6;
	////RTRSR1solver->Shrinked_tau = 0.5;
	////RTRSR1solver->CheckParams();
	//RTRSR1solver->Run();
	//delete RTRSR1solver;

	delete[] Ls;
};

#ifdef MATLAB_MEX_FILE

std::map<integer *, integer> *CheckMemoryDeleted;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs < 5)
	{
		mexErrMsgTxt("The number of arguments should be at least five.\n");
	}
	double *Lss, *X, *Xopt;
	integer n, N, HasHHR, metric;
	Lss = mxGetPr(prhs[0]);
	X = mxGetPr(prhs[1]);
	n = mxGetM(prhs[1]);
	const mwSize *ptrdims = mxGetDimensions(prhs[0]);
	if (mxGetNumberOfDimensions(prhs[0]) == 2)
		N = 1;
	else
		N = ptrdims[2];
	realdp *Ls = new realdp[n * n * N];
	for (integer i = 0; i < n * n * N; i++)
		Ls[i] = Lss[i];
	HasHHR = static_cast<integer> (mxGetScalar(prhs[2]));
	metric = static_cast<integer> (mxGetScalar(prhs[4]));

	if (ptrdims[1] != n || ptrdims[0] != n)
	{
		mexErrMsgTxt("The size of matrix C is not correct.\n");
	}
	if (mxGetM(prhs[1]) != n || mxGetN(prhs[1]) != n)
	{
		mexErrMsgTxt("The size of the initial X is not correct!\n");
	}

	genrandseed(0);

	CheckMemoryDeleted = new std::map<integer *, integer>;
	//testStieSoftICA(Cs, n, p, N, X, Xopt);

	// Obtain an initial iterate from the input
	SPDVariable *SPDX = new SPDVariable(n);
	realdp *initialX = SPDX->ObtainWriteEntireData();
    
	for (integer i = 0; i < n * n; i++)
	{
		initialX[i] = X[i];
	}

	SPDVariable *SPDsoln = nullptr;
	if (nrhs >= 6)
	{
		double *soln = mxGetPr(prhs[5]); /*soln: n by p*/
		SPDsoln = new SPDVariable(n);
		realdp *SPDsolnptr = SPDsoln->ObtainWriteEntireData();
		for (integer i = 0; i < n * n; i++)
		{
			SPDsolnptr[i] = soln[i];
		}
	}

	// Define the manifold
	SPDManifold Domain(n);

	if (metric == 1)
		Domain.ChooseSPDParamsSet1();
	else
		Domain.ChooseSPDParamsSet2();

	// Define the SPDMean problem
	SPDMean Prob(Ls, n, N);
	Prob.SetDomain(&Domain);
	Domain.SetHasHHR((HasHHR != 0));
	ParseSolverParamsAndOptimizing(prhs[3], &Prob, SPDX, SPDsoln, plhs);
	//Domain.CheckParams();
	std::map<integer *, integer>::iterator iter = CheckMemoryDeleted->begin();
	for (iter = CheckMemoryDeleted->begin(); iter != CheckMemoryDeleted->end(); iter++)
	{
		if (iter->second != 1)
			printf("Global address: %p, sharedtimes: %d\n", iter->first, iter->second);
	}
	delete CheckMemoryDeleted;
	delete SPDX;
	delete SPDsoln;
	delete[] Ls;
	return;
}

#endif
