//  Created by Xinru Yuan on 10/3/16.
//  Copyright 2016 Xinru Yuan. All rights reserved.

/*Help to debug the code*/
// #include "ForDebug.h"
// #include <iostream>
// #include <ctime>
// #include "Timer.h"
// #include "def.h"
// #include "MyMatrix.h"
// #include "SPDMean.h"
// #include "SolversLS.h"

#include "TestSPDMeanL1EW.h"

using namespace ROPTLIB;

realdp dis3(realdp *X, integer n);
realdp distwo3(realdp *X, realdp *Y, integer n);

#if !defined(MATLAB_MEX_FILE) && defined(TESTSPDMEANL1EW)

int main()
{
    testSPDMeanL1EW();
    
#ifdef _WIN64
#ifdef _DEBUG
    _CrtDumpMemoryLeaks();
#endif
#endif
    return 0;
}

#endif


void testSPDMeanL1EW(void)
{
    printf("============= Median computation by EW method =============\n");
    // choose a random seed
    unsigned tt = (unsigned)time(NULL);
    tt = 0;
//	genrandseed(tt);
    
    /*Randomly generate a point on the SPD manifold*/
    integer n = 3, num = 3;
    realdp *initial_X = new realdp[n * n];
    for (integer i = 0; i < n; i++)
    {
        for (integer j = 0; j < n; j++)
        {
            initial_X[i + j * n] = 0;
        }
        initial_X[i + i * n] = 1;
    }
    
    
    integer N = n, length = n * n, info;
    integer maxiter = 500;
    
    
    realdp *Ls = new realdp[n * n * num];
    realdp *Rstmp = new realdp[n * n];
    
    for (integer i = 0; i < num; i++)
    {
        for (integer j = 0; j < n * n; j++)
            Rstmp[j] = genrandnormal();// genrand_gaussian(); genrandnormal();
        gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, Rstmp, &n, Rstmp, &n, &GLOBAL::DZERO, Ls + i * n * n, &n);
        
        
        potrf_(GLOBAL::L, &n, Ls + n * n * i, &n, &info);
        for (integer j = 0; j < n; j++)
            for (integer k = j + 1; k < n; k++)
                Ls[j + k * n + i * n * n] = 0;
        
    }
    
    
    unsigned long starttime = getTickCount();
    
    realdp minval, negone = -1.0, t;
    realdp ComTime;
    integer idx;
    
    realdp *X1 = new realdp[n * n];
    copy_(&length, initial_X, &GLOBAL::IONE, X1, &GLOBAL::IONE);
    
    realdp *timeSeries = new realdp[maxiter];
    realdp *funsSeries = new realdp[maxiter];
    realdp *gradSeries = new realdp[maxiter];
    realdp *disSeries = new realdp[maxiter];
    
    realdp *logLXL = new realdp[n * n * num];
    realdp *coef = new realdp[num];
    realdp *eta = new realdp[n * n];
    realdp *Lx = new realdp[n * n];
    realdp *tmp = new realdp[n * n];
    realdp *X2 = new realdp[n * n];
    
    
// Start the loop
    for (integer iter = 0; iter < maxiter; iter++)
    {
//         Cholesky factorization of X1, and store in Lx
        copy_(&length, X1, &GLOBAL::IONE, Lx, &GLOBAL::IONE);
        
        potrf_(GLOBAL::L, &n, Lx, &n, &info);
        for (integer j = 0; j < n; j++)
            for (integer k = j + 1; k < n; k++)
                Lx[j + k * n] = 0;
        
//   --------------  distance from A_i to X_k  --------------
        for (integer i = 0; i < num; i++)
        {
            copy_(&length, Lx, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
            /*Solve the linear system Li X = Lx, i.e., X = Li^{-1} Lx. The solution X is stored in LiiLx.
             * Note that Li is a lower triangular matrix.
             * Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
            trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Ls + n * n * i, &N, tmp, &N, &info);
            if (info != 0)
            {
                printf("The cholesky decompsotion in SPDMean::f failed with info:%d!\n", info);
            }
            gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, tmp, &N, tmp, &N, &GLOBAL::DZERO, logLXL + n * n * i, &N);
            Matrix MMt(logLXL + n * n * i, n, n);
            Matrix::LogSymmetricM(GLOBAL::L, MMt, MMt);
            coef[i] = nrm2_(&length, logLXL + n * n * i, &GLOBAL::IONE);
        }
        
        
        minval = coef[0];
        idx = 0;
        for (integer i = 1; i < num; i++)
        {
            if(coef[i] < minval)
            {
                minval = coef[i];
                idx = i;
            }
        }
        
        
        funsSeries[iter] = 0;
        for (integer i = 0; i < num; i++) funsSeries[iter] = funsSeries[iter] + coef[i];      
        
//          --------------  compute eta  --------------
//          logLXL * L^T -> tmp
        gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, logLXL + n * n * idx, &N, Ls + n * n * idx, &N, &GLOBAL::DZERO, tmp, &N);
        
//          L^{-T} * tmp -> tmp, i.e.,  L^{-T} * logLXL * L^T -> tmp
        trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, Ls + n * n * idx, &N, tmp, &N, &info);
        
        if (info != 0)
        {
            printf("The cholesky decompsotion in SPDMean::RieGrad failed with info:%d!\n", info);
        }
        gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, X1, &N, tmp, &N, &GLOBAL::DZERO, eta, &N);
        
        
//      eta -> -eta
        scal_(&length, &negone, eta, &GLOBAL::IONE);
        
//          --------------  update next iterate  --------------
        t = static_cast<realdp> (1.0/(1.0 + iter));
        scal_(&length, &t, eta, &GLOBAL::IONE);
        
//      Cholesky factorization of eta, and stored in eta
        potrf_(GLOBAL::L, &n, eta, &n, &info);
        for (integer j = 0; j < n; j++)
            for (integer k = j + 1; k < n; k++)
                eta[j + k * n] = 0;
        
        trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Ls + n * n * idx, &N, eta, &N, &info);
        if (info != 0)
        {
            printf("The cholesky decompsotion in SPDMean::f failed with info:%d!\n", info);
        }
        gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, eta, &N, eta, &N, &GLOBAL::DZERO, X2, &N);
        Matrix MMt(X2, n, n);
        Matrix::ExpSymmetricM(GLOBAL::L, MMt, MMt);
        gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, Lx, &N, X2, &N, &GLOBAL::DONE, X2, &N);
        gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, X2, &N, Lx, &N, &GLOBAL::DONE, X2, &N);
        
        //         update X1 <- X2
        copy_(&length, X2, &GLOBAL::IONE, X1, &GLOBAL::IONE);
        timeSeries[iter] = static_cast<realdp>(getTickCount() - starttime) / CLK_PS; // record the time for each iteration
        
    }
    
    ComTime = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
    
    delete[] timeSeries;
    delete[] disSeries;
    delete[] gradSeries;
    delete[] funsSeries;
    
    delete[] logLXL;
    delete[] coef;
    delete[] eta;
    delete[] Lx;
    delete[] tmp;
    delete[] X1;
    delete[] X2;
    delete[] Rstmp;
    
};




#ifdef MATLAB_MEX_FILE

std::map<integer *, integer> *CheckMemoryDeleted;

/*This function checks the number and formats of input parameters.
 * nlhs: the number of output in mxArray format
 * plhs: the output objects in mxArray format*
 * nrhs: the number of input in mxArray format
 * prhs: the input objects in mxArray format */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    
    printf("============= Median computation by EW method =============\n");
    // deal with input
    realdp *Ls, *initial_X, *Xtrue;
    integer n, num, maxiter, Debug;
    realdp TOL, stepsize, alpha;
    Ls = mxGetPr(prhs[0]);
    initial_X = mxGetPr(prhs[1]);
    Xtrue = mxGetPr(prhs[2]);
    n = static_cast<integer>(mxGetScalar(prhs[3]));
    num = static_cast<integer>(mxGetScalar(prhs[4]));
    maxiter = static_cast<integer>(mxGetScalar(prhs[5]));
    stepsize = static_cast<double>(mxGetScalar(prhs[6]));
    TOL = static_cast<double>(mxGetScalar(prhs[7]));
    Debug = static_cast<integer>(mxGetScalar(prhs[8]));
    
    
//     time start
    unsigned long starttime = getTickCount();
    CheckMemoryDeleted = new std::map<integer *, integer>;
    
    
    integer length = n * n, N = n, info, enditer;
    realdp minval, negone = -1.0, coefa;
    realdp ComTime, err_g, sigma;
    integer idx;
    
    realdp *X1 = new realdp[n * n];
    copy_(&length, initial_X, &GLOBAL::IONE, X1, &GLOBAL::IONE);
    
    realdp *timeSeries = new realdp[maxiter];
    realdp *funsSeries = new realdp[maxiter];
    realdp *gradSeries = new realdp[maxiter];
    realdp *disSeries = new realdp[maxiter];
    
    realdp *logLXL = new realdp[n * n * num];
    realdp *coef = new realdp[num];
    realdp *eta = new realdp[n * n];
    realdp *Lx = new realdp[n * n];
    realdp *tmp = new realdp[n * n];
    realdp *X2 = new realdp[n * n];
    
    
    realdp *In = new realdp[n * n]; // make identity matrix
    for (integer i = 0; i < length; i++) In[i] = 0.0;
    for (integer i = 0; i < n; i++) In[i*n+i] = 1.0;
    
    
    if(Debug >= 3)
    {
        // compute distance between current iterate and identity matrix
        disSeries[0] = distwo3(X1, Xtrue, n);
        gradSeries[0] = 0.0;
    }
    
// Start the loop
    for (integer iter = 0; iter < maxiter; iter++)
    {
//         Cholesky factorization of X1, and store in Lx
        copy_(&length, X1, &GLOBAL::IONE, Lx, &GLOBAL::IONE);
        
        potrf_(GLOBAL::L, &n, Lx, &n, &info);
        for (integer j = 0; j < n; j++)
            for (integer k = j + 1; k < n; k++)
                Lx[j + k * n] = 0;
        
                
//   --------------  distance from A_i to X_k  --------------
        for (integer i = 0; i < num; i++)
        {
            copy_(&length, Lx, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
            /*Solve the linear system Li X = Lx, i.e., X = Li^{-1} Lx. The solution X is stored in LiiLx.
             * Note that Li is a lower triangular matrix.
             * Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
            trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Ls + n * n * i, &N, tmp, &N, &info);
            if (info != 0)
            {
                printf("The cholesky decompsotion in SPDMean::f failed with info:%d!\n", info);
            }
            gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, tmp, &N, tmp, &N, &GLOBAL::DZERO, logLXL + n * n * i, &N);
            Matrix MMt(logLXL + n * n * i, n, n);
            Matrix::LogSymmetricM(GLOBAL::L, MMt, MMt);
            coef[i] = nrm2_(&length, logLXL + n * n * i, &GLOBAL::IONE);
        }
        
//   --------------  check if dist(Xk, Ai) == 0: find the smallest dist(Xk, Ai)  --------------
        minval = coef[0];
        idx = 0;
        for (integer i = 1; i < num; i++)
        {
            if(coef[i] < minval)
            {
                minval = coef[i];
                idx = i;
            }
        }
        
        
//          --------------  compute eta  --------------
        
        if (minval > 1e-16)
        {
            // compute sigma = sum 1/dist(Xk, Ai)
            sigma = 0.0;
            for (integer i = 0; i < num; i++)
            {
                sigma = sigma + 1.0/coef[i];
            }
            sigma = 1.0/sigma;
            
            for (integer i = 0; i < n * n; i++) eta[i] = 0;
            
            
            for (integer i = 0; i < num; i++)
            {
                /*tmp <-- log(Li^{-1} X Li^{-T}) Li^T */
                gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, logLXL + n * n * i, &N,
                        Ls + n * n * i, &N, &GLOBAL::DZERO, tmp, &N);
                
                /*Solve the linear system Li^T X = tmp, i.e., X = Li^{-T} log(Li^{-1} X Li^{-T}) Li^T. The solution X is stored in tmp.*/
                trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, Ls + n * n * i, &N, tmp, &N, &info);
                if (info != 0)
                {
                    std::cout << "The cholesky decompsotion in SPDMean::RieGrad failed with info:" << info << "!" << std::endl;
                }
                coefa = 1.0/coef[i];
                gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &coefa, X1, &N, tmp, &N, &GLOBAL::DONE, eta, &N);
            }
            scal_(&length, &sigma, eta, &GLOBAL::IONE);
            
        }
        else // eta_k = Log(Aj)
        {
//          logLXL * Li^T -> tmp
            gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, logLXL + n * n * idx, &N, Ls + n * n * idx, &N, &GLOBAL::DZERO, tmp, &N);
            
//          Li^{-T} * tmp -> tmp, i.e.,  Li^{-T} * logLXL * Li^T -> tmp
            trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, Ls + n * n * idx, &N, tmp, &N, &info);
            
            if (info != 0)
            {
                printf("The cholesky decompsotion in SPDMean::RieGrad failed with info:%d!\n", info);
            }
//          X * tmp -> eta
            gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, X1, &N, tmp, &N, &GLOBAL::DZERO, eta, &N);
        }

        
        funsSeries[iter] = 0;
        for (integer i = 0; i < num; i++) funsSeries[iter] = funsSeries[iter] + coef[i];
        

//      eta -> -eta
        scal_(&length, &negone, eta, &GLOBAL::IONE);
        
//          --------------  update next iterate  --------------
        scal_(&length, &stepsize, eta, &GLOBAL::IONE);
        
//      Lx^{-1} * eta -> eta
        trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Lx, &N, eta, &N, &info);
//      eta -> eta^T
        copy_(&length, eta, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        gemm_(GLOBAL::T, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, tmp, &N, In, &N, &GLOBAL::DZERO, eta, &N);
        trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Lx, &N, eta, &N, &info);
        
        copy_(&length, eta, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        gemm_(GLOBAL::T, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, tmp, &N, In, &N, &GLOBAL::DZERO, eta, &N);        
        
        copy_(&length, eta, &GLOBAL::IONE, X2, &GLOBAL::IONE);
        
        Matrix MMt(X2, n, n);
        Matrix::ExpSymmetricM(GLOBAL::L, MMt, MMt);
        
        copy_(&length, X2, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, Lx, &N, tmp, &N, &GLOBAL::DZERO, X2, &N);
        
        copy_(&length, X2, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, tmp, &N, Lx, &N, &GLOBAL::DZERO, X2, &N);
        
        //         update X1 <- X2
        copy_(&length, X2, &GLOBAL::IONE, X1, &GLOBAL::IONE);        
        
        if(Debug >= 3 && iter > 0)
        {
            disSeries[iter] = distwo3(X2, Xtrue, n);
            gradSeries[iter] = 0.0;
        }
        
        timeSeries[iter] = static_cast<double>(getTickCount() - starttime) / CLK_PS; // record the time for each iteration
        
        // check stopping criteria
        if(iter == 0)
        {
            err_g = std::abs(funsSeries[0]);
        }
        else
        {
            err_g = std::abs(funsSeries[iter] - funsSeries[iter - 1])/funsSeries[0];
        }
        enditer = iter;
        
//         if(err_g < TOL)
//         {
//             enditer = iter;
//             printf("error: %.4f\n", err_g);
//             printf("Tolerance reached!");
//             break;
//         }
    }
    
    
    ComTime = static_cast<double>(getTickCount() - starttime) / CLK_PS;
    
    
// create output
    plhs[0] = mxCreateDoubleMatrix(maxiter, 1, mxREAL); // computation time
    plhs[1] = mxCreateDoubleScalar(ComTime); //  total computational time
    plhs[2] = mxCreateDoubleScalar(enditer); // number of iterations
    
    plhs[3] = mxCreateDoubleMatrix(length, 1, mxREAL); // X
    realdp *output3 = mxGetPr(plhs[3]);
    copy_(&length, X1, &GLOBAL::IONE, output3, &GLOBAL::IONE);
    
    plhs[4] = mxCreateDoubleMatrix(maxiter, 1, mxREAL); // distance
    plhs[5] = mxCreateDoubleMatrix(maxiter, 1, mxREAL); // function
    plhs[6] = mxCreateDoubleMatrix(maxiter, 1, mxREAL); // grad
    plhs[7] = mxCreateDoubleScalar(err_g);
    
    realdp *plhstime = mxGetPr(plhs[0]);
    realdp *plhsdis = mxGetPr(plhs[4]);
    realdp *plhsfun = mxGetPr(plhs[5]);
    realdp *plhsgrad = mxGetPr(plhs[6]);
    
    for(integer i = 0; i < maxiter; i++)
    {
        plhstime[i] = timeSeries[i];
        plhsdis[i] = disSeries[i];
        plhsfun[i] = funsSeries[i];
        plhsgrad[i] = gradSeries[i];
//         printf("there: %.4e, %.4e, %.4e, %.4e\n", timeSeries[i], disSeries[i], stepsizeSeries[i], truestepsizeSeries[i]);
    }
    
    delete[] timeSeries;
    delete[] disSeries;
    delete[] gradSeries;
    delete[] funsSeries;
    
    delete[] logLXL;
    delete[] coef;
    delete[] eta;
    delete[] Lx;
    delete[] tmp;
    delete[] X1;
    delete[] X2;
    delete[] In;
    
    
    std::map<integer *, integer>::iterator iteriter = CheckMemoryDeleted->begin();
    for (iteriter = CheckMemoryDeleted->begin(); iteriter != CheckMemoryDeleted->end(); iteriter++)
    {
        if (iteriter->second != 1)
            printf("Global address: %d", iteriter->first);
        printf(", shared times: %d\n", iteriter->second);
    }
    delete CheckMemoryDeleted;
    return;
    
}

#endif




realdp dis3(realdp *X, integer n)
{
    // compute distance between current iterate and the identity matrix
    realdp *eigenvaluesXnew = new realdp[n + n * n];
    realdp *eigenvectorsXnew = eigenvaluesXnew + n;
    realdp *tmp = new realdp[n * n];
    
    realdp disSeries;
    integer N = n * n;
    
    copy_(&N, X, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
    Matrix Enew(eigenvaluesXnew, n, 1), Vnew(eigenvectorsXnew, n, n);
    Matrix MMtnew(tmp, n, n);
    Matrix::EigenSymmetricM(GLOBAL::L, MMtnew, Enew, Vnew);
    disSeries = 0.0;
    for (integer ii = 0; ii < n; ii++)
    {
        disSeries = disSeries+log(eigenvaluesXnew[ii])*log(eigenvaluesXnew[ii]);
    }
    disSeries = sqrt(disSeries);
    delete[] eigenvaluesXnew;
    delete[] tmp;
    return disSeries;
    
};




realdp distwo3(realdp *X, realdp *Y, integer n)
{
    realdp *Xold = new realdp[n * n];
    realdp *Xnew = new realdp[n * n];
    realdp *tmp = new realdp[n * n];
    realdp *eigenvaluesXold = new realdp[n + n * n];
    realdp *eigenvectorsXold = eigenvaluesXold + n;
    realdp *eigenvaluesXnew = new realdp[n + n * n];
    realdp *eigenvectorsXnew = eigenvaluesXnew + n;
    integer N = n * n;
    integer info;
    realdp truestepsize;
    
    copy_(&N, X, &GLOBAL::IONE, Xold, &GLOBAL::IONE);
    copy_(&N, Y, &GLOBAL::IONE, Xnew, &GLOBAL::IONE);
    
    // get cholesky of Xnew and Xold
    potrf_(GLOBAL::L, &n, Xnew, &n, &info);
    potrf_(GLOBAL::L, &n, Xold, &n, &info);
    for (integer j = 0; j < n; j++)
    {
        for (integer k = j + 1; k < n; k++)
        {
            Xnew[j + k * n] = 0; //manually make the upper triangle 0: X0 = X0 * X0^T
            Xold[j + k * n] = 0;
        }
    }
    // Solve linear system Xnew * W = Xold, i.e. W = Xnew^{-1}*Xold -> Xold
    trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &n, &n, Xnew, &n, Xold, &n, &info);
    
    // Compute Xold <- Xold * Xold'
    copy_(&N, Xold, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
    gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Xold, &n);
    
    // eigenvalue decomposition of Xold
    Matrix Eold(eigenvaluesXold, n, 1), Vold(eigenvectorsXold, n, n);
    Matrix MMtold(Xold, n, n);
    Matrix::EigenSymmetricM(GLOBAL::L, MMtold, Eold, Vold);
    
    truestepsize = 0.0;
    for (integer ii = 0; ii < n; ii++)
    {
        truestepsize = truestepsize+log(eigenvaluesXold[ii])*log(eigenvaluesXold[ii]);
    }
    truestepsize = sqrt(truestepsize);
    
    delete[] Xold;
    delete[] Xnew;
    delete[] tmp;
    delete[] eigenvaluesXold;
    delete[] eigenvaluesXnew;
    
    
    return truestepsize;
};







