
/*Help to debug the code*/
// #include "ForDebug.h"
// #include <iostream>
// #include <ctime>
// #include "Timer.h"
// #include "def.h"
// #include "MyMatrix.h"
// #include "SPDMean.h"
// #include "SolversLS.h"

#include "TestSPDMeanMajorization.h"

using namespace ROPTLIB;
// using namespace std;

realdp distwo6(realdp *X, realdp *Y, integer n);

#if !defined(MATLAB_MEX_FILE) && defined(TESTSPDMEANRSD)

int main()
{
	testSPDMeanMajorization();
    
#ifdef _WIN64
#ifdef _DEBUG
    _CrtDumpMemoryLeaks();
#endif
#endif
    return 0;
}

#endif

void testSPDMeanMajorization(void)
{
	/*Randomly generate a point on the SPD manifold*/
	integer n = 3, num = 5;
	SPDVariable SPDX(n);
	realdp *initialX = SPDX.ObtainWriteEntireData();
	for (integer i = 0; i < n; i++)
	{
		for (integer j = 0; j < n; j++)
		{
			initialX[i + j * n] = 0;
		}
		initialX[i + i * n] = 1;
	}

	// Define the manifold
	SPDManifold Domain(n);
	Domain.SetHasHHR(false); /*set whether the manifold uses the idea in [HGA2015, Section 4.3] or not*/

	realdp *Ls = new realdp[n * n * num + n * n];
	realdp *tmp = Ls + n * n * num;
	integer info;
	for (integer i = 0; i < num; i++)
	{
		for (integer j = 0; j < n * n; j++)
			tmp[j] = genrandnormal();

		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Ls + i * n * n, &n);


		potrf_(GLOBAL::L, &n, Ls + i * n * n, &n, &info);
		if (info != 0)
		{
			printf("Warning: TestSPDMean Cholesky decomposition fails with info:%d!\n", info);
		}
		for (integer j = 0; j < n; j++)
			for (integer k = j + 1; k < n; k++)
				Ls[j + k * n + i * n * n] = 0;
	}
	integer maxiter = 100;
	realdp *Xout = new realdp[n * n + maxiter * 2 + 2];
	realdp *timeSeries = Xout + n * n;
	realdp *distSeries = timeSeries + 1 + maxiter;
	integer totaliter = 0, Debug = 3;
	realdp ComTime = 0, tol = static_cast<realdp> (1e-17);

	MajMinAlgo(Ls, n, num, initialX, maxiter, tol, Debug, Xout, timeSeries, distSeries, totaliter, ComTime);
	delete[] Xout;
	delete[] Ls;
};

void MajMinAlgo(realdp *Ls, integer n, integer num, realdp *initX, integer maxiter, realdp tol, 
	integer Debug, realdp *Xout, realdp *timeSeries, realdp *distSeries, integer &totaliter, realdp &ComTime, realdp *Xtrue)
{
	realdp *x1 = new realdp[n * n * 2];
	realdp *Lx = x1 + n * n;
	integer length = n * n;

	unsigned long starttime = getTickCount();
	copy_(&length, initX, &GLOBAL::IONE, x1, &GLOBAL::IONE);

	/*iterations*/
	realdp *f1X = new realdp[n * n * 7 + 2 * n];
	realdp *f2X = f1X + n * n;
	realdp *XinvgX = f2X + n * n;
	realdp *V = XinvgX + n * n;
	realdp *D = V + n * n;
	realdp *tmp = D + n;
	realdp *tmp2 = tmp + n * n;
	realdp *V2 = tmp2 + n * n;
	realdp *D2 = V2 + n * n;

	integer iter = 0;
	Computef1Xf2XXinvgX(Ls, n, num, x1, XinvgX, f1X, f2X);
	realdp ngf0 = sqrt(dot_(&length, XinvgX, &GLOBAL::IONE, XinvgX, &GLOBAL::IONE));
	realdp ngf = ngf0;
	if (Debug >= 2)
	{
		distSeries[iter] = ((Xtrue == nullptr) ? 0 : distwo6(x1, Xtrue, n));
		timeSeries[iter] = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
	}
	while (iter < maxiter && ngf / ngf0 > tol)
	{
		//ForDebug::Print("XinvgX:", XinvgX, n, n);//---
		//ForDebug::Print("f1X:", f1X, n, n);//---
		//ForDebug::Print("f2X:", f2X, n, n);//---
		Matrix M_f2X(f2X, n, n), M_D(D, n, 1), M_V(V, n, n);
		Matrix::EigenSymmetricM(GLOBAL::L, M_f2X, M_D, M_V);
		for (integer i = 0; i < n; i++)
			D[i] = std::sqrt(D[i]);

		gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, f1X, &n, V, &n, &GLOBAL::DZERO, tmp, &n);
		gemm_(GLOBAL::T, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, V, &n, tmp, &n, &GLOBAL::DZERO, tmp2, &n);
		for (integer i = 0; i < n; i++)
		{
			scal_(&n, &D[i], tmp2 + i * n, &GLOBAL::IONE);
			scal_(&n, &D[i], tmp2 + i, &n);
		}
		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp2, &n, V, &n, &GLOBAL::DZERO, tmp, &n);
		gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, V, &n, tmp, &n, &GLOBAL::DZERO, tmp2, &n); /*now, tmp2 = f2X^{-1/2} * f1X * f2X^{-1/2} */

		Matrix M_f2f1f2(tmp2, n, n), M_D2(D2, n, 1), M_V2(V2, n, n);
		Matrix::EigenSymmetricM(GLOBAL::L, M_f2f1f2, M_D2, M_V2);

		for (integer i = 0; i < n; i++)
			D2[i] = static_cast<realdp> (1) / std::sqrt(D2[i]);

		/*x2 = V * D * V^T * V2 * D2 * V2^T * V * D* V^T */
		copy_(&length, V2, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
		for (integer i = 0; i < n; i++)
			scal_(&n, &D2[i], tmp + i * n, &GLOBAL::IONE);
		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, V2, &n, &GLOBAL::DZERO, tmp2, &n);
		gemm_(GLOBAL::T, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, V, &n, tmp2, &n, &GLOBAL::DZERO, tmp, &n);
		gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, tmp, &n, V, &n, &GLOBAL::DZERO, tmp2, &n);
		for (integer i = 0; i < n; i++)
		{
			scal_(&n, &D[i], tmp2 + i * n, &GLOBAL::IONE);
			scal_(&n, &D[i], tmp2 + i, &n);
		}
		gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, V, &n, tmp2, &n, &GLOBAL::DZERO, tmp, &n);
		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, V, &n, &GLOBAL::DZERO, x1, &n);
		iter++;
		//ForDebug::Print("x1:", x1, n, n);//---
		Computef1Xf2XXinvgX(Ls, n, num, x1, XinvgX, f1X, f2X);
		ngf = sqrt(dot_(&length, XinvgX, &GLOBAL::IONE, XinvgX, &GLOBAL::IONE));

		if (Debug >= 2)
		{
			distSeries[iter] = ((Xtrue == nullptr) ? 0 : distwo6(x1, Xtrue, n));
			timeSeries[iter] = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
		}
		if (Debug >= 3)
			printf("Iter:%d,|gf|:%.3e,time:%.2e\n", iter, ngf, timeSeries[iter]);
	}
	if(Xout != nullptr)
		copy_(&length, x1, &GLOBAL::IONE, Xout, &GLOBAL::IONE);
	ComTime = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
	totaliter = iter + 1;
	if (Debug >= 1)
		printf("Iter:%d,|gf|:%.3e,|gf|/|gf0|:%.3e,time:%.2e\n", iter, ngf, ngf / ngf0, ComTime);

	delete[] f1X;
	delete[] x1;
};

void Computef1Xf2XXinvgX(realdp *Ls, integer n, integer num, realdp *x, realdp *XinvgX, realdp *f1X, realdp *f2X)
{
	for (integer i = 0; i < n * n; i++)
	{
		XinvgX[i] = 0;
		f1X[i] = 0;
		f2X[i] = 0;
	}

	realdp *Lx = new realdp[n * n];
	integer length = n * n;
	/*iterations*/
	for (integer i = 0; i < n; i++)
	{
		for (integer j = i; j < n; j++)
		{
			Lx[i + j * n] = 0;
			Lx[j + i * n] = x[j + i * n];
		}
	}
	integer info;
	potrf_(GLOBAL::L, &n, Lx, &n, &info); /*Cholesky decomposition*/
	//ForDebug::Print("Lx:", Lx, n, n);//---
	realdp *V = new realdp[n * n * 8 + n];
	realdp *LXL = V + n * n;
	realdp *VD1 = LXL + n * n;
	realdp *VD2 = VD1 + n * n;
	realdp *VD3 = VD2 + n * n;
	realdp *tmp = VD3 + n * n;
	realdp *tmp2 = tmp + n * n;
	realdp *Ltmp = tmp2 + n * n;
	realdp *D = Ltmp + n * n;
	for (integer i = 0; i < num; i++)
	{
		copy_(&length, const_cast<realdp *> (Lx), &GLOBAL::IONE, Ltmp, &GLOBAL::IONE);
		/*Solve the linear system Li X = Lx, i.e., X = Li^{-1} Lx. The solution X is stored in LiiLx.
		Note that Li is a lower triangular matrix.
		Details: http://www.netlib.org/
		/explore-html/d6/d6f/trtrs_8f.html */
		trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &n, &n, Ls + n * n * i, &n, Ltmp, &n, &info);
		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, Ltmp, &n, Ltmp, &n, &GLOBAL::DZERO, LXL, &n);
		//std::cout << "i:" << i << std::endl;//---
		//ForDebug::Print("LXL:", LXL, n, n);//---
		Matrix MMt(LXL, n, n), M_D(D, n, 1), M_V(V, n, n);
		Matrix::EigenSymmetricM(GLOBAL::L, MMt, M_D, M_V);

		copy_(&length, V, &GLOBAL::IONE, VD1, &GLOBAL::IONE);
		copy_(&length, V, &GLOBAL::IONE, VD2, &GLOBAL::IONE);
		copy_(&length, V, &GLOBAL::IONE, VD3, &GLOBAL::IONE);
		for (integer i = 0; i < n; i++)
		{
			if (D[i] <= 0)
			{
				printf("Error: The matrix for MajMinAlgo is not symmetric positive definite!!\n");
				return;
			}
			realdp a = log(D[i]);
			scal_(&n, &a, VD1 + i * n, &GLOBAL::IONE);
			a = (std::sqrt(log(D[i]) * log(D[i]) + 1) + log(D[i])) / D[i];
			scal_(&n, &a, VD2 + i * n, &GLOBAL::IONE);
			a = (std::sqrt(log(D[i]) * log(D[i]) + 1) - log(D[i])) * D[i];
			scal_(&n, &a, VD3 + i * n, &GLOBAL::IONE);
		}

		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, VD1, &n, V, &n, &GLOBAL::DZERO, tmp, &n);
		//ForDebug::Print("1 tmp:", tmp, n, n);//----
		/*tmp <-- log(Li^{-1} X Li^{-T}) Li^T */
		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, Ls + n * n * i, &n, &GLOBAL::DZERO, tmp2, &n);
		trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &n, &n, Ls + n * n * i, &n, tmp2, &n, &info);
		//ForDebug::Print("2 tmp:", tmp2, n, n);//----
		axpy_(&length, &GLOBAL::DONE, tmp2, &GLOBAL::IONE, XinvgX, &GLOBAL::IONE); /*XinvgX += Li^{-T} log(Li^{-1} X Li^{-T}) Li^T*/
//		gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, x, &n, tmp2, &n, &GLOBAL::DONE, gX, &n); /*gX += X Li^{-T} log(Li^{-1} X Li^{-T}) Li^T*/

		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, VD2, &n, V, &n, &GLOBAL::DZERO, tmp, &n);
		//ForDebug::Print("3 tmp:", tmp, n, n);//----
		trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &n, &n, Ls + n * n * i, &n, tmp, &n, &info);
		//ForDebug::Print("4 tmp:", tmp, n, n);//----
		realdp tmpv = 0;
		for (integer i = 0; i < n; i++)
		{
			for (integer j = i; j < n; j++)
			{
				tmpv = tmp[i + j * n];
				tmp[i + j * n] = tmp[j + i * n];
				tmp[j + i * n] = tmpv;
			}
		}
		trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &n, &n, Ls + n * n * i, &n, tmp, &n, &info);
		//ForDebug::Print("5 tmp:", tmp, n, n);//----
		axpy_(&length, &GLOBAL::DONE, tmp, &GLOBAL::IONE, f1X, &GLOBAL::IONE);

		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, VD3, &n, V, &n, &GLOBAL::DZERO, tmp, &n);
		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, Ls + n * n * i, &n, &GLOBAL::DZERO, tmp2, &n); /*V is used as a temporary space*/
		gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, Ls + n * n * i, &n, tmp2, &n, &GLOBAL::DONE, f2X, &n);
	}
	delete[] Lx;
	delete[] V;
}


#ifdef MATLAB_MEX_FILE

std::map<integer *, integer> *CheckMemoryDeleted;

/*This function checks the number and formats of input parameters.
 * nlhs: the number of output in mxArray format
 * plhs: the output objects in mxArray format*
 * nrhs: the number of input in mxArray format
 * prhs: the input objects in mxArray format */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs < 5)
	{
		mexErrMsgTxt("The number of arguments should be at least five.\n");
	}
	double *LLs, *XX;
	integer n, k, maxiter, Debug;
	realdp tol;
	LLs = mxGetPr(prhs[0]);
	XX = mxGetPr(prhs[1]);
	n = mxGetM(prhs[1]);
	/* dimensions of input matrices */
	const mwSize *ptrdims = mxGetDimensions(prhs[0]);
	k = ptrdims[2];
	if (ptrdims[0] != n || ptrdims[1] != n || mxGetN(prhs[1]) != n)
	{
		mexErrMsgTxt("error: the dimensions of Ls and X do not match!\n");
		return;
	}
	maxiter = static_cast<integer> (mxGetScalar(prhs[2]));
	tol = mxGetScalar(prhs[3]);
	Debug = static_cast<integer> (mxGetScalar(prhs[4]));
	
	realdp *Xout = new realdp[n * n + maxiter * 2 + 2];
	realdp *timeSeries = Xout + n * n;
	realdp *distSeries = timeSeries + 1 + maxiter;
	integer totaliter = 0, length = n * n;
	realdp ComTime = 0;
    realdp *Ls = new realdp[n * n * k + n * n * 2];
    realdp *X = Ls + n * n * k;
    realdp *XXtrue = X + n * n;
    for(integer i = 0; i < n * n * k; i++)
        Ls[i] = LLs[i];
    for(integer i = 0; i < n * n; i++)
        X[i] = XX[i];

	double *Xtrue = nullptr;
	if (nrhs >= 6)
    {
		Xtrue = mxGetPr(prhs[5]);
        for(integer i = 0; i < n * n; i++)
            XXtrue[i] = Xtrue[i];
    }
	MajMinAlgo(Ls, n, k, X, maxiter, tol, Debug, Xout, timeSeries, distSeries, totaliter, ComTime, XXtrue);
	// create output
	plhs[0] = mxCreateDoubleMatrix(totaliter, 1, mxREAL); // computation time
	plhs[1] = mxCreateDoubleScalar(ComTime); //  total computational time
	plhs[2] = mxCreateDoubleScalar(totaliter); // number of iterations
	plhs[3] = mxCreateDoubleMatrix(n * n, 1, mxREAL); // X
	plhs[4] = mxCreateDoubleMatrix(totaliter, 1, mxREAL); // computation time
	double *output3 = mxGetPr(plhs[3]);
    for(integer i = 0; i < length; i++)
        output3[i] = Xout[i];
//	copy_(&length, Xout, &GLOBAL::IONE, output3, &GLOBAL::IONE);

	double *output0 = mxGetPr(plhs[0]);
	double *output4 = mxGetPr(plhs[4]);
    for(integer i = 0; i < totaliter; i++)
    {
        output0[i] = timeSeries[i];
        output4[i] = distSeries[i];
    }
//	copy_(&totaliter, timeSeries, &GLOBAL::IONE, output0, &GLOBAL::IONE);
//	copy_(&totaliter, distSeries, &GLOBAL::IONE, output4, &GLOBAL::IONE);

	delete[] Xout;
    delete[] Ls;

    return;
    
}

#endif


realdp distwo6(realdp *X, realdp *Y, integer n)
{
	realdp *Xold = new realdp[n * n];
	realdp *Xnew = new realdp[n * n];
	realdp *tmp = new realdp[n * n];
	realdp *eigenvaluesXold = new realdp[n + n * n];
	realdp *eigenvectorsXold = eigenvaluesXold + n;
	realdp *eigenvaluesXnew = new realdp[n + n * n];
	realdp *eigenvectorsXnew = eigenvaluesXnew + n;
	integer N = n * n;
	integer info;
	realdp truestepsize;

	copy_(&N, X, &GLOBAL::IONE, Xold, &GLOBAL::IONE);
	copy_(&N, Y, &GLOBAL::IONE, Xnew, &GLOBAL::IONE);

	// get cholesky of Xnew and Xold
	potrf_(GLOBAL::L, &n, Xnew, &n, &info);
	potrf_(GLOBAL::L, &n, Xold, &n, &info);
	for (integer j = 0; j < n; j++)
	{
		for (integer k = j + 1; k < n; k++)
		{
			Xnew[j + k * n] = 0; //manually make the upper triangle 0: X0 = X0 * X0^T
			Xold[j + k * n] = 0;
		}
	}
	// Solve linear system Xnew * W = Xold, i.e. W = Xnew^{-1}*Xold -> Xold
	trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &n, &n, Xnew, &n, Xold, &n, &info);

	// Compute Xold <- Xold * Xold'
	copy_(&N, Xold, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
	gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Xold, &n);

	// eigenvalue decomposition of Xold
	Matrix Eold(eigenvaluesXold, n, 1), Vold(eigenvectorsXold, n, n);
	Matrix MMtold(Xold, n, n);
	Matrix::EigenSymmetricM(GLOBAL::L, MMtold, Eold, Vold);

	truestepsize = 0.0;
	for (integer ii = 0; ii < n; ii++)
	{
		truestepsize = truestepsize + log(eigenvaluesXold[ii])*log(eigenvaluesXold[ii]);
	}
	truestepsize = sqrt(truestepsize);

	delete[] Xold;
	delete[] Xnew;
	delete[] tmp;
	delete[] eigenvaluesXold;
	delete[] eigenvaluesXnew;

	return truestepsize;
};

