//
//
//  Cleaned the code compared with version 2.
//
//  Created by Xinru Yuan on 10/3/16.
//  Copyright 2016 Xinru Yuan. All rights reserved.
//

// #include "TestSPDMeanRSD.h"
// 
// using namespace ROPTLIB;



/*Help to debug the code*/
// #include "ForDebug.h"
// #include <iostream>
// #include <ctime>
// #include "Timer.h"
// #include "def.h"
// #include "MyMatrix.h"
// #include "SPDMean.h"
// #include "SolversLS.h"

#include "TestSPDMeanRSD.h"

using namespace ROPTLIB;
// using namespace std;

realdp dis5(realdp *X, integer n);
realdp distwo5(realdp *X, realdp *Y, integer n);

#if !defined(MATLAB_MEX_FILE) && defined(TESTSPDMEANRSD)

int main()
{
     testSPDMeanRSD();
    
#ifdef _WIN64
#ifdef _DEBUG
    _CrtDumpMemoryLeaks();
#endif
#endif
    return 0;
}

#endif





void testSPDMeanRSD(void)
{
	printf("============= RSD: Using C++: needs to be implemented =============\n");
	// choose a random seed
	unsigned tt = (unsigned)time(NULL);
	tt = 0;
//	genrandseed(tt);

	/*Randomly generate a point on the SPD manifold*/
	integer n = 100, num = 4;
    realdp *initial_X = new realdp[n * n];
	for (integer i = 0; i < n; i++)
	{
		for (integer j = 0; j < n; j++)
		{
			initial_X[i + j * n] = 0;
		}
		initial_X[i + i * n] = 1;
	}

    
    integer N = n * n;
    

	realdp *Rs = new realdp[n * n * num];
	realdp *Rstmp = new realdp[n * n];
	for (integer i = 0; i < num; i++)
	{
		for (integer j = 0; j < n * n; j++)
			Rstmp[j] = genrandnormal();
		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, Rstmp, &n, Rstmp, &n, &GLOBAL::DZERO, Rs + i * n * n, &n);
	}
    delete[] Rstmp;
    delete[] Rs;
    delete[] initial_X;

};




#ifdef MATLAB_MEX_FILE

std::map<integer *, integer> *CheckMemoryDeleted;

/*This function checks the number and formats of input parameters.
 * nlhs: the number of output in mxArray format
 * plhs: the output objects in mxArray format*
 * nrhs: the number of input in mxArray format
 * prhs: the input objects in mxArray format */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    
    printf("============= RSD =============\n");
    // input
    double *RRs, *initial_X, *XXtrue;
    integer n, num, maxiter, Retraction, StepType, Debug;
    realdp TOL;
    RRs = mxGetPr(prhs[0]);
    initial_X = mxGetPr(prhs[1]);
    XXtrue = mxGetPr(prhs[2]);
    n = static_cast<integer>(mxGetScalar(prhs[3]));
    num = static_cast<integer>(mxGetScalar(prhs[4]));
    maxiter = static_cast<integer>(mxGetScalar(prhs[5]));
    TOL = static_cast<realdp>(mxGetScalar(prhs[6]));
    Retraction = static_cast<integer>(mxGetScalar(prhs[7]));
    StepType = static_cast<integer>(mxGetScalar(prhs[8]));
    Debug = static_cast<integer>(mxGetScalar(prhs[9]));
    
    
    unsigned long starttime = getTickCount();
    CheckMemoryDeleted = new std::map<integer *, integer>;
    
    
    //parameters
    realdp a, stepsize, ch, Delta, onehalf = 0.5;
    realdp ComTime;
    integer N = n * n, enditer, info;
    
    realdp *Xtrue = new realdp[N];
    for(integer i = 0; i < N; i++)
        Xtrue[i] = XXtrue[i];
    
    realdp *X1 = new realdp[N];
    for(integer i = 0; i < N; i++)
        X1[i] = initial_X[i];
//    copy_(&N, initial_X, &GLOBAL::IONE, X1, &GLOBAL::IONE);
    
    realdp *Rs = new realdp[N * num];
    integer Nnum = N * num;
    for(integer i = 0; i < Nnum; i++)
        Rs[i] = RRs[i];
//    copy_(&Nnum, RRs, &GLOBAL::IONE, Rs, &GLOBAL::IONE);
        
    
    realdp *timeSeries = new realdp[maxiter];
    realdp *gradSeries = new realdp[maxiter];
    realdp *disSeries = new realdp[maxiter];
    realdp *truestepsizeSeries = new realdp[maxiter];
    realdp *stepsizeSeries = new realdp[maxiter];


    realdp *Lx = new realdp[N]; // store the Cholesky factorization of each iterate X
    realdp *Z = new realdp[N * num];  // store R0^{-1}R
    realdp *S = new realdp[N];  // store the gradient
    realdp *LiStmp = new realdp[N];
    realdp *Y = new realdp[N];
    realdp *tmp = new realdp[N];
    realdp *eigenvalues = new realdp[n + 2 * N];
    realdp *eigenvectors = eigenvalues + n;
    realdp *eigenvectorsD = eigenvectors + N;
    Matrix E(eigenvalues, n, 1), V(eigenvectors, n, n), VD(eigenvectorsD, n, n);

    
    //make identity matrix
    realdp *I = new realdp[N];
    for(integer i = 0; i < N; i++) I[i] = 0.0;
    for(integer i = 0; i < n; i++) I[i*n+i] = 1.0;

    
    integer iter = 0;
    
    
     if(Debug >= 3)
     {
         // compute distance between current iterate and the true solution
        disSeries[iter] = distwo5(X1, Xtrue, n);
     }
    
         
// Start the loop
    for (integer iter = 0; iter < maxiter; iter++)
    {
        
        // get cholesky of X0 and stored in Lx
        copy_(&N, X1, &GLOBAL::IONE, Lx, &GLOBAL::IONE);
        potrf_(GLOBAL::L, &n, Lx, &n, &info);
        for (integer j = 0; j < n; j++)
        {
            for (integer k = j + 1; k < n; k++)
            {
                Lx[j + k * n] = 0; //manually make the upper triangle 0: X0 = X0 * X0^T
            }
        }
        

        Delta = 0.0;
        for(integer h = 0; h < N; h++) S[h] = 0.0; // S is used to store the gradient
        
        
        for (integer i = 0; i < num; i++)
        {
            // solve linear system R0*Z=Rs, then Z=R0^{-1}Rs
            copy_(&N, Rs + i * N, &GLOBAL::IONE, Z + i * N, &GLOBAL::IONE);
            trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &n, &n, Lx, &n, Z + i * N, &n, &info);
            
            // compute Z{i} * Z{i}' and store in Z{i}
            copy_(&N, Z + i * N, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
            gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Z + i * N, &n);
            
            // eigenvalue decomposition Z{i}
            Matrix MMt(Z + i * N, n, n);
            Matrix::EigenSymmetricM(GLOBAL::L, MMt, E, V);
            
            // compute the condition number of Z
            ch = eigenvalues[n-1]/eigenvalues[0];
            
            Delta = Delta + ((ch + 1.0) * log(ch))/(ch - 1.0);
            
            //compute Z <- log(Z'*Z)
            copy_(&N, eigenvectors, &GLOBAL::IONE, VD.matrix, &GLOBAL::IONE);
            for (integer i = 0; i < n; i++)
            {
                a = log(eigenvalues[i]);
                scal_(&n, &a, eigenvectors + i * n, &GLOBAL::IONE);
            }
            Matrix MMt2(Z + i * N, n, n);
            Matrix::DGEMM(GLOBAL::DONE, V, false, VD, true, GLOBAL::DZERO, MMt2);
            
            // Z <- (Z + Z')/2
            copy_(&N, Z + i * N, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
            gemm_(GLOBAL::T, GLOBAL::N, &n, &n, &n, &onehalf, tmp, &n, I, &n, &onehalf, Z + i * N, &n);
            
            // S = S + Z
            gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, Z + i * N, &n, I, &n, &GLOBAL::DONE, S, &n);
        }        
        
        
        // compute stepsize
        if (StepType == 1) 
        {
            stepsize = 2.0/Delta;  // RL
        }
        else if (StepType == 0)
        {
            stepsize = 4.0/(2.0 * num + Delta); //QR
        }
        
        stepsizeSeries[iter] = stepsize;
        
        
        // compute Xk <- R0 * S * R0'//
        // S <- R0 * S
        copy_(&N, S, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, Lx, &n, tmp, &n, &GLOBAL::DZERO, S, &n);
        
        // S <- S * R0'
        copy_(&N, S, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, Lx, &n, &GLOBAL::DZERO, S, &n);
        
        
        // ----------------- compute the norm of gradient
        // compute |S| = tr(S * X^{-1} * S * X^{-1})
        // compute Y1 <- L^{-1}S by solving L*Y1 = S
        copy_(&N, S, &GLOBAL::IONE, Y, &GLOBAL::IONE);
        trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &n, &n, Lx, &n, Y, &n, &info);
        
        // transpose Y1
        copy_(&N, Y, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        gemm_(GLOBAL::T, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, tmp, &n, I, &n, &GLOBAL::DZERO, Y, &n);
        
        // compute L * Y = Y1
        trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &n, &n, Lx, &n, Y, &n, &info);
        
        // compute Y <- Y * Y
        copy_(&N, Y, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Y, &n);
        
        gradSeries[iter] = 0.0;
        for(integer j = 0; j < n; j++) gradSeries[iter] = gradSeries[iter] + Y[j*n+j];
        gradSeries[iter] = std::sqrt(gradSeries[iter]);
        
//        printf("iter: %d, |grad|: %.4e\n", iter, gradSeries[iter]);
        
        timeSeries[iter] = static_cast<double>(getTickCount() - starttime) / CLK_PS; // record the time for each iteration                
        
        
        // S <- theta * S
        copy_(&N, S, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
        gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &stepsize, tmp, &n, I, &n, &GLOBAL::DZERO, S, &n);
        

        //Retraction:  S <- X0 + S
        if (Retraction == 1)
        {
          gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, X1, &n, I, &n, &GLOBAL::DONE, S, &n);
        }
        else if(Retraction == 2)
        {
            // X2 <- X1 + eta + 0.5 * eta * X^{-1} * eta
            copy_(&N, S, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
            // Solve the linear system L X = E, i.e., X = Lx^{-1} eta. The solution X is stored in LiE
            trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &n, &n, Lx, &n, tmp, &n, &info);
            
            //Compute X2 = LiE^T LiE = E L^{-T} L^{-1} E
            gemm_(GLOBAL::T, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, LiStmp, &n);
            
            /*X2 <-- 0.5 * X2*/
            scal_(&N, &onehalf, LiStmp, &GLOBAL::IONE);
            
            /*S <-- S + LiStmp */
            gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, LiStmp, &n, I, &n, &GLOBAL::DONE, S, &n);
            
            /*S <-- x + S*/
            gemm_(GLOBAL::N, GLOBAL::N, &n, &n, &n, &GLOBAL::DONE, X1, &n, I, &n, &GLOBAL::DONE, S, &n);
        }
                
            if(Debug >= 3 && iter > 0)
            {
                // compute distance between current iterate and identity matrix
                disSeries[iter] = distwo5(X1, Xtrue, n);
                truestepsizeSeries[iter] = distwo5(X1, S, n);
            }
        
//        printf("dis: %.4e\n", disSeries[iter]);
  
        
        // Update X_old <- X_new: S is the new iterate, initial_X is the old iterate
        copy_(&N, S, &GLOBAL::IONE, X1, &GLOBAL::IONE);

        // check stopping criteria
        if(gradSeries[iter] < TOL)
        {
            enditer = iter;
            break;
        }
        
    }
    
    ComTime = static_cast<double>(getTickCount() - starttime) / CLK_PS;

    

    // create output
    plhs[0] = mxCreateDoubleMatrix(maxiter, 1, mxREAL); // computation time
    plhs[1] = mxCreateDoubleScalar(ComTime); //  total computational time
    plhs[2] = mxCreateDoubleScalar(maxiter); // number of iterations
    
    plhs[3] = mxCreateDoubleMatrix(N, 1, mxREAL); // X
    double *output3 = mxGetPr(plhs[3]);
    for(integer i = 0; i < N; i++)
        output3[i] = initial_X[i];
//    copy_(&N, initial_X, &GLOBAL::IONE, output3, &GLOBAL::IONE);
    
    plhs[4] = mxCreateDoubleMatrix(maxiter, 1, mxREAL); // distance
    plhs[5] = mxCreateDoubleMatrix(maxiter, 1, mxREAL); // stepsize
    plhs[6] = mxCreateDoubleMatrix(maxiter, 1, mxREAL); // true stepsize
    plhs[7] = mxCreateDoubleMatrix(maxiter, 1, mxREAL); // grad
    
    
    double *plhstime = mxGetPr(plhs[0]);
    double *plhsdis = mxGetPr(plhs[4]);
    double *plhsstepsize = mxGetPr(plhs[5]);
    double *plhstruestepsize = mxGetPr(plhs[6]);
    double *plhsgrad = mxGetPr(plhs[7]);

    for(integer i = 0; i < maxiter; i++)
    {
        plhstime[i] = timeSeries[i];
        plhsdis[i] = disSeries[i];
        plhsstepsize[i] = stepsizeSeries[i];
        plhstruestepsize[i] = truestepsizeSeries[i];
        plhsgrad[i] = gradSeries[i];
//         printf("there: %.4e, %.4e, %.4e, %.4e\n", timeSeries[i], disSeries[i], stepsizeSeries[i], truestepsizeSeries[i]);
    }
      
    delete[] timeSeries;
    delete[] disSeries;  
    delete[] stepsizeSeries;
    delete[] truestepsizeSeries;
    delete[] gradSeries;
    delete[] Z;
    delete[] S;
    delete[] X1;
    delete[] Rs;
    delete[] I;
    delete[] LiStmp;
    delete[] Lx;
    delete[] Y;
    delete[] Xtrue;
    
    
    std::map<integer *, integer>::iterator iteriter = CheckMemoryDeleted->begin();
	for (iteriter = CheckMemoryDeleted->begin(); iteriter != CheckMemoryDeleted->end(); iteriter++)
	{
		if (iteriter->second != 1)
            printf("Global address: %d", iteriter->first);
            printf(", shared times: %d\n", iteriter->second);
	}
	delete CheckMemoryDeleted;
    return;
    
}

#endif




realdp dis5(realdp *X, integer n)
{
    // compute distance between current iterate and identity matrix
    realdp *eigenvaluesXnew = new realdp[n + n * n];
    realdp *eigenvectorsXnew = eigenvaluesXnew + n;
    realdp *tmp = new realdp[n * n];
    
    realdp disSeries;
    integer N = n * n;
    
    copy_(&N, X, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
    Matrix Enew(eigenvaluesXnew, n, 1), Vnew(eigenvectorsXnew, n, n);
    Matrix MMtnew(tmp, n, n);
    Matrix::EigenSymmetricM(GLOBAL::L, MMtnew, Enew, Vnew);
    disSeries = 0.0;
    for (integer ii = 0; ii < n; ii++)
    {
        disSeries = disSeries+log(eigenvaluesXnew[ii])*log(eigenvaluesXnew[ii]);
    }
    disSeries = sqrt(disSeries);
    delete[] eigenvaluesXnew;
    delete[] tmp;
    return disSeries;
    
};




realdp distwo5(realdp *X, realdp *Y, integer n)
{
    realdp *Xold = new realdp[n * n];
    realdp *Xnew = new realdp[n * n];
    realdp *tmp = new realdp[n * n];
    realdp *eigenvaluesXold = new realdp[n + n * n];
    realdp *eigenvectorsXold = eigenvaluesXold + n;
    realdp *eigenvaluesXnew = new realdp[n + n * n];
    realdp *eigenvectorsXnew = eigenvaluesXnew + n;
    integer N = n * n;
    integer info;
    realdp truestepsize;
    
    copy_(&N, X, &GLOBAL::IONE, Xold, &GLOBAL::IONE);
    copy_(&N, Y, &GLOBAL::IONE, Xnew, &GLOBAL::IONE);
    
    // get cholesky of Xnew and Xold
    potrf_(GLOBAL::L, &n, Xnew, &n, &info);
    potrf_(GLOBAL::L, &n, Xold, &n, &info);
    for (integer j = 0; j < n; j++)
    {
        for (integer k = j + 1; k < n; k++)
        {
            Xnew[j + k * n] = 0; //manually make the upper triangle 0: X0 = X0 * X0^T
            Xold[j + k * n] = 0;
        }
    }
    // Solve linear system Xnew * W = Xold, i.e. W = Xnew^{-1}*Xold -> Xold
    trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &n, &n, Xnew, &n, Xold, &n, &info);
    
    // Compute Xold <- Xold * Xold'
    copy_(&N, Xold, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
    gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Xold, &n);
    
    // eigenvalue decomposition of Xold
    Matrix Eold(eigenvaluesXold, n, 1), Vold(eigenvectorsXold, n, n);
    Matrix MMtold(Xold, n, n);
    Matrix::EigenSymmetricM(GLOBAL::L, MMtold, Eold, Vold);
    
    truestepsize = 0.0;
    for (integer ii = 0; ii < n; ii++)
    {
        truestepsize = truestepsize+log(eigenvaluesXold[ii])*log(eigenvaluesXold[ii]);
    }
    truestepsize = sqrt(truestepsize);
    
    delete[] Xold;
    delete[] Xnew;
    delete[] tmp;
    delete[] eigenvaluesXold;
    delete[] eigenvaluesXnew;
    
    return truestepsize;
};







